
/*
    Copyright (C) 2000, 2009 Vansteenhuyse Hubert <hubert.vansteenhuyse@freecode.be>

    This is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with HXE. If not, see <http://www.gnu.org/licenses/>.
*/

/***********************************************************************************************************************
 ***********************************************************************************************************************
 **	PUBLISH via	http://gnomefiles.org/
 ** software repository of gnome
 ***********************************************************************************************************************
 ***********************************************************************************************************************
 * GTK Implementation of the GUI interface defined in platform.h
 * -------------------------------------------------------------
 * CAIRO is used for rendering text and shapes.
 * PANGO is not used at all.
 * GTK routines are used for widget manipulations.
 * GDK routines are used to retrieve default system values.
 *
 * The code should be GTK 2.3 compatible despite all the GTK depreactions on their core routines.
 * I am considering to leave this GTK port as is and focus to another widely used GUI API for UNIX alike systems.
 * GTK doesn't seem to work on the longer turn without having to modify end-programmer code.
 * This is no longer acceptable.
 *
 * For stability reasons thread usage is NOT implemented!
 *
 ***********************************************************************************************************************
 ***********************************************************************************************************************
 * Schema of the main widget
 * -------------------------
 *
 * +wMain-----------------------------------------+
 * | +vBoxMain----------------------------------+ |
 * | | (wToolBar)                               | |
 * | | -----------------------------------------| |
 * | | +vBoxMainCenter------------------------+ | |
 * | | | (wNoteBook + tabs, ...)              | | |
 * | | |--------------------------------------| | |
 * | | | +vBoxDrawContainer-----------------+ | | |
 * | | | |  vBoxInfoBox (wInfoBoxes, ...)   | | | |
 * | | | |----------------------------------| | | |
 * | | | | +hBoxDrawingArea---------------+ | | | |
 * | | | | | +vBoxDrawingArea-------+  .  | | | | |
 * | | | | | | wDrawingArea         | / \ | | | | |
 * | | | | | |                      | (v  | | | | |
 * | | | | | |                      |  s  | | | | |
 * | | | | | |                      |  c  | | | | |
 * | | | | | |                      |  r  | | | | |
 * | | | | | |                      |  o  | | | | |
 * | | | | | |                      |  l  | | | | |
 * | | | | | |----------------------|  l) | | | | |
 * | | | | | |<     (hscroll)     > | \ / | | | | |
 * | | | | | +----------------------+  .  | | | | |
 * | | | | +------------------------------+ | | | |
 * | | | +----------------------------------+ | | |
 * | | +--------------------------------------+ | |
 * | |------------------------------------------| |
 * | | (hBoxStatus)                             | |
 * | +----------------------------------------+ | |
 * +----------------------------------------------+
 *
 * Widgets between parenthesis can be hidden or shown.
 * Horizontal and Vertical Boxes are never visible.
 * wDrawingArea is drawing area widgets.
 * vBoxInfoBox is used to attach InfoBox widgets
 * (small information on screen e.g. display the progress).
 * wNotebook is shown or hidden transparently to the
 * end developer when the notebook is empty.
 * The block vBoxDrawContainer is re-parent transparently
 * to the end developer into the pages of the notebook and
 * vBoxMainCenter.
 **/

#include <gdk/gdk.h>
#include <gtk/gtk.h>
//#include <gtk/gtkunixprint.h>

#include <glib.h>
#include <cairo.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/resource.h>
#include <unistd.h>

#include <sys/stat.h>
#include <sys/types.h>

#include "platform.h"

/**
 * Definitions and Macros
 **/
#define PLATFORM_GTK_FG_GC(widget) ( widget->style->fg_gc[gtk_widget_get_state(widget)])
#define PLATFORM_GTK_BG_GC(widget) ( widget->style->bg_gc[gtk_widget_get_state(widget)])

#if GTK_MAJOR_VERSION >= 3			/*GTK version >= 3*/
	#define PIXMAP_CAIRO_SURFACE
	typedef cairo_surface_t PIXELMAP;
#else								/*GTK version < 3*/
	typedef GdkPixmap PIXELMAP;
#endif

/**
 * Forward declarations
 */
int maxx(void);
int maxy(void);
int on_destroy(GtkWidget *widget, gpointer data);

/**
 * Local values
 **/
PLATFORM_APP *fwplat = 0;
int __sw_platform_exit = 0, __sw_platform_init = 0, __exit_code = 0, __sw_got_focus = 1;
float	__line_width = 1.f, __pixel_size = 1.f;
int sw_in_keyboardroutine = 0, __k_ctrl_in = 0, __k_shift_in = 0, __k_alt_in = 0;
GdkEventKey *__key;

char			*defFont = (char*)"Sans",
				*defFixed = (char*)"Monospace";

char 	fileRootFolder[1024 * 64], selectedFont[1024 * 2];
int		swValidTootFolder = 0;

/**
 * System colors
 **/
C_RGBA		color_fg, color_bg,				/*system color client text area*/
			color_btn_bg, color_btn_fg,		/*system color button*/
			color_sel_bg, color_sel_fg,		/*system color selected text*/
			color_tip_bg, color_tip_fg,		/*system color tool tip*/

			colorf, colorb;					/*currently selected colors*/

int			SysMaxColorVal = 0,
			SysColorDepth = -1;

double		__alpha = 1.f, __intensity = 1.f;

/**
 * Mouse and keyboard
 **/
GdkCursor 	*cursors[3];	//0 = NULL default, 1 = IBEAM, 2 ...
GdkModifierType mouseB = 0;
int kbd = 0, mouseX = 0, mouseY = 0;

/**
 * Cairo Context
 **/
cairo_t					*CairoCtx = 0;

/**
 * Main Widgets
 **/
GtkWidget		*wMain = 0,
				*vBoxMain = 0,
				*vBoxMainCenter = 0,
				*vBoxDrawContainer = 0,
				*vBoxInfoBox = 0,
				*wDrawingArea = 0,
				*hBoxDrawingArea = 0,
				*vBoxDrawingArea = 0,
				*hBoxStatus = 0;
GtkNotebook		*wNoteBook = 0;
GtkToolbar		*wToolBar = 0;
GtkVScrollbar 	*vscroll = 0;
GtkHScrollbar 	*hscroll = 0;
GtkAdjustment 	*hadj = 0,
				*vadj = 0;
PIXELMAP 		*pixmap = 0;


/**
 * caret
 **/
#define CARET_BLINK 			"gtk-cursor-blink"
#define CARET_BLINK_TIME 		"gtk-cursor-blink-time"
#define CARET_BLINK_TIMEOUT 	"gtk-cursor-blink-timeout"
#define CARET_INSERT_WDT 		1
#define USE_CARET_TRANSPARENT	/*if NOT defined an opaque caret cursor will be used instead*/
GdkRectangle 	CaretPos;
IPLAT			__caret_mode = 0,
				__caret_onoff = CARET_OFF;
int 			__caret_sw_show = 0,
				__caret_sw_blink = 0,
				__caret_sw_moved = 0;
void*			__caret_img = 0;

int				__caret_blink_time = 0;


cairo_font_options_t 	*fo = 0;
char 	current_fontface[TXTOUT_MAX];
double	font_dpi = 96, dpi_scale = 1.f, current_font_size = 0, requested_font_size = 0;
int		current_font_wdt = 0, current_font_hgt = 0;
int __sw_opaque = 0, __sw_bold = 0, __sw_italic = 0, __sw_underlined = 0;

/**
 * Controls
 **/
LIST *lstToolBar = 0,*lstStatusBar = 0;

/***********************************************************************************************************
 * LOCAL ROUTINES
 ***********************************************************************************************************/
int load_systems_defaults(){
	return 0;
}

/* Draw context used to draw behind the scenes */
cairo_t* local_create_dc_from_pixelmap(PIXELMAP *pm){
	cairo_t *cr = 0;
	if(pm){
		#ifdef PIXMAP_CAIRO_SURFACE
		cr = cairo_create(pm);		/*create a cairo context from a cairo_surface_t*/
		#else
		cr = gdk_cairo_create(pm);	/*create a cairo context from a GdkPixmap*/
		#endif
	}
	return cr;
}

/* Draw context used to draw directly on the screen */
cairo_t* local_create_dc_from_window(GdkWindow *window){
	cairo_t *cr = 0;
	if(window)
		cr = gdk_cairo_create(window);
	return cr;
}

cairo_t* local_destroy_dc(cairo_t *cr){
	if(cr) cairo_destroy(cr);
	return 0;
}

void gtk_impl_exit(int exitcode){
	__exit_code = exitcode;
	on_destroy(wMain, 0);
}

double total_memory(void){
	double mb = 0;
	struct rlimit rl;
		if(!getrlimit(RLIMIT_AS, &rl))
			mb = (double)rl.rlim_cur / (double)(1024 * 1024);
		return mb;
}

gint int_get_setting(GtkWidget *widget, const char *name){
	gint n = 0;
	GtkSettings *settings;
	if(widget){
		settings = gtk_widget_get_settings(widget);
		g_object_get(settings, name, &n, NULL);
	}
	return n;
}
/* returned value must be freed*/
gchar* char_get_setting(GtkWidget *widget, const char *name){
	gint  n = 0;
	gchar *ch = 0;
	GtkSettings *settings;
	if(widget){
		settings = gtk_widget_get_settings(widget);
		g_object_get(	settings,
						name, &n,
						name, &ch,
						NULL);
	}
	return ch;
}
gchar* char_get_setting_free(gchar *strval){
	if(strval)g_free(strval);
	return 0;
}

int gtk_send_redraw(PLAT_RECT *r){
	GdkRectangle gr;
	if(!wDrawingArea || !GTK_IS_WIDGET(wDrawingArea) || !gtk_widget_is_drawable(wDrawingArea)) return 0;
	if(!r){
		gtk_widget_queue_draw(wDrawingArea);
	} else {
		gr.x = r->x; gr.y = r->y;
		gr.width = PLAT_RECT_WP(r);
		gr.height = PLAT_RECT_HP(r);
		gtk_widget_draw (wDrawingArea, &gr);
	}

	return 1;
}

void doevents(){
	while(gtk_events_pending()){
		gtk_main_iteration();
	};
}

struct timeval __impl_time(){
	struct timeval tv;
	if(gettimeofday(&tv, 0) == -1){
		exit(0);
		tv.tv_sec = 0;
		tv.tv_usec = 0;
	}
	return tv;
}

#define FSEC 1.0E-6
#define USEC 1.0E+6
float __impl_time_delta(struct timeval tv){
	struct timeval now = __impl_time();
	float 	sec = (now.tv_sec - tv.tv_sec - 1),
			usec = FSEC * (USEC + now.tv_usec - tv.tv_usec);
	//printf("sec %.3f, usec %.3f = %.3f\n", sec, usec, sec + usec);
	return sec + usec;
}

int __impl_gtk_timer_call(int msec, __timer_routine tm_routine, void *arg){
	uint gres = 0;

	if(msec && tm_routine){
		gres = g_timeout_add(msec, tm_routine, arg);
	}

	return gres;
}

/***********************************************************************************************************
 * FONT ROUTINES
 ***********************************************************************************************************/
int fontwdt(void){
	return current_font_wdt;
}

int fonthgt(void){
	return current_font_hgt;
}

void updates_font_related(cairo_t *cr){
	if(cr == CairoCtx){
	}
}

void cairo_setfont_execute(){
	double pix_x = 1.f, pix_y = 1.f;

	PangoFontDescription	*desc;
	gchar					*fnamecairo;
	cairo_font_extents_t 	ft;
	if(!CairoCtx) return;
	desc = pango_font_description_from_string(current_fontface);
	if(!desc) return;

	__pixel_size = 1.f;
	cairo_device_to_user_distance(CairoCtx, &pix_x, &pix_y);
	if(pix_x > pix_y)	__pixel_size = pix_x;
	else				__pixel_size = pix_y;

	font_dpi = wMain ? pango_cairo_context_get_resolution (gtk_widget_get_pango_context (wMain)) : 0;
	if(font_dpi > 0)	dpi_scale = font_dpi / 72.f;
	else				dpi_scale = 1.f;

	current_font_size = (requested_font_size ? requested_font_size : pango_font_description_get_size(desc) / PANGO_SCALE );
	current_font_size *= dpi_scale;

	fnamecairo = (gchar*)pango_font_description_get_family(desc);
	cairo_select_font_face(CairoCtx, fnamecairo, __sw_italic ? CAIRO_FONT_SLANT_ITALIC : CAIRO_FONT_SLANT_NORMAL, __sw_bold ? CAIRO_FONT_WEIGHT_BOLD : CAIRO_FONT_WEIGHT_NORMAL);
	cairo_set_font_size(CairoCtx, current_font_size);
	if(fo) cairo_set_font_options(CairoCtx, fo);

	cairo_font_extents(CairoCtx, &ft);
	current_font_hgt = ft.height + 2;	/* + 2 is the extra space between lines, also makes the cursor bigger thus more visible*/
	current_font_wdt = ft.ascent - ft.descent;

	pango_font_description_free(desc);
	updates_font_related(CairoCtx);
}

void font_use(IPLAT ifont){
	PangoContext 			*pc = 0;
	PangoFontDescription	*pd = 0;
	gint					sze = 12;

	if(wMain){
		pc = gtk_widget_get_pango_context (wMain);
		if(pc){
			pd = pango_context_get_font_description (pc);
			if(pd) sze = pango_font_description_get_size (pd) / PANGO_SCALE;
		}
	}

	if(ifont == FONT_MONOSPACE)
		snprintf(current_fontface, TXTOUT_MAX, "%s %d", defFixed, sze);
	else if(ifont == FONT_SYSTEM)
		snprintf(current_fontface, TXTOUT_MAX, "%s %d", defFont, sze);
	else {
		if(strlen(selectedFont))
			snprintf(current_fontface, TXTOUT_MAX, "%s", selectedFont);
		else
			snprintf(current_fontface, TXTOUT_MAX, "%s %d", defFont, sze);
	}
	cairo_setfont_execute();

}

void font_face(const char *face, int sze){
	if(sze <= 0) sze = 13;
	requested_font_size = sze;
	snprintf(current_fontface, TXTOUT_MAX, "%s", face);
	cairo_setfont_execute();
}

/***********************************************************************************************************
 * COLOR ROUTINES
 ***********************************************************************************************************/
void load_system_color_sizes(){
	GdkColor	c;
	//X11 color names
	gdk_color_parse((const gchar*)"White", &c);
	SysMaxColorVal = c.red;
	if(fwplat) fwplat->gui.color_maxval = SysMaxColorVal;
}

#define FROM_GDK_COLOR(out, in){ out.r = in.red; out.g = in.green; out.b = in.blue; out.a = SysMaxColorVal; }

void load_system_colors(){
	GtkWidget	*wControl = 0;
	GtkStyle 	*gs = 0;

	load_system_color_sizes();

	if(wDrawingArea){
		gtk_widget_ensure_style(wDrawingArea);
		gs = gtk_widget_get_style(wDrawingArea);
		if(gs){
			color_bg.r = gs->base[GTK_STATE_NORMAL].red;
			color_bg.g = gs->base[GTK_STATE_NORMAL].green;
			color_bg.b = gs->base[GTK_STATE_NORMAL].blue;

			color_fg.r = gs->fg[GTK_STATE_NORMAL].red;
			color_fg.g = gs->fg[GTK_STATE_NORMAL].green;
			color_fg.b = gs->fg[GTK_STATE_NORMAL].blue;

			color_sel_bg.r = gs->base[GTK_STATE_SELECTED].red;
			color_sel_bg.g = gs->base[GTK_STATE_SELECTED].green;
			color_sel_bg.b = gs->base[GTK_STATE_SELECTED].blue;

			color_sel_fg.r = gs->fg[GTK_STATE_SELECTED].red;
			color_sel_fg.g = gs->fg[GTK_STATE_SELECTED].green;
			color_sel_fg.b = gs->fg[GTK_STATE_SELECTED].blue;
		}

		wControl = gtk_window_new(GTK_WINDOW_POPUP);
		gtk_widget_set_name(wControl, "gtk-tooltip");
		if(wControl){
			gtk_widget_ensure_style(wControl);
			gs = gtk_widget_get_style(wControl);
			if(gs){
				color_tip_bg.r = gs->bg[GTK_STATE_NORMAL].red;
				color_tip_bg.g = gs->bg[GTK_STATE_NORMAL].green;
				color_tip_bg.b = gs->bg[GTK_STATE_NORMAL].blue;

				color_tip_fg.r = gs->fg[GTK_STATE_NORMAL].red;
				color_tip_fg.g = gs->fg[GTK_STATE_NORMAL].green;
				color_tip_fg.b = gs->fg[GTK_STATE_NORMAL].blue;
			}
			gtk_widget_destroy(wControl);
		}

		/*BUTTON COLOR*/
		wControl = gtk_window_new(GTK_WINDOW_POPUP);
		gtk_widget_set_name(wControl, "gtk-button");
		if(wControl){
			gtk_widget_ensure_style(wControl);
			gs = gtk_widget_get_style(wControl);
			if(gs){
				color_btn_bg.r = gs->bg[GTK_STATE_NORMAL].red;
				color_btn_bg.g = gs->bg[GTK_STATE_NORMAL].green;
				color_btn_bg.b = gs->bg[GTK_STATE_NORMAL].blue;

				color_btn_fg.r = gs->fg[GTK_STATE_NORMAL].red;
				color_btn_fg.g = gs->fg[GTK_STATE_NORMAL].green;
				color_btn_fg.b = gs->fg[GTK_STATE_NORMAL].blue;
			}
			gtk_widget_destroy(wControl);
		}
	}
}

void set_cairo_color(cairo_t *cr, C_RGBA *col){
	double div = (double)(SysMaxColorVal), r, g, b;
	if(!cr || !col) return;
	r = (double)(col->r) / div;
	g = (double)(col->g) / div;
	b = (double)(col->b) / div;
	cairo_set_source_rgba(cr, r * __intensity, g * __intensity, b * __intensity, __alpha);
}

void set_fcolor(double nr, double ng, double nb){
	colorf.r = CLAMP(nr, 0.0, 1.0) * SysMaxColorVal;
	colorf.g = CLAMP(ng, 0.0, 1.0) * SysMaxColorVal;
	colorf.b = CLAMP(nb, 0.0, 1.0) * SysMaxColorVal;
}
void set_bcolor(double nr, double ng, double nb){
	colorb.r = CLAMP(nr, 0.0, 1.0) * SysMaxColorVal;
	colorb.g = CLAMP(ng, 0.0, 1.0) * SysMaxColorVal;
	colorb.b = CLAMP(nb, 0.0, 1.0) * SysMaxColorVal;
}
void set_alpha(double na){
	__alpha = CLAMP(na, 0.0, 1.0);
}
void set_intensity(double ni){
	__intensity = CLAMP(ni, 0.0, 1.0);

}
void set_sys_color(IPLAT f, IPLAT b){
	double	dr, dg, db, da;
	C_RGBA 	F, B, *fg = &color_fg, *bg = &color_bg;

	/*foreground*/
	if(f >= SYSCOL_BLACK && f <= SYSCOL_CYANL){
		plat_normal_color(f, &dr, &dg, &db, &da);
		F.r = dr * SysMaxColorVal;
		F.g = dg * SysMaxColorVal;
		F.b = db * SysMaxColorVal;
		F.a = da * SysMaxColorVal;
		fg = &F;
	} else if(f == SYSCOL_TEXT_F)		fg = &color_fg;
	else if(f == SYSCOL_TEXT_B)			fg = &color_bg;
	else if(f == SYSCOL_TEXT_SELETED_F)	fg = &color_sel_fg;
	else if(f == SYSCOL_TEXT_SELETED_B)	fg = &color_sel_bg;
	else if(f == SYSCOL_TOOLTIP_F)		fg = &color_tip_fg;
	else if(f == SYSCOL_TOOLTIP_B)		fg = &color_tip_bg;
	else if(f == SYSCOL_BUTTON_F)		fg = &color_btn_fg;
	else if(f == SYSCOL_BUTTON_B)		fg = &color_btn_bg;

	/*background*/
	if(b >= SYSCOL_BLACK && b <= SYSCOL_CYANL){
		plat_normal_color(b, &dr, &dg, &db, &da);
		B.r = dr * SysMaxColorVal;
		B.g = dg * SysMaxColorVal;
		B.b = db * SysMaxColorVal;
		B.a = da * SysMaxColorVal;
		bg = &B;
	} else if(b == SYSCOL_TEXT_F)				bg = &color_fg;
	else if(b == SYSCOL_TEXT_B)			bg = &color_bg;
	else if(b == SYSCOL_TEXT_SELETED_F)	bg = &color_sel_fg;
	else if(b == SYSCOL_TEXT_SELETED_B)	bg = &color_sel_bg;
	else if(b == SYSCOL_TOOLTIP_F)		bg = &color_tip_fg;
	else if(b == SYSCOL_TOOLTIP_B)		bg = &color_tip_bg;
	else if(b == SYSCOL_BUTTON_F)		bg = &color_btn_fg;
	else if(b == SYSCOL_BUTTON_B)		bg = &color_btn_bg;

	colorb = *bg;
	colorf = *fg;
}

void swap_colors(void){
	C_RGBA	color = colorf;

	colorf = colorb;
	colorb = color;
}

/***********************************************************************************************************
 * SHAPE RENDERING ROUTINES
 ***********************************************************************************************************/
void rectangle(int swFilled, int x, int y, int wdt, int hgt){
	if(!CairoCtx) return;
	if(swFilled)
		set_cairo_color(CairoCtx, &colorb);
	else
		set_cairo_color(CairoCtx, &colorf);

	cairo_rectangle(CairoCtx, x, y, wdt, hgt);
	if(swFilled)
		cairo_fill(CairoCtx);
	else
		cairo_stroke(CairoCtx);
}
void arc(int x, int y, double r, double start, double end){
	if(!CairoCtx) return;
	set_cairo_color(CairoCtx, &colorf);
	cairo_arc_negative(CairoCtx, x, y, r, start, end);
	cairo_stroke(CairoCtx);
}

void circle(int swFilled, int x, int y, int r){
	arc(x, y, r, 2 * M_PI, 0);
}

void polygon(int swFilled, int elements, double *x, double *y){
	int i = 0;
	if(!CairoCtx || !elements || !x || !y) return;
	set_cairo_color(CairoCtx, &colorf);

	cairo_new_path(CairoCtx);
	if(elements){
		cairo_move_to(CairoCtx, x[i], y[i]);
	}
	for(i = 1; i < elements; i++){
		cairo_line_to(CairoCtx, x[i], y[i]);
	}

	cairo_close_path(CairoCtx);
	if(swFilled)
		cairo_fill(CairoCtx);
	else
		cairo_stroke(CairoCtx);

}

void linepath(int elements, double *x, double *y){
	if(!CairoCtx || elements < 2) return;
	set_cairo_color(CairoCtx,  &colorf);
	cairo_set_line_width(CairoCtx, __pixel_size * __line_width);

	cairo_move_to(CairoCtx, x[0], y[0]);
	for(int i = 1; i < elements; i++){
		cairo_line_to(CairoCtx, x[i], y[i]);
	}
	cairo_stroke(CairoCtx);
}


void line(int x1, int y1, int x2, int y2){
	if(!CairoCtx) return;
	set_cairo_color(CairoCtx,  &colorf);
	cairo_set_line_width(CairoCtx, __pixel_size * __line_width);
	cairo_move_to(CairoCtx, x1, y1);
	/*A cairo line is relative to last cairo_move_to(xr, x, y) call */
	cairo_rel_line_to(CairoCtx, x2 - x1, y2 - y1);
	cairo_stroke(CairoCtx);
}

void putpixel(int x, int y){
	if(!CairoCtx) return;
	set_cairo_color(CairoCtx,  &colorf);
	cairo_rectangle(CairoCtx, x, y, __pixel_size, __pixel_size);
	cairo_fill(CairoCtx);
}

double dashes[] = {
		10.0,  /* ink */
		10.0,  /* skip */
		10.0,  /* ink */
		10.0   /* skip*/
};
void dashline(int x1, int y1, int x2, int y2){
	int ndash = sizeof (dashes) / sizeof(dashes[0]);
	double offset = -10.0;
	if(!CairoCtx) return;
	cairo_set_dash(CairoCtx, dashes, ndash, offset);
	line(x1, y1, x2, y2);
	cairo_set_dash(CairoCtx, 0, 0, 0);
}

/***********************************************************************************************************
 * TEXT RENDERING ROUTINES
 ***********************************************************************************************************/
void clrscr(){
	if(!CairoCtx) return;
	set_cairo_color(CairoCtx, &colorb);
	cairo_paint(CairoCtx);
}

void textmode(IPLAT text_mode){
	if(!wDrawingArea) return;
	if(text_mode == TEXT_OPAQUE)
		__sw_opaque = 1;
	else if(text_mode == TEXT_TRANSPARENT)
		__sw_opaque = 0;
	else if(text_mode == TEXT_BOLD)
		__sw_bold = 1;
	else if(text_mode == TEXT_ITALIC)
		__sw_italic = 1;
	else if(text_mode == TEXT_UNDERLINED)
		__sw_underlined = 1;
	else if(text_mode == TEXT_NORMAL)
		__sw_bold = __sw_italic = __sw_underlined = 0;
	cairo_setfont_execute();
}

int cpixels(int *wdt, int *hgt, char *str){
	int n = (str ? strlen(str) : 0);
	cairo_text_extents_t ct;

	if(!CairoCtx) return 0;

		if(!n){
			if(wdt) *wdt = 0;
			if(hgt) *hgt = 0;
		} else {
			cairo_text_extents(CairoCtx, str, &ct);
			if(!ct.x_advance) ct.x_advance = n * fontwdt();
			if(wdt) *wdt = (int)( ct.x_advance);
			if(hgt) *hgt = fonthgt();
		}
	return 1;
}

int lenpixels(int *wdt, int *hgt, char *fmt, ...){
	char buf[TXTOUT_MAX];
	va_list args;

	if(!CairoCtx && !fmt) return 0;

	va_start(args, fmt);
	vsnprintf(buf, TXTOUT_MAX, fmt, args);
	va_end(args);

	return cpixels(wdt, hgt, buf);
}

void cout(int x, int y, char *str){
	cairo_font_extents_t ft;
	int n = (str ? strlen(str) : 0), xx, yy;


	if(n && CairoCtx){

		cpixels(&xx, &yy, str);
		if(__sw_opaque)	rectangle(1, x, y, xx, yy);

		cairo_font_extents(CairoCtx, &ft);
		set_cairo_color(CairoCtx, &colorf);
		cairo_move_to(CairoCtx, x, y + yy - ft.descent);
		cairo_show_text(CairoCtx, str);

		//if(CAIRO_STATUS_SUCCESS != cairo_status(CairoCtx)) printf("%s | %s\n", cairo_status_to_string(cairo_status(CairoCtx)), str);

		cairo_stroke(CairoCtx);
		if(__sw_underlined){
			if(__sw_bold)
				__line_width = 3.0f;
			else
				__line_width = 1.0f;
			line(x, y + yy, x + xx, y + yy);
			__line_width = 1.0f;
		}
	}
}

void textout(int x, int y, char *fmt, ...){
	int n;
	char buf[TXTOUT_MAX];
	va_list args;

	if(!wDrawingArea || !fmt) return;
	va_start(args, fmt);
	n = vsnprintf(buf, TXTOUT_MAX, fmt, args);
	va_end(args);
	if(n) cout(x, y, buf);
}

/***********************************************************************************************************
 * BITMAP ROUTINES
 ***********************************************************************************************************/
PIXELMAP* pixelmap_create(GdkWindow *window, int wdt, int hgt){
	if(!window) return 0;
	#ifdef PIXMAP_CAIRO_SURFACE
	 	 return gdk_window_create_similar_surface(window, CAIRO_CONTENT_COLOR_ALPHA, wdt, hgt);
	#else
	 	 return gdk_pixmap_new(window, wdt, hgt, -1);
	#endif
}

PIXELMAP* pixelmap_create_from_window(GdkWindow *window){
	int w, h;
	if(!window) return 0;
	gdk_drawable_get_size(window, &w, &h);
	return pixelmap_create(window, w, h);
}


PIXELMAP* pixelmap_destroy(PIXELMAP* pm){
	if(pm){
		#ifdef PIXMAP_CAIRO_SURFACE
			cairo_surface_destroy(pm);
		#else
			 g_object_unref(pm);
		#endif
	}
	return 0;
}

void* screen_copy(int x, int y, int wdt, int hgt){
	cairo_t 	*cr;
	PIXELMAP 	*pmap;

	if(__sw_platform_exit || !wDrawingArea) return 0;

	/*create the GdkPixmap and cairo context*/
	pmap = pixelmap_create(wDrawingArea->window, wdt, hgt);
	if(!pmap) return 0;
	cr = local_create_dc_from_pixelmap(pmap);
	if(!cr){
		g_object_unref(pmap);
		return 0;
	}

	cairo_scale (cr, 1, 1);
	/*copy the screen to a GdkPixmap using cairo*/
	#ifdef PIXMAP_CAIRO_SURFACE
		cairo_set_source_surface(cr, pmap, -x, -y);
		//cairo_set_source_surface(cr, pmap, x, y);
	#else
		gdk_cairo_set_source_pixmap(cr, wDrawingArea->window, -x, -y); 	//<---<< -x, -y ok wanneer gn double buffer
		//gdk_cairo_set_source_pixmap(cr, wDrawingArea->window, x, y);		//<---<< x, y +/- ok wanneer wel double buffer
	#endif

	cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
	cairo_pattern_set_extend(cairo_get_source(cr), CAIRO_EXTEND_NONE);
	cairo_rectangle(cr, 0, 0, x + wdt, y + hgt);
	cairo_fill(cr);
	cr = local_destroy_dc(cr);

	return pmap;
}

void screen_put(void *img, int scrx, int scry, int imgx, int imgy, int wdt, int hgt){
	cairo_t *cr;

	if(__sw_platform_exit || !img || !wDrawingArea) return;

	/*draw the image using GdkPixmap and cairo*/
	cr = local_create_dc_from_window(wDrawingArea->window);
	if(!cr) return;
	cairo_scale (cr, 1, 1);
	gdk_cairo_set_source_pixmap(cr, (PIXELMAP*)img, scrx, scry);
	cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
	cairo_pattern_set_extend(cairo_get_source(cr), CAIRO_EXTEND_NONE);
	cairo_rectangle(cr, scrx, scry, wdt, hgt);
	cairo_fill(cr);
	cr = local_destroy_dc(cr);
}

void* screen_destroy(void *img){
	img = pixelmap_destroy(img);
	return img;
}

GdkPixbuf* screen_get_pixels(GtkWidget *widget, int x, int y, int wdt, int hgt){
	GdkPixbuf 	*pixbuf = 0;
	if(!__sw_platform_exit && widget && widget->window){
		#if GTK_MAJOR_VERSION >= 3			/*GTK version >= 3*/
		pixbuf = gdk_pixbuf_get_from_window(wDrawingArea->window, x, y, wdt, hgt);
		#else
		pixbuf = gdk_pixbuf_get_from_drawable(NULL, wDrawingArea->window, NULL, x, y, 0, 0, wdt, hgt);
		#endif
	}
	return pixbuf;
}
/***********************************************************************************************************
 * PRINT ROUTINES (call backs for on_toolbar_item_clicked TOOLBAR_ITEM_DOCPRINT)
 * https://developer.gnome.org/gtk2/stable/gtk2-High-level-Printing-API.html
 ***********************************************************************************************************/
PLAT_RECT __plat_gtk_print_page_area(GtkPageSetup *ps){
	PLAT_RECT 	r;
	GtkUnit 	u = GTK_UNIT_POINTS;

	r.x = gtk_page_setup_get_left_margin (ps, u);
	r.xx = gtk_page_setup_get_page_width(ps, u) + gtk_page_setup_get_right_margin(ps, u);
	r.y = gtk_page_setup_get_top_margin(ps, u);
	r.yy = gtk_page_setup_get_page_height(ps, u) + gtk_page_setup_get_bottom_margin(ps, u);

	return r;
}
gboolean __plat_gtk_print_preview (
		GtkPrintOperation        *operation,
		GtkPrintOperationPreview *preview,
		GtkPrintContext          *context,
		GtkWindow                *parent,
		gpointer                  user_data
){

	return FALSE;
}

int __imp_nPrintStartPage = 0;
void __plat_gtk_print_begin(
		GtkPrintOperation *operation,
		GtkPrintContext   *context,
		gpointer           user_data
){
	GtkPrintSettings *ps = gtk_print_operation_get_print_settings (operation);
	GtkPrintPages 	pp = gtk_print_settings_get_print_pages(ps);
	GtkPageRange 	*pr = 0;
	gint			npr = 0;
	int				swAbort = 0;
	int 			pages = 0;
	PLAT_RECT 		rPaper;

	__imp_nPrintStartPage = 0;

	switch(pp){
	case GTK_PRINT_PAGES_ALL:
		rPaper = __plat_gtk_print_page_area(gtk_print_context_get_page_setup(context));
		pages = onprint_setup(fwplat, &rPaper);
		break;
	case GTK_PRINT_PAGES_RANGES:
		pr = gtk_print_settings_get_page_ranges (ps, &npr);
		if(npr <= 0 || npr > 1 || !pr){
			swAbort = 1;
		} else {
			__imp_nPrintStartPage = pr->start;
			 pages = pr->end + 2 - __imp_nPrintStartPage;
		}
		if(pr) g_free(pr);
		break;
	default:
		swAbort = 1;
		break;
	}

	if(swAbort){
		gtk_print_operation_cancel(operation);
		fwplat->ctl.error_msg("Printing", "Page selection not supported.\nAll pages, 1 page or 1 range are supported.");
		return;
	}
	if(pages <= 0){
		gtk_print_operation_cancel(operation);
		return;
	}

	gtk_print_operation_set_n_pages(operation, pages);
}
void __plat_gtk_print_page(
		GtkPrintOperation *operation,
		GtkPrintContext *context,
		int page_nr
){
	cairo_t	*saveCairoCtx = CairoCtx;
	PLAT_RECT rPaper;

	CairoCtx = gtk_print_context_get_cairo_context(context);

	rPaper = __plat_gtk_print_page_area(gtk_print_context_get_page_setup (context));
	cairo_setfont_execute();
	if(!CairoCtx || onprint_page(fwplat, &rPaper, __imp_nPrintStartPage + page_nr) == 0){
		//cancel print
		gtk_print_operation_cancel(operation);
	}

	CairoCtx = saveCairoCtx;
}

void __plat_gtk_print_end (
		GtkPrintOperation *operation,
		GtkPrintContext   *context,
		gpointer           user_data
){
}
/***********************************************************************************************************
 * CARET ROUTINES (keyboard cursor)
 ***********************************************************************************************************/
void __cared_bg_update(int swRepaintOld, int x, int y){
	if(__caret_onoff == CARET_OFF || !__sw_got_focus){
		if(__caret_img) __caret_img = screen_destroy(__caret_img);
		return;
	}
	if(swRepaintOld)
		screen_put(__caret_img, CaretPos.x, CaretPos.y, 0, 0, CaretPos.width, CaretPos.height);

	__caret_img = screen_destroy(__caret_img);
	__caret_img = screen_copy(x, y, CaretPos.width, CaretPos.height);
}

void __caret_update(){
	__caret_sw_moved = 0;
	if(CairoCtx || !wDrawingArea || __sw_platform_exit) return;
	if(!__caret_sw_show || __caret_onoff == CARET_OFF) return;

	CairoCtx = local_create_dc_from_window(wDrawingArea->window);
	if(!CairoCtx) return;

	if(__caret_sw_blink) {
		if(__sw_got_focus){
			/**
			 * Transparent caret.
			 * Read pixels from screen at caret position, invert them and draw the inverted pixels
			 * Read using GDK, write using CAIRO.
			 **/
			GdkPixbuf 	*pixbuf;
			int x, y, n, m, pos, stride, bps, chn, tail, wdt, hgt;
			int r, g, b, a, p;
			guchar *pixels;

			x = CaretPos.x;
			wdt = __caret_mode == CARET_OVW ? CaretPos.width : CARET_INSERT_WDT;
			y = CaretPos.y;
			hgt = CaretPos.height;
			pixbuf = screen_get_pixels(wDrawingArea, x, y, wdt+2, hgt); //+2 blijkbaar laatste kolom pixels niet goed gecopieerd (geeft blauwe schijn aan cursor INSERT mode)
			if(pixbuf){
				pixels = gdk_pixbuf_get_pixels(pixbuf);
				if(pixels){
					bps = gdk_pixbuf_get_bits_per_sample(pixbuf);
					chn = gdk_pixbuf_get_n_channels(pixbuf);
					stride = gdk_pixbuf_get_rowstride(pixbuf);
					hgt = gdk_pixbuf_get_height(pixbuf);
					tail = stride * hgt;
					wdt = gdk_pixbuf_get_width(pixbuf);
					for(y = 0; y < hgt; y++){
						for(x = 0; x < wdt-2; x++){			//-2, laatste 2 kolomen pixels niet meegerekend cf uitleg hier boven
							pos = stride * y + x * chn;
							r = g = b = a = 0;
							for(n = 0; n < chn; n++){
								pos += n;
								if(pos >= tail) break;
								for(m = 0; m < bps / 8; m++){
									pos += m;
									p = pixels[pos] << (m * bps);
									switch(n){
									case 0:
										r |= p;
										break;
									case 1:
										g |= p;
										break;
									case 2:
										b |= p;
										break;
									case 3:
										a |= p;
										break;
									default:
										break;
									}
								}
							}
							if(pos >= tail) break;
							set_fcolor(	1 - ( (r << bps) / (float)SysMaxColorVal),
										1 - ( (g << bps) / (float)SysMaxColorVal),
										1 - ( (b << bps) / (float)SysMaxColorVal)
							);
							putpixel(CaretPos.x + x, CaretPos.y + y);
						}
						if(pos >= tail) break;
					}
				}
				g_object_unref(pixbuf);
			}
		}
	} else
		screen_put(__caret_img, CaretPos.x, CaretPos.y, 0, 0, CaretPos.width, CaretPos.height);

	CairoCtx = local_destroy_dc(CairoCtx);
}

void caret_set(IPLAT caret_onoff, IPLAT caret_mode, int x, int y, char letter){
	if(x < 0) x = 0;
	if(y < 0) y = 0;

	if(!CairoCtx && (CaretPos.x != x || CaretPos.y != y || caret_mode != __caret_mode)){
		__caret_sw_moved = 1;
		__cared_bg_update(1, x, y);
	}

	if(letter){
		char letters[2];
		letters[0] = letter;
		letters[1] = 0;
		fwplat->gui.lenpixels(&CaretPos.width, &CaretPos.height, letters);
	} else {
		CaretPos.width = fwplat->gui.fontwdt();
		CaretPos.width = fwplat->gui.fonthgt();
	}

	CaretPos.x = x;
	CaretPos.y = y;
	__caret_sw_show = (caret_onoff == CARET_ON);
	__caret_mode = caret_mode;

	__caret_onoff = caret_onoff;

	__caret_sw_blink = 1;
	if(__caret_sw_show) __caret_update();
}

/**
 * gedit cursor blinks like this, don't know why but it is more visible
 * blink time on = time + 1 / 4
 * blink time off = time - 1 / 4
 **/
static gint __caret_blink_routine(gpointer arg){
	__caret_sw_blink = (__caret_sw_blink ? 0 : 1);
	__caret_update();
	if(__caret_sw_blink)
		g_timeout_add(__caret_blink_time + __caret_blink_time / 4, __caret_blink_routine, 0);
	else
		g_timeout_add(__caret_blink_time * 3 / 4, __caret_blink_routine, 0);

	return FALSE;
}

/***********************************************************************************************************
 * MOUSE ROUTINES (mouse cursor)
 ***********************************************************************************************************/
IPLAT i_mouse_cursor = MOUSE_CURSOR_DEFAULT;
void set_mouse_cursor(IPLAT mouse_cursor){
	GdkCursor *cs = cursors[0];

	if(mouse_cursor == MOUSE_CURSOR_IBEAM) cs = cursors[1];
	else if(mouse_cursor == MOUSE_CURSOR_WAIT) cs = cursors[2];

	i_mouse_cursor = mouse_cursor;
	if(wDrawingArea && GDK_IS_WINDOW(wDrawingArea->window))
		gdk_window_set_cursor(wDrawingArea->window, cs);
}
IPLAT get_mouse_cursor(void){
	return i_mouse_cursor;
}


/***********************************************************************************************************
 * CLIPBOARD ROUTINES
 ***********************************************************************************************************/
int cboard_text_put(char *buf, int len){
	GtkClipboard *brd = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
	if(buf && len && brd){
		gtk_clipboard_set_text(brd, buf, len);
		return len;
	}
	return 0;
}

int cboard_text_sze(void){
	char *chr;
	int n = 0;
	GtkClipboard *brd = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
	if(!brd || !gtk_clipboard_wait_is_text_available(brd)) return n;

	chr = gtk_clipboard_wait_for_text(brd);
	if(!chr) return n;
	n = strlen(chr);
	g_free(chr);
	return n;
}

char* cboard_text_ptr(void){
	char *chr;
	GtkClipboard *brd;
	brd = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
	if(!brd || !gtk_clipboard_wait_is_text_available(brd)) return 0;
	chr = gtk_clipboard_wait_for_text(brd);
	return chr;
}
char* cboard_text_free(char *gtkPtr){
	if(gtkPtr) g_free(gtkPtr);
	return 0;
}

/***********************************************************************************************************
 * DRAG AND DROP ROUTINES
 ***********************************************************************************************************/
enum { TARGET_URI_LIST = 100 };
static void on_drag_data_received(GtkWidget *widget, GdkDragContext *context, gint x, gint y, GtkSelectionData *selection_data, guint target_type, guint time, gpointer data){
	gchar **uris;
	gchar *uri;
	int i, n = 0;
	IPLAT dndType = DND_UNDEFINED;

	if(target_type == TARGET_URI_LIST){
		uris = g_uri_list_extract_uris((gchar*)gtk_selection_data_get_data(selection_data));
		for(i = 0; uris[i] != NULL; i++){
			uri = uris[i];
			if(fwplat){
				if(0 == strncmp(uri, "file://", 7)){ dndType = DND_URI_FILE; n = 7; }
				else if(0 == strncmp(uri, "http://", 7)){ dndType = DND_URI_WWW; n = 7; }
				else if(0 == strncmp(uri, "ftp://", 6)){ dndType = DND_URI_FTP; n = 6; }
				ondrop(fwplat, dndType, uri + n, uri);
			}
		}
	 }
}


/***********************************************************************************************************
 * WIDGET CALLBACK ROUTINES
 ***********************************************************************************************************/
int sw_in_event_after = 0;
void on_wmain_event_after(GtkWidget *widget, GdkEvent  *event, gpointer user_data){
	int sw;
	GtkToggleToolButton *togglebtn;

	LIST_INDEX i;
	CTL_TOOLBAR_ITEM *tbi;
	GdkEvent evt;
	gint return_val = 0;
	GdkEventKey *ekey = (GdkEventKey*)event;

	if(sw_in_event_after) return;
	sw_in_event_after = 1;
	if(event){
		if(event->type == GDK_KEY_PRESS){
			//check keybindings
			for(i = 0; lstToolBar && i < lstToolBar->cnt; i++){
				tbi = (CTL_TOOLBAR_ITEM*)list_at(lstToolBar, i);
				if(tbi && LCASE(tbi->ctrl_key) == LCASE(ekey->keyval) && ekey->state & GDK_CONTROL_MASK){
					if(tbi->sw_active){
						//blokeert tot de clicked item in de toolbar volledig uitgevoerd is, man man man wat een ramp die gtk...
						if(tbi->iplat_toolbar == TOOLBAR_PUSHBUTTON) {
							togglebtn = (GtkToggleToolButton*)GTK_OBJECT(tbi->internal);
							if(togglebtn){
								sw = gtk_toggle_tool_button_get_active(togglebtn);
								gtk_toggle_tool_button_set_active(togglebtn, !sw);
								gtk_signal_emit(GTK_OBJECT(tbi->internal), gtk_signal_lookup("toggled", GTK_OBJECT_TYPE(GTK_OBJECT(tbi->internal))), &evt, &return_val);
							}
						} else
							gtk_signal_emit(GTK_OBJECT(tbi->internal), gtk_signal_lookup("clicked", GTK_OBJECT_TYPE(GTK_OBJECT(tbi->internal))), &evt, &return_val);
						break;
					}
				}
			}
		}
	}
	sw_in_event_after = 0;
}
void on_state_changed(GtkWidget *widget, GtkStateType state, gpointer user_data){
	load_system_colors();
	gtk_send_redraw(0);
}

/* Create a new backing pixmap of the appropriate size */
static gboolean on_configure(GtkWidget *widget, GdkEventConfigure *event){
	if(event && wDrawingArea && widget == wDrawingArea){
		if(fwplat) fwplat->redraw_reason = REDRAW_RESIZE;

		pixmap = pixelmap_destroy(pixmap);
		//if(fwplat->sw_double_buffer)	pixmap = pixelmap_create_from_window(widget->window);

		if(vscroll){
			vadj->upper = widget->allocation.height;
			vadj->page_increment = widget->allocation.height / 100;
			vadj->step_increment = .5;
			gtk_adjustment_changed(vadj);
		}
		if(hscroll){
			hadj->upper = widget->allocation.width;
			hadj->page_increment = widget->allocation.width / 100;
			hadj->step_increment = .5f;
			gtk_adjustment_changed(hadj);
		}

		/**
		 * De systeem kleuren herinlezen (kan zijn dat de theme aangepast werd)
		 */

		load_system_colors();

		if(load_systems_defaults()){
			/*system settings changed*/
			cairo_setfont_execute();
		}
	}
	return FALSE;
}

int on_destroy(GtkWidget *widget, gpointer data){
	if(ongui_exit(fwplat)){
		__sw_platform_exit = 1;
		gtk_main_quit();
		wMain = 0;
		return 0;
	}
	return 1;
}

static gboolean on_draw(GtkWidget *widget, GdkEventExpose *event){
	PLAT_RECT	r;

	if(fwplat && !CairoCtx && GTK_IS_WIDGET(wDrawingArea) && gtk_widget_is_drawable(wDrawingArea)){
		/**
		* Routine new_draw_area calls gtk_widget_set_double_buffered.
		* GTK automatically takes care of double buffering.
		* Nothing else must be done...
		*/
		CairoCtx = local_create_dc_from_window(wDrawingArea->window);

		if(CairoCtx){
			cairo_set_antialias(CairoCtx, CAIRO_ANTIALIAS_SUBPIXEL);
			cairo_setfont_execute();

			//dc clipping...
			PLAT_RECT_XYWH(r, event->area.x, event->area.y, event->area.width, event->area.height)
			//gdk_cairo_region(CairoCtx, event->region);

			/*redraw to the pixmap*/
			onredraw(fwplat, &r);

			/*re-read the wToolBar items*/
			fwplat->ctl.toolbar_refresh();
			fwplat->ctl.statusbar_refresh();
		}
		CairoCtx = local_destroy_dc(CairoCtx);

		__cared_bg_update(0, CaretPos.x, CaretPos.y);
		__caret_sw_blink = 1;
		__caret_update();
	}

	fwplat->redraw_reason = IPLAT_NULL;
	return TRUE;
}

IPLAT mB = IPLAT_NULL;
static gboolean on_mousebutton_pressed(GtkWidget * widget, GdkEventButton* event, gpointer pUserData ) {
	if(event){
		if(event->button == 1)
			mB = MBTN_LEFT_IN;
		else
			mB = MBTN_RIGHT_IN;
		mouseX = (int)event->x;
		mouseY = (int)event->y;
		onmouse(fwplat, mouseX, mouseY, mB);
	}
	return FALSE;
}

static gboolean on_mousebutton_released(GtkWidget * widget, GdkEventButton* event, gpointer pUserData ) {
	if(event){
		if(event->button == 1)
			mB = MBTN_LEFT_UP;
		else
			mB = MBTN_RIGHT_UP;
		mouseX = (int)event->x;
		mouseY = (int)event->y;
		onmouse(fwplat, mouseX, mouseY, mB);
		mB = IPLAT_NULL;
	}
	return FALSE;
}

static void on_mouse(GtkWidget *widget, GdkEventMotion *event){
	if(wDrawingArea && widget == wDrawingArea && GTK_IS_WIDGET(widget)){
		gdk_window_get_pointer(event->window, &mouseX, &mouseY, 0);
		onmouse(fwplat, mouseX, mouseY, mB);
	}
}

static gboolean on_mousewheel(GtkWidget * widget, GdkEventScroll * event, gpointer pUserData ) {
	if(wDrawingArea && widget == wMain && GTK_IS_WIDGET(widget)){
		if(event->direction == GDK_SCROLL_UP){
			onmouse(fwplat, mouseX, mouseY, MWHEEL_UP);
		} else if(event->direction == GDK_SCROLL_DOWN){
			onmouse(fwplat, mouseX, mouseY, MWHEEL_DN);

		}
	}
	return FALSE;
}

void run_special_keys(GdkEventKey *ekey){
	if(!__k_ctrl_in && ekey->state & GDK_CONTROL_MASK){
		__k_ctrl_in = 1;
		onskey(fwplat, PCONTROL, 1);
	} else if(__k_ctrl_in && !(ekey->state & GDK_CONTROL_MASK)){
		__k_ctrl_in = 0;
		onskey(fwplat, PCONTROL, 0);
	}

	if(!__k_shift_in && ekey->state & GDK_SHIFT_MASK){
		__k_shift_in = 1;
		onskey(fwplat, PSHIFT, 1);
	} else if(__k_shift_in && !(ekey->state & GDK_SHIFT_MASK)){
		__k_shift_in = 0;
		onskey(fwplat, PSHIFT, 0);
	}

	if(!__k_alt_in && ekey->state & GDK_MOD1_MASK){
		__k_alt_in = 1;
		onskey(fwplat, PALT, 1);
	} else if(__k_alt_in && !(ekey->state & GDK_MOD1_MASK)){
		__k_alt_in = 0;
		onskey(fwplat, PALT, 0);
	}

}

static gboolean on_key(GtkWidget *widget, GdkEventKey *event, gpointer user_data){
	PLATKEY pk = PKAN;
	char *kname;

	if(sw_in_keyboardroutine) return TRUE;
	sw_in_keyboardroutine = 1;
	if(fwplat && event){

		run_special_keys(event);
		kname = gdk_keyval_name(event->keyval);

		if(kname){
			if(0 == strcasecmp(kname, "return"))event->string[(event->length ? event->length - 1 : 0)] = '\n';
			else if(0 == strcasecmp(kname, "down"))pk = PKDN;
			else if(0 == strcasecmp(kname, "up"))pk = PKUP;
			else if(0 == strcasecmp(kname, "left"))pk = PKLE;
			else if(0 == strcasecmp(kname, "right"))pk = PKRI;
			else if(0 == strcasecmp(kname, "page_down"))pk = PKPD;
			else if(0 == strcasecmp(kname, "page_up"))pk = PKPU;
			else if(0 == strcasecmp(kname, "insert"))pk = PKIN;
			else if(0 == strcasecmp(kname, "home"))pk = PKHM;
			else if(0 == strcasecmp(kname, "end"))pk = PKED;
			else if(0 == strcasecmp(kname, "delete"))pk = PKDL;
			else if(0 == strcasecmp(kname, "backspace"))pk = PKBK;
			else if(0 == strcasecmp(kname, "escape"))pk = PESC;
			else if(0 == strcasecmp(kname, "tab"))pk = PTAB;
			else if(0 == strcasecmp(kname, "control_l") || 0 == strcasecmp(kname, "control_r")){
				__k_ctrl_in = 1;
				pk = PCONTROL;
			}
			else if(0 == strcasecmp(kname, "shift_l") || 0 == strcasecmp(kname, "shift_r")){
				__k_shift_in = 1;
				pk = PSHIFT;
			}
			else if(0 == strcasecmp(kname, "alt_l") || 0 == strcasecmp(kname, "alt_r")){
				__k_alt_in = 1;
				pk = PALT;
			}
			else if(0 == strcasecmp(kname, "super_l") || 0 == strcasecmp(kname, "super_r"))pk = PSUPER;
		}

		if(pk == PKAN){
			if(event->keyval < 255)
				onkey(fwplat, event->keyval);
			else
				onkey(fwplat, event->string[(event->length ? event->length - 1 : 0)]);
		} else
			onskey(fwplat, pk, 1);


	}
	sw_in_keyboardroutine = 0;
	return TRUE;
}

static gboolean on_key_released(GtkWidget *widget, GdkEventKey *event, gpointer user_data){
	PLATKEY pk = PKAN;
	char *kname;

	if(fwplat && event){

		run_special_keys(event);

		kname = gdk_keyval_name(event->keyval);
		if(kname){
			//printf("key %s released\n",kname);
			if(0 == strcasecmp(kname, "down"))pk = PKDN;
			else if(0 == strcasecmp(kname, "up"))pk = PKUP;
			else if(0 == strcasecmp(kname, "left"))pk = PKLE;
			else if(0 == strcasecmp(kname, "right"))pk = PKRI;
			else if(0 == strcasecmp(kname, "page_down"))pk = PKPD;
			else if(0 == strcasecmp(kname, "page_up"))pk = PKPU;
			else if(0 == strcasecmp(kname, "insert"))pk = PKIN;
			else if(0 == strcasecmp(kname, "home"))pk = PKHM;
			else if(0 == strcasecmp(kname, "end"))pk = PKED;
			else if(0 == strcasecmp(kname, "delete"))pk = PKDL;
			else if(0 == strcasecmp(kname, "backspace"))pk = PKBK;
			else if(0 == strcasecmp(kname, "escape"))pk = PESC;
			else if(0 == strcasecmp(kname, "tab"))pk = PTAB;
			else if(0 == strcasecmp(kname, "control_l") || 0 == strcasecmp(kname, "control_r")){
				__k_ctrl_in = 0;
				pk = PCONTROL;
			}
			else if(0 == strcasecmp(kname, "shift_l") || 0 == strcasecmp(kname, "shift_r")){
				__k_shift_in = 0;
				pk = PSHIFT;
			}
			else if(0 == strcasecmp(kname, "alt_l") || 0 == strcasecmp(kname, "alt_r")){
				__k_alt_in = 0;
				pk = PALT;
			}
			else if(0 == strcasecmp(kname, "super_l") || 0 == strcasecmp(kname, "super_r"))pk = PSUPER;
		}
		if(pk != PKAN) onskey(fwplat, pk, 0);
	}
	return TRUE;
}

static gboolean on_focus_out(GtkWidget *widget, GdkEventMotion *event){
	if(widget == wMain){
		__caret_sw_blink = 0;
		__caret_update();
		__sw_got_focus = 0;
		__caret_sw_show = 0;
	}
	return FALSE;
}

static gboolean on_focus_in(GtkWidget *widget, GdkEventMotion *event){
	if(widget == wMain){
		//__cared_bg_update(0, CaretPos.x, CaretPos.y);
		__sw_got_focus = 1;
		__caret_sw_show = 1;
		__caret_sw_blink = 1;
		//__caret_update();
		gtk_send_redraw(0);
	}
	return FALSE;
}

/***********************************************************************************************************
 * WIDGET ROUTINES
 ***********************************************************************************************************/
int maxx(void){
	return (wDrawingArea ? wDrawingArea->allocation.width : 0);
}

int maxy(void){
	return (wDrawingArea? wDrawingArea->allocation.height : 0);
}

int resolutionx(void){
	int w=0, h=0;
	GdkWindow *rw = gtk_widget_get_root_window(wMain);
	gdk_drawable_get_size(rw, &w, &h);
	return w;
}

int resolutiony(void){
	int w=0, h=0;
	GdkWindow *rw = gtk_widget_get_root_window(wMain);
	gdk_drawable_get_size(rw, &w, &h);
	return h;
}

void set_title(char *t){
	if(wMain)
		gtk_window_set_title(GTK_WINDOW(wMain), (t ? t : "GtkApp"));
}

void set_icon(char* path){
	GdkPixbuf *icon;
	icon = gdk_pixbuf_new_from_file(path, NULL);
	if(icon) gtk_window_set_icon(GTK_WINDOW(wMain), icon);
}

GtkWidget* new_draw_area(){
	GtkWidget *da = 0;
	GtkTargetList *tl;

	da = gtk_drawing_area_new();
	if(da){
		gtk_widget_set_app_paintable(da, TRUE);
        gtk_drag_dest_set (
                da,
				GTK_DEST_DEFAULT_MOTION |
				GTK_DEST_DEFAULT_HIGHLIGHT |
				GTK_DEST_DEFAULT_DROP,
                0,
                0,
                GDK_ACTION_COPY
        );

		tl = gtk_drag_dest_get_target_list(GTK_WIDGET(da));
		if(!tl){
			tl = gtk_target_list_new (NULL, 0);
			gtk_drag_dest_set_target_list(GTK_WIDGET(da), tl);
			gtk_target_list_unref(tl);
		}
		gtk_target_list_add_uri_targets (tl, TARGET_URI_LIST);

		gtk_widget_set_events(da, GDK_ALL_EVENTS_MASK);
		g_signal_connect(G_OBJECT(da), "motion_notify_event", G_CALLBACK(on_mouse), NULL);
		g_signal_connect(G_OBJECT(da), "button-press-event", G_CALLBACK(on_mousebutton_pressed), NULL);
		g_signal_connect(G_OBJECT(da), "button-release-event", G_CALLBACK(on_mousebutton_released), NULL);
		g_signal_connect(G_OBJECT(da), "expose_event", G_CALLBACK(on_draw), NULL);
		g_signal_connect(G_OBJECT(da), "configure_event", G_CALLBACK(on_configure), NULL);
        g_signal_connect(G_OBJECT(da), "drag-data-received", G_CALLBACK(on_drag_data_received), NULL);
		g_signal_connect(G_OBJECT(da), "state-changed", G_CALLBACK(on_state_changed), NULL);

		gtk_widget_set_double_buffered(GTK_WIDGET(da), fwplat->sw_double_buffer);
		gtk_widget_show(GTK_WIDGET(da));

	}
	return da;
}

/***********************************************************************************************************
 * SYSTEM MESSAGE BOXES
 ***********************************************************************************************************/
char msgbuf[1024 * 16];
void error_msg(const char *title, char *fmt, ...){
	va_list args;
	GtkWidget *werr;

	//doevents();

	msgbuf[0] = 0;
	va_start(args, fmt);
	vsprintf(msgbuf, fmt, args);
	va_end(args);
	if(!strlen(msgbuf)) return;

	werr = gtk_message_dialog_new(GTK_WINDOW(wMain), GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, "%s", msgbuf);
	if(werr){
		gtk_window_set_title(GTK_WINDOW(werr), title ? title : "Error");
		gtk_dialog_run(GTK_DIALOG(werr));
		gtk_widget_destroy(werr);
	}
}
IPLAT yesno_msg(const char *title, char *fmt, ...){
	va_list args;
	GtkWidget *wmsg;
	IPLAT rsp = DLG_CANCEL;

	//doevents();

	msgbuf[0] = 0;
	va_start(args, fmt);
	vsprintf(msgbuf, fmt, args);
	va_end(args);
	if(!strlen(msgbuf)) return rsp;

	wmsg = gtk_message_dialog_new(GTK_WINDOW(wMain), GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, "%s", msgbuf);
	if(wmsg){
		gtk_window_set_title(GTK_WINDOW(wmsg), title ? title : "Please confirm");
		if(gtk_dialog_run(GTK_DIALOG(wmsg)) == GTK_RESPONSE_YES)
			rsp = DLG_OK;
		gtk_widget_destroy(wmsg);
	}
	return rsp;
}

IPLAT msgbox(const char *title, char *fmt, ...){
	va_list args;
	GtkWidget *wmsg;
	IPLAT rsp = DLG_CANCEL;

	//doevents();

	msgbuf[0] = 0;
	va_start(args, fmt);
	vsprintf(msgbuf, fmt, args);
	va_end(args);
	if(!strlen(msgbuf)) return rsp;


	wmsg = gtk_message_dialog_new(GTK_WINDOW(wMain), GTK_DIALOG_MODAL, GTK_MESSAGE_INFO, GTK_BUTTONS_OK, "%s", msgbuf);
	if(wmsg){
		gtk_window_set_title(GTK_WINDOW(wmsg), title ? title : "Info");
		if(gtk_dialog_run(GTK_DIALOG(wmsg)) == GTK_RESPONSE_YES)
			rsp = DLG_OK;
		gtk_widget_destroy(wmsg);
	}
	return rsp;
}

/***********************************************************************************************************
 * TOOLBAR
 ***********************************************************************************************************/
void toolbar_refresh(){
	int i;
	CTL_TOOLBAR_ITEM *itm;
	GtkWidget *w = 0;
	GtkToggleToolButton *togglebtn;

	for(i = 0; !__sw_platform_exit && i < lstToolBar->cnt; i++){
		itm = (CTL_TOOLBAR_ITEM*)list_at(lstToolBar, i);
		if(itm){
			w = (GtkWidget*)(itm->internal);
			if(w){
				if(itm->sw_show && !GTK_WIDGET_VISIBLE(w))
					gtk_widget_show(w);
				else if(!itm->sw_show && GTK_WIDGET_VISIBLE(w))
					gtk_widget_hide(w);

				if(itm->sw_active && w->state == GTK_STATE_INSENSITIVE){
					gtk_widget_set_sensitive(w, 1);
				} else if(!itm->sw_active && w->state != GTK_STATE_INSENSITIVE)
					gtk_widget_set_sensitive(w, 0);

				if(itm->iplat_toolbar == TOOLBAR_PUSHBUTTON){
					togglebtn = GTK_TOGGLE_TOOL_BUTTON(w);
					if(itm->sw_pushed && !gtk_toggle_tool_button_get_active(togglebtn))
						gtk_toggle_tool_button_set_active(togglebtn, 1);
					else if(!itm->sw_pushed && gtk_toggle_tool_button_get_active(togglebtn))
							gtk_toggle_tool_button_set_active(togglebtn, 0);
				}
			}
		}
	}
}

void on_toolbar_item_clicked(GtkToolButton *toolbutton, gpointer user_data){
	GtkWidget *w = 0;
	GtkToggleToolButton *togglebtn;
    char *filename = 0, *folder = 0, *h, *t, filtername[256], filter[256];
    GtkFileFilter *ff = 0;
    int len;

    CTL_TOOLBAR_ITEM *i = (CTL_TOOLBAR_ITEM*)user_data;
	if(i){
		i->response = 0;
		if(i->iplat_toolbar == TOOLBAR_ITEM_FONTSELECTION){
			w = gtk_font_selection_dialog_new("Use Font");
			if(strlen(selectedFont))
				gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG (w), selectedFont);
			else if(strlen(current_fontface))
				gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG (w), current_fontface);

			len = gtk_dialog_run(GTK_DIALOG(w));

			if(len == GTK_RESPONSE_ACCEPT || len == GTK_RESPONSE_OK){
				filename = gtk_font_selection_dialog_get_font_name( GTK_FONT_SELECTION_DIALOG (w) );
				if(filename){
					sprintf(selectedFont, "%s", filename);
					i->response = &selectedFont;
					ontoolbar(fwplat, i);
				}
			}
			gtk_widget_destroy(w);

		} else if(i->iplat_toolbar == TOOLBAR_ITEM_OPENFILE){
			w = gtk_file_chooser_dialog_new(	"Open File",
												GTK_WINDOW(wMain),
												GTK_FILE_CHOOSER_ACTION_OPEN,
												GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
												GTK_STOCK_OPEN,	GTK_RESPONSE_ACCEPT,
												(char*)0);

			if(i->filefilter){
				len = strlen(i->filefilter);
				t = i->filefilter;
				while(t < i->filefilter + len){
					ff = gtk_file_filter_new();
					h = t;
					t = strchr(h, ';') + 1;
					snprintf(filtername, t - h, "%s", h);
					gtk_file_filter_set_name(ff, filtername);
					h = t;
					if(t < i->filefilter + len){
						t = strchr(h, ';') + 1; //get the pattern
						snprintf(filter, t - h, "%s", h);
						gtk_file_filter_add_pattern(ff, filter);
					}
					gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(w), ff);
				}
			}

			if(swValidTootFolder){
				gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(w), fileRootFolder);
			} else {
				gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(w), getenv("HOME"));
			}
			if(gtk_dialog_run(GTK_DIALOG(w)) == GTK_RESPONSE_ACCEPT) filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(w));
			folder = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER(w));
			if(folder){
				sprintf(fileRootFolder, "%s", folder);
				free(folder);
				swValidTootFolder = 1;
			}
			gtk_widget_destroy(w);
			i->response = filename;
		} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCSAVEAS){
			w = gtk_file_chooser_dialog_new(	"Save File As",
												GTK_WINDOW(wMain),
												GTK_FILE_CHOOSER_ACTION_SAVE,
												GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
												GTK_STOCK_SAVE_AS,	GTK_RESPONSE_ACCEPT,
												(char*)0);
			if(swValidTootFolder){
				gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(w), fileRootFolder);
			} else {
				gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(w), getenv("HOME"));
			}
			if(i->filefilter){
				len = strlen(i->filefilter);
				t = i->filefilter;
				while(t < i->filefilter + len){
					ff = gtk_file_filter_new();
					h = t;
					t = strchr(h, ';') + 1;
					snprintf(filtername, t - h, "%s", h);
					gtk_file_filter_set_name(ff, filtername);
					h = t;
					if(t < i->filefilter + len){
						t = strchr(h, ';') + 1; //get the pattern
						snprintf(filter, t - h, "%s", h);
						gtk_file_filter_add_pattern(ff, filter);
					}
					gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(w), ff);
				}
			}

			if(gtk_dialog_run(GTK_DIALOG(w)) == GTK_RESPONSE_ACCEPT) filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(w));
			folder = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER(w));
			if(folder){
				sprintf(fileRootFolder, "%s", folder);
				free(folder);
				swValidTootFolder = 1;
			}
			gtk_widget_destroy(w);
			i->response = filename;
		} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCPRINT){
			//https://developer.gnome.org/gtk2/stable/gtk2-High-level-Printing-API.html
			//https://fossies.org/linux/gtk+/tests/testprint.c
			GtkPrintOperation *print = gtk_print_operation_new ();
			if(print){
				g_signal_connect (print, "begin-print", G_CALLBACK (__plat_gtk_print_begin), NULL);
				g_signal_connect (print, "draw-page", G_CALLBACK (__plat_gtk_print_page), NULL);
				g_signal_connect (print, "end-print", G_CALLBACK (__plat_gtk_print_end), NULL);
				g_signal_connect (print, "preview", G_CALLBACK (__plat_gtk_print_preview), NULL);

				gtk_print_operation_set_use_full_page (print, TRUE);
				gtk_print_operation_set_unit (print, GTK_UNIT_PIXEL);
				len = gtk_print_operation_run(print, GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG, GTK_WINDOW(wMain), NULL);
				if(len == GTK_PRINT_OPERATION_RESULT_APPLY){
					 //gtk_print_operation_run (GTK_PRINT_OPERATION (print), GTK_PRINT_OPERATION_ACTION_PRINT, GTK_WINDOW(wMain), NULL);
				}
				g_object_unref(print);
			}
		}

		if(fwplat){
			ontoolbar(fwplat, i);
			fwplat->redraw_reason = REDRAW_MENU_ACTION;
		}

		if(i->iplat_toolbar == TOOLBAR_PUSHBUTTON){
			togglebtn = GTK_TOGGLE_TOOL_BUTTON(i->internal);
			i->sw_pushed = gtk_toggle_tool_button_get_active(togglebtn);
			if(fwplat){
				ontoolbar(fwplat, i);
				fwplat->redraw_reason = REDRAW_MENU_ACTION;
			}
		}

	}

	if(filename)free(filename);
}

int toolbar_add(CTL_TOOLBAR_ITEM *i){
	GtkToolButton *itm = 0;
	GtkToggleToolButton *togglebtn;
	GtkToolItem *isep = 0;
	GtkWidget *wIcon = 0;
	char *iconpath = 0;

	if(i){
		lstToolBar = list_add(lstToolBar, LIST_ELEMENT_CAST(i));
		if(i->iplat_toolbar == TOOLBAR_ITEM_OPENFILE){
			itm = GTK_TOOL_BUTTON(gtk_tool_button_new_from_stock(GTK_STOCK_OPEN));
			if(!i->ctrl_key) i->ctrl_key = 'o';
			if(i->text && strlen(i->text)) gtk_tool_button_set_label(GTK_TOOL_BUTTON(itm), i->text);
		} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCNEW){
			itm = GTK_TOOL_BUTTON(gtk_tool_button_new_from_stock(GTK_STOCK_NEW));
			if(i->text && strlen(i->text)) gtk_tool_button_set_label(GTK_TOOL_BUTTON(itm), i->text);
			if(!i->ctrl_key) i->ctrl_key = 'n';
		} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCSAVE){
			itm = GTK_TOOL_BUTTON(gtk_tool_button_new_from_stock(GTK_STOCK_SAVE));
			if(i->text && strlen(i->text)) gtk_tool_button_set_label(GTK_TOOL_BUTTON(itm), i->text);
			if(!i->ctrl_key) i->ctrl_key = 's';
		} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCSAVEAS){
			itm = GTK_TOOL_BUTTON(gtk_tool_button_new_from_stock(GTK_STOCK_SAVE_AS));
			if(i->text && strlen(i->text)) gtk_tool_button_set_label(GTK_TOOL_BUTTON(itm), i->text);
			if(!i->ctrl_key) i->ctrl_key = 'a';
		} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCSEARCH){
			itm = GTK_TOOL_BUTTON(gtk_tool_button_new_from_stock(GTK_STOCK_FIND));
			if(i->text && strlen(i->text)) gtk_tool_button_set_label(GTK_TOOL_BUTTON(itm), i->text);
			if(!i->ctrl_key) i->ctrl_key = 'f';
		} else if(i->iplat_toolbar == TOOLBAR_BUTTON ){
			itm = GTK_TOOL_BUTTON(gtk_tool_button_new_from_stock(GTK_STOCK_FILE));
			if(i->icon && strlen(i->icon)){
				wIcon = gtk_image_new_from_file((const gchar*)i->icon);
				if(wIcon){
					gtk_widget_set_visible(wIcon, TRUE);
					gtk_tool_button_set_icon_widget(itm, wIcon);
				}
			}
			gtk_tool_button_set_label(GTK_TOOL_BUTTON(itm), i->text);
		} else if(i->iplat_toolbar == TOOLBAR_PUSHBUTTON){
			if(i->icon && strlen(i->icon)) iconpath = i->icon;
			if(iconpath)
				togglebtn = GTK_TOGGLE_TOOL_BUTTON(gtk_toggle_tool_button_new());
			else
				togglebtn = GTK_TOGGLE_TOOL_BUTTON(gtk_toggle_tool_button_new_from_stock(GTK_STOCK_FILE));
			itm = GTK_TOOL_BUTTON(togglebtn);
			if(i->icon && strlen(i->icon)){
				wIcon = gtk_image_new_from_file((const gchar*)iconpath);
				if(wIcon){
					//gtk_toolbar_set_icon_size(wToolBar, GTK_ICON_SIZE_MENU);
					gtk_widget_set_visible(wIcon, TRUE);
					gtk_tool_button_set_icon_widget(itm, wIcon);
				}
			}
			gtk_tool_button_set_label(itm, i->text);
		} else if(i->iplat_toolbar == TOOLBAR_ITEM_SEPARATOR){
			isep = gtk_separator_tool_item_new();
			if(i->sw_show) gtk_widget_show(GTK_WIDGET(isep));
			gtk_toolbar_insert (wToolBar, isep,  gtk_toolbar_get_n_items(wToolBar));
		} else if(i->iplat_toolbar == TOOLBAR_ITEM_FONTSELECTION){
			isep = gtk_tool_button_new_from_stock(GTK_STOCK_SELECT_FONT);
			i->internal = (void*)isep;
			if(i->sw_show) gtk_widget_show(GTK_WIDGET(isep));
			if(!i->sw_active)gtk_widget_set_state(GTK_WIDGET(isep), GTK_STATE_INSENSITIVE);
			gtk_toolbar_insert (wToolBar, isep,  gtk_toolbar_get_n_items(wToolBar));
			gtk_tool_item_set_homogeneous(GTK_TOOL_ITEM(isep), 0);
			g_signal_connect(G_OBJECT(isep), "clicked", G_CALLBACK(on_toolbar_item_clicked), i);

		} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCPRINT){
			isep = gtk_tool_button_new_from_stock(GTK_STOCK_PRINT);
			i->internal = (void*)isep;
			if(i->sw_show) gtk_widget_show(GTK_WIDGET(isep));
			if(!i->sw_active)gtk_widget_set_state(GTK_WIDGET(isep), GTK_STATE_INSENSITIVE);
			gtk_toolbar_insert (wToolBar, isep,  gtk_toolbar_get_n_items(wToolBar));
			gtk_tool_item_set_homogeneous(GTK_TOOL_ITEM(isep), 0);
			g_signal_connect(G_OBJECT(isep), "clicked", G_CALLBACK(on_toolbar_item_clicked), i);

		}

		if(itm){
			gtk_tool_item_set_homogeneous(GTK_TOOL_ITEM(itm), 0);
			if(i->iplat_toolbar == TOOLBAR_PUSHBUTTON){
				g_signal_connect(G_OBJECT(togglebtn), "toggled", G_CALLBACK(on_toolbar_item_clicked), i);
			} else
				g_signal_connect(G_OBJECT(itm), "clicked", G_CALLBACK(on_toolbar_item_clicked), i);

			i->internal = (void*)itm;
			if(i->sw_show) gtk_widget_show(GTK_WIDGET(itm));
			if(!i->sw_active)gtk_widget_set_state(GTK_WIDGET(itm), GTK_STATE_INSENSITIVE);
			if(i->tooltip) gtk_tool_item_set_tooltip_text(GTK_TOOL_ITEM(itm), i->tooltip);
			gtk_toolbar_insert (wToolBar, GTK_TOOL_ITEM(itm),  gtk_toolbar_get_n_items(wToolBar));
		}
	}

	if(lstToolBar->cnt && !GTK_WIDGET_VISIBLE(GTK_WIDGET(wToolBar)))
		gtk_widget_show(GTK_WIDGET(wToolBar));

	//return 1 for success and 0 for failure
	return itm || isep ? 1 : 0;
}

/***********************************************************************************************************
 *  NOTEBOOK (tab windows)
 ***********************************************************************************************************/

/**
 * The aim is to re-parent vBoxDrawContainer as a whole.
 * The parent is either vBoxMainCenter or a notebook page.
 **/
void reparent_draw_container(GtkBox *aBox){
	GtkWidget *wParent, *wDrawAreaRef;

	if(aBox){
		wDrawAreaRef = g_object_ref(vBoxDrawContainer);
		wParent = gtk_widget_get_parent(vBoxDrawContainer);
		if(wParent != GTK_WIDGET(aBox)){
			if(wParent) gtk_container_remove(GTK_CONTAINER(wParent), vBoxDrawContainer);
			gtk_box_pack_start(aBox, vBoxDrawContainer, TRUE, TRUE, 0);
		}
		g_object_unref(wDrawAreaRef);
	}
}
void on_close_tab(GtkButton *button, gpointer drawarr){

	if(fwplat && ontabclose(fwplat, (void*)drawarr)){
		reparent_draw_container(GTK_BOX(vBoxMainCenter));
		gtk_notebook_remove_page(wNoteBook, gtk_notebook_get_current_page(wNoteBook));
	}

	if(gtk_notebook_get_n_pages(wNoteBook) == 0){
		gtk_widget_hide(GTK_WIDGET(wNoteBook));
	}
}

void tabwnd_selectpage(GtkNotebook *notebook, GtkNotebookPage *page, guint n, gpointer user_data){
	GtkBox *pg;

	pg = GTK_BOX(gtk_notebook_get_nth_page(wNoteBook, n));
	reparent_draw_container(pg);
	if(fwplat) ontabactive(fwplat, (void*)pg);
}

void* tabwnd_create(char *title){
	GtkHBox  *hb = 0, *hbDraw = 0;
	GtkButton *btnClose;
	GtkWidget *image, *lbl;
	int n;

	if(gtk_notebook_get_n_pages(wNoteBook) == 0){
		gtk_widget_show(GTK_WIDGET(wNoteBook));
	}

	hb = GTK_HBOX(gtk_hbox_new(FALSE, 0));
	lbl = gtk_label_new((const gchar*)title);
	btnClose = GTK_BUTTON(gtk_button_new());
	hbDraw = GTK_HBOX(gtk_hbox_new(FALSE, 0));
	reparent_draw_container(GTK_BOX(hbDraw));

	image = gtk_image_new_from_stock(GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
	gtk_container_add(GTK_CONTAINER(btnClose), image);
	gtk_button_set_relief(GTK_BUTTON(btnClose), GTK_RELIEF_NONE);

	g_signal_connect(G_OBJECT(btnClose), "clicked", G_CALLBACK(on_close_tab), hbDraw);
	gtk_button_set_focus_on_click(GTK_BUTTON(btnClose), FALSE);
	gtk_container_add (GTK_CONTAINER(hb), lbl);
	gtk_container_add (GTK_CONTAINER(hb), GTK_WIDGET(btnClose));
	gtk_widget_show(lbl);
	gtk_widget_show(GTK_WIDGET(btnClose));
	gtk_widget_show(GTK_WIDGET(image));
	gtk_widget_show(GTK_WIDGET(hbDraw));

	n = gtk_notebook_append_page(wNoteBook, GTK_WIDGET(hbDraw), GTK_WIDGET(hb));
	gtk_notebook_set_current_page(wNoteBook, n);

	gtk_notebook_set_tab_reorderable(wNoteBook, GTK_WIDGET(hbDraw), TRUE);

	return (void*)hbDraw;
}

void tabwnd_set_lock(int sw){
	gtk_widget_set_sensitive(GTK_WIDGET(wNoteBook), !sw);
}

void* tabwnd_set_title(void* screenID, char *title){
	GtkWidget *pg;
	GtkHBox *hbw;
	GList *lst;
	guint i;
	gpointer data;
	GtkLabel *lbl;
	if(!wNoteBook || !title || !strlen(title)) return screenID;
	pg = gtk_notebook_get_nth_page(wNoteBook, gtk_notebook_get_current_page(wNoteBook));
	lbl = 0;
	if(pg){
		hbw = GTK_HBOX(gtk_notebook_get_tab_label(wNoteBook, GTK_WIDGET(screenID)));
		if(!hbw) return screenID;
		lst = gtk_container_get_children(GTK_CONTAINER(hbw));
		if(!lst) return screenID;
		for(i = 0; i <  g_list_length(lst); i++){
			data = g_list_nth_data(lst, i);
			if(data && GTK_IS_LABEL(data)){
				lbl = GTK_LABEL(data);
				break;
			}
		}
		if(lbl){
			gtk_label_set_text(lbl, title);
		}
		g_list_free(lst);
	}
	return screenID;
}

/***********************************************************************************************************
 * SCROLLBARS
 ***********************************************************************************************************/
void sbar_set(IPLAT sbar, double pct){
	double val;
	if(sbar == SBARV && vadj){
		if(!wDrawingArea) pct = 0;
		val = maxy() / 100.f * pct;
		vadj->value = val;
		gtk_adjustment_changed(vadj);
	} else if(sbar == SBARH && hadj){
		if(!wDrawingArea) pct = 0;
		val = maxx() / 100.f * pct;
		hadj->value = val;
		gtk_adjustment_changed(hadj);
	}
}
void on_srollbar_changed(GtkAdjustment *adjustment, gpointer user_data){
	double val;
	/*not during a drawing operation*/
	if(!CairoCtx && fwplat){
		if(adjustment == hadj){
			val = (double)(adjustment->value) / (double)maxx() * 100.f;
			if(!wDrawingArea) val = 0;
			onscrollbar(fwplat, SBARH, val);
			gtk_adjustment_changed(hadj);
		}
		else if(adjustment == vadj){
			val = (double)(adjustment->value) / (double)maxy() * 100.f;
			if(!wDrawingArea) val = 0;
			onscrollbar(fwplat, SBARV, val);
			gtk_adjustment_changed(vadj);
		}
	}
}

void sbar_show(IPLAT sbar, int sw){
	gboolean v;

	if(sbar == SBARH){
		v = GTK_WIDGET_VISIBLE(GTK_WIDGET(hscroll));
		if(sw) {
			if(!v) gtk_widget_show(GTK_WIDGET(hscroll));
		} else {
			if(v) gtk_widget_hide(GTK_WIDGET(hscroll));
		}
	} else if(sbar == SBARV){
		v = GTK_WIDGET_VISIBLE(GTK_WIDGET(vscroll));
		if(sw){
			if(!v) gtk_widget_show(GTK_WIDGET(vscroll));
		} else {
			if(v) gtk_widget_hide(GTK_WIDGET(vscroll));
		}
	}

}

/***********************************************************************************************************
 * STATUSBAR
 ***********************************************************************************************************/
void statusbar_refresh(){
	CTL_STATUSBAR_ITEM	*itm;
	GtkStatusbar *statbar;
	GtkWidget *w;
	int i;

	if(!GTK_IS_WIDGET(hBoxStatus) || !lstStatusBar) return;
	for(i = 0; !__sw_platform_exit && i < lstStatusBar->cnt; i++){
		itm = (CTL_STATUSBAR_ITEM*)list_at(lstStatusBar, i);
		if(itm){
			statbar = (GtkStatusbar*)GTK_STATUSBAR(itm->internal);
			w = GTK_WIDGET(statbar);
			gtk_statusbar_pop(statbar, 1);
			gtk_statusbar_push(statbar, 1, itm->text);

			if(i < lstStatusBar->cnt - 1)
				gtk_statusbar_set_has_resize_grip(statbar, FALSE);
			else
				gtk_statusbar_set_has_resize_grip(statbar, TRUE);

			if(itm->sw_active && w->state == GTK_STATE_INSENSITIVE)
				gtk_widget_set_sensitive(w, 1);
			else if(!itm->sw_active && w->state != GTK_STATE_INSENSITIVE)
				gtk_widget_set_sensitive(w, 0);

		}
	}
}

void statusbar_add(CTL_STATUSBAR_ITEM *i){
	GtkStatusbar	*statbar = 0;
	if(i){
		statbar = GTK_STATUSBAR(gtk_statusbar_new());
		gtk_widget_show(GTK_WIDGET(statbar));
		gtk_box_pack_start(GTK_BOX(hBoxStatus), GTK_WIDGET(statbar), TRUE, TRUE, 0);

		i->internal = (void*)statbar;
		gtk_statusbar_push(statbar, 1, i->text);
		lstStatusBar = list_add(lstStatusBar, LIST_ELEMENT_CAST(i));
		if(lstStatusBar->cnt && !GTK_WIDGET_VISIBLE(GTK_WIDGET(hBoxStatus)))
			gtk_widget_show(GTK_WIDGET(hBoxStatus));
	}
}

/***********************************************************************************************************
 * CUSTOM DIALOG BOXES
 ***********************************************************************************************************/
#define DIALOGBOX_SPACING 3

struct DIALOG {
	GtkWidget		*wDialog, *vBox;
	LIST 			*lstCtlItems;
	dialog_routine	on_altered;
	IPLAT 			cb_resp;
	DIALOGCONTROL	itmFocus;
	void			*externaldata;
}DIALOG;
struct DIALOG_ITEM {
	IPLAT 			controltype;
	GtkWidget 		*wItem, *sw, *wExtra;
	unsigned int	id;

	int nColumns;

	char *SelectedID;
}DIALOG_ITEM;

struct DIALOG *CurrentDialog = 0;

void* dlg_get_userdata(DIALOGBOX dialog){
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	return (dlg ? dlg->externaldata : 0);
}
void* dlg_set_userdata(DIALOGBOX dialog, void *userdata){
	void *currdata = 0;
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	if(dlg){
		currdata = dlg->externaldata;
		dlg->externaldata = userdata;
	}
	return currdata;
}

DIALOGCONTROL dlg_next_item(DIALOGBOX dialog, DIALOGCONTROL lastitem){
	int sw = 0;
	LIST_INDEX i;
	struct DIALOG_ITEM *itm;
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	if(dlg){
		for(i = 0; i < dlg->lstCtlItems->cnt; i++){
			itm = (struct DIALOG_ITEM*)list_at(dlg->lstCtlItems, i);
			if(!itm) continue;
			if(!lastitem || sw) return itm;
			if(itm == lastitem) sw = 1;
		}
	}
	return 0;
}

struct DIALOG_ITEM* dlg_runtime_find_control(struct DIALOG *dlg, GtkWidget *w){
	struct DIALOG_ITEM *itm = 0;

	do{
		itm = (struct DIALOG_ITEM*)dlg_next_item(dlg, itm);
		if(itm && itm->wItem == w) return itm;
	}while(itm);

	return 0;
}

struct DIALOG_ITEM* dlg_runtime_find(DIALOGBOX dialog, DIALOGCONTROL ctl){
	LIST_INDEX i;
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	if(ctl && dlg && dlg->lstCtlItems){
		for(i = 0; i < dlg->lstCtlItems->cnt; i++){
			if((struct DIALOG_ITEM*)ctl == (struct DIALOG_ITEM*)list_at(dlg->lstCtlItems, i))
				return (struct DIALOG_ITEM*)list_at(dlg->lstCtlItems, i);
		}
	}
	return 0;
}

DIALOGCONTROL dlg_find_item(DIALOGBOX dialog, unsigned int id){
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	struct DIALOG_ITEM *itm = 0;
	LIST_INDEX i;
	if(id && dlg && dlg->lstCtlItems){
		for(i = 0; i < dlg->lstCtlItems->cnt; i++){
			itm = (struct DIALOG_ITEM*)list_at(dlg->lstCtlItems, i);
			if(itm && itm->id == id)
				return itm;
		}
	}
	return itm;
}

void dlg_runtime_btn_clicked(GtkButton *button, gpointer user_data){
	struct DIALOG_ITEM *itm;
	IPLAT ires;
	struct DIALOG *dlg = (struct DIALOG*)user_data;
	if(dlg && dlg == CurrentDialog && dlg->wDialog && dlg->on_altered){
		dlg->cb_resp = IPLAT_NULL;
		itm = dlg_runtime_find_control(dlg, GTK_WIDGET(button));
		if(itm){
			ires = dlg->on_altered(dlg, itm, CTL_STAT_CLICKED);
			if(ires == CTL_CB_CLOSEDLG){
				gtk_dialog_response(GTK_DIALOG(dlg->wDialog), GTK_RESPONSE_ACCEPT);
				gtk_widget_hide(dlg->wDialog);
			}
		}
	}

}

gboolean dlg_runtime_txtbox_onkey(GtkWidget *widget, GdkEventKey *event, gpointer user_data){
	struct DIALOG_ITEM *itm;
	IPLAT ires;
	struct DIALOG *dlg = (struct DIALOG*)user_data;
	if(dlg && dlg == CurrentDialog && dlg->wDialog && dlg->on_altered){
		dlg->cb_resp = IPLAT_NULL;
		itm = dlg_runtime_find_control(dlg, widget);

		ires = dlg->on_altered(dlg, itm, CTL_STAT_TEXT_ALTERED);
		if(ires == CTL_CB_CLOSEDLG)
			gtk_dialog_response(GTK_DIALOG(dlg->wDialog), GTK_RESPONSE_ACCEPT);
	}
	return FALSE;
}

void dlg_runtime_combo_selchanged(GtkComboBox *widget, gpointer user_data){
	struct DIALOG_ITEM *itm;
	IPLAT ires;
	struct DIALOG *dlg = (struct DIALOG*)user_data;
	if(dlg && dlg == CurrentDialog && dlg->wDialog && dlg->on_altered){
		dlg->cb_resp = IPLAT_NULL;
		itm = dlg_runtime_find_control(dlg, GTK_WIDGET(widget));
		ires = dlg->on_altered(dlg, itm, CTL_STAT_CLICKED);
		if(ires == CTL_CB_CLOSEDLG)
			gtk_dialog_response(GTK_DIALOG(dlg->wDialog), GTK_RESPONSE_ACCEPT);
	}
}

void dlg_runtime_listbox_selchanged(GtkTreeView *tree_view, gpointer user_data){
	IPLAT ires;
	GtkTreeModel *model;
	GtkTreePath *p = 0;
	GtkTreeIter		iter;
	struct DIALOG_ITEM *itm = (struct DIALOG_ITEM*)user_data;

	if(tree_view && itm && CurrentDialog && CurrentDialog->on_altered){
		model = gtk_tree_view_get_model(tree_view);
		gtk_tree_view_get_cursor(tree_view, &p, 0);
		gtk_tree_model_get_iter(model, &iter, p);
		gtk_tree_model_get(model, &iter, 0, &itm->SelectedID, -1);
		gtk_tree_path_free(p);
		ires = CurrentDialog->on_altered(CurrentDialog, itm, CTL_STAT_CLICKED);
		if(ires == CTL_CB_CLOSEDLG)
			gtk_dialog_response(GTK_DIALOG(CurrentDialog->wDialog), GTK_RESPONSE_ACCEPT);
	}
}

IPLAT dlg_item_type(DIALOGCONTROL ctl){
	struct DIALOG_ITEM *itm = (struct DIALOG_ITEM*)(ctl);
	return (itm ? itm->controltype : IPLAT_NULL);
}

void dlg_item_show(DIALOGCONTROL ctl, int sw){
	struct DIALOG_ITEM *itm = (struct DIALOG_ITEM*)(ctl);
	if(itm && itm->wItem){
		if(!sw)
			gtk_widget_hide(itm->wItem);
		else
			gtk_widget_show(itm->wItem);
	}
}
int dlg_item_shown(DIALOGCONTROL ctl){
	struct DIALOG_ITEM *itm = (struct DIALOG_ITEM*)(ctl);
	return (itm && itm->wItem ? gtk_widget_get_visible(itm->wItem) : 0);
}
void dlg_item_enable(DIALOGCONTROL ctl, int sw){
	struct DIALOG_ITEM *itm = (struct DIALOG_ITEM*)(ctl);
	if(itm && itm->wItem) gtk_widget_set_sensitive(itm->wItem, sw);
}

int dlg_item_enabled(DIALOGCONTROL ctl){
	struct DIALOG_ITEM *itm = (struct DIALOG_ITEM*)(ctl);
	return (itm && itm->wItem ? gtk_widget_is_sensitive(itm->wItem) : 0);
}
void dlg_item_set_value(DIALOGCONTROL ctl, char *value){
	GtkTreeStore 	*store;
	GtkTreeIter		iter;
	GType			*gtypes = 0;
	char 			tmp[TXTOUT_MAX], *head, *tail, *str;
	int				s, n, c;
	struct DIALOG_ITEM *itm = (struct DIALOG_ITEM*)(ctl);

	GtkTextBuffer *buffer;

	if(itm && itm->wItem && value){
		s = strlen(value);

		switch(itm->controltype){
		case CTL_CHECK_BUTTON:
		case CTL_PUSH_BUTTON:
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(itm->wItem), !strcmp(value, "1"));
			break;

		case CTL_TEXT:
			gtk_entry_set_text(GTK_ENTRY(itm->wItem), value);
		    break;

		case CTL_TEXT_MULTI_LINE:
			buffer = GTK_TEXT_VIEW(itm->wItem)->buffer;
			gtk_text_buffer_set_text(buffer, value, strlen(value));
		    break;

		case CTL_DROPBOX:
			gtk_combo_box_remove_text((GtkComboBox*)itm->wItem, 0);
			s = strlen(value);
			if(s){
				head = value;
				tail = head + s;
				while(head < tail){
					str = strchr(head, '\n');
					n = str - head;
					if(n){
						memset(&tmp, 0, TXTOUT_MAX);
						memcpy(tmp, head, n);
						gtk_combo_box_append_text((GtkComboBox*)itm->wItem, (const gchar*)tmp);
					}
					head = str + 1;
				}
			}

		    break;

		case CTL_LIST_BOX:
			n = sizeof(G_TYPE_STRING) * (itm->nColumns + 2);
			gtypes = malloc(n);
			memset(gtypes, 0, n);
			c = itm->nColumns + 1;
			for(n = 0; n < c; n++) gtypes[n] = G_TYPE_STRING;
			store = gtk_tree_store_newv(c, gtypes);
			free(gtypes);

			/*value format is:
			 * id;value column 1\nvalue column 2\n...\n
			 * id;value column 1\nvalue column 2\n...\n
			 * ... */
			head = value;
			while(head < head + s){
				tail = strchr(head, ';');
				if(!tail) break;
				n = tail - head;
				if(n > TXTOUT_MAX) n = TXTOUT_MAX;
				if(!n) break;
				memset(&tmp, 0, TXTOUT_MAX);
				memcpy(tmp, head, n);
				gtk_tree_store_append(store, (GtkTreeIter*)&iter, NULL);
				gtk_tree_store_set(store, (GtkTreeIter*)&iter, 0, tmp,-1);

				head = tail + 1;
				for(c = 0; head < head + s && c < itm->nColumns; c++){
					tail = strchr(head, '\n');
					if(!tail) break;
					n = tail - head;
					if(n > TXTOUT_MAX) n = TXTOUT_MAX;
					memset(&tmp, 0, TXTOUT_MAX);
					memcpy(tmp, head, n);

					gtk_tree_store_set(store, (GtkTreeIter*)&iter, c + 1, tmp, -1);
					head = tail + 1;
				}
			}

			gtk_tree_view_set_model(GTK_TREE_VIEW(itm->wItem), GTK_TREE_MODEL(store));
			g_object_unref(GTK_TREE_MODEL(store));
		    break;

		default:
			break;
		}
	}
}
char* dlg_item_get_value(DIALOGCONTROL ctl){
	struct DIALOG_ITEM *itm = (struct DIALOG_ITEM*)(ctl);
	if(itm && itm->wItem){
		switch(itm->controltype){
		case CTL_CHECK_BUTTON:
		case CTL_PUSH_BUTTON:
			return (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(itm->wItem)) ? "1" : "0");
			break;

		case CTL_TEXT:
		case CTL_TEXT_MULTI_LINE:
			return (char*)gtk_entry_get_text(GTK_ENTRY(itm->wItem));
		    break;

		case CTL_DROPBOX:
			return gtk_combo_box_get_active_text(GTK_COMBO_BOX(itm->wItem));
		    break;

		case CTL_LIST_BOX:
			return itm->SelectedID;
		    break;

		default:
			break;
		}
	}
	return 0;
}
int dlg_item_get_value_size(DIALOGCONTROL ctl){
	struct DIALOG_ITEM *itm = (struct DIALOG_ITEM*)(ctl);
	if(itm && itm->wItem){
		switch(itm->controltype){
		case CTL_CHECK_BUTTON:
			return 1;
			break;

		case CTL_PUSH_BUTTON:
			return 1;
			break;

		case CTL_TEXT:
		case CTL_TEXT_MULTI_LINE:
			return (gtk_entry_get_text(GTK_ENTRY(itm->wItem)) ? strlen((char*)gtk_entry_get_text(GTK_ENTRY(itm->wItem))) : 0);
		    break;

		case CTL_DROPBOX:
			return (gtk_combo_box_get_active_text(GTK_COMBO_BOX(itm->wItem)) ? strlen(gtk_combo_box_get_active_text(GTK_COMBO_BOX(itm->wItem))) : 0);
		    break;

		case CTL_LIST_BOX:
			return (itm->SelectedID ? strlen(itm->SelectedID) : 0);
		    break;

		default:
			break;
		}
	}
	return 0;
}

unsigned int dlg_item_get_id(DIALOGCONTROL ctl){
	struct DIALOG_ITEM *itm = (struct DIALOG_ITEM*)(ctl);
	return (itm ? itm->id : 0);
}

void dlg_runtime_draw_dialog(DIALOGBOX dialog, char *title){
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	GtkWidget *wAct;
	LIST_INDEX i;
	struct DIALOG_ITEM *itm;
	GtkWidget *wPar, *hBox;

	if(dlg->vBox) return;

	if(dlg && dlg == CurrentDialog && dlg->wDialog){
		gtk_window_set_title(GTK_WINDOW(dlg->wDialog), title);

		for(i = 0; i < dlg->lstCtlItems->cnt; i++){
			itm = (struct DIALOG_ITEM *)list_at(dlg->lstCtlItems, i);
			if(itm->controltype != CTL_NEWLINE && itm->wItem && !itm->sw){
				wPar = gtk_widget_get_parent(itm->wItem);
				if(wPar) gtk_container_remove(GTK_CONTAINER(wPar), itm->wItem);
			}
		}

		if(dlg->vBox){
			gtk_widget_destroy(dlg->vBox);
			dlg->vBox = 0;
		}

	    /*A gtkFixed is not an option here :(, can't figure out the size of the controls*/
	    wAct =  gtk_dialog_get_content_area(GTK_DIALOG(dlg->wDialog));
	    dlg->vBox = gtk_vbox_new(FALSE, DIALOGBOX_SPACING);
	    gtk_container_add(GTK_CONTAINER(wAct), dlg->vBox);

	    hBox = gtk_hbox_new(FALSE, DIALOGBOX_SPACING);
	    gtk_container_add(GTK_CONTAINER(dlg->vBox), hBox);

		for(i = 0; i < dlg->lstCtlItems->cnt; i++){
			itm = (struct DIALOG_ITEM*)list_at(dlg->lstCtlItems, i);
			if(!itm) continue;
			if(itm->controltype != CTL_NEWLINE && !dlg_item_shown(itm)) continue;
			if(itm->controltype == CTL_NEWLINE || itm->controltype == CTL_NEWLINE_DRAW){
			    hBox = gtk_hbox_new(FALSE, DIALOGBOX_SPACING);
			    gtk_container_add(GTK_CONTAINER(dlg->vBox), hBox);
			    if(itm->controltype == CTL_NEWLINE_DRAW){
			    	gtk_box_pack_start(GTK_BOX(hBox), itm->wItem, TRUE, TRUE, 0);
				    gtk_box_pack_start(GTK_BOX(hBox), itm->wExtra, FALSE, FALSE, 0);
				    hBox = gtk_hbox_new(FALSE, DIALOGBOX_SPACING);
				    gtk_container_add(GTK_CONTAINER(dlg->vBox), hBox);
			    }
			} else if(itm->wItem){
				if(itm->sw){
					gtk_box_pack_start(GTK_BOX(hBox), itm->sw, TRUE, TRUE, 0);
				}
				else
					gtk_box_pack_start(GTK_BOX(hBox), itm->wItem, TRUE, TRUE, 0);
			}
		}

	    gtk_widget_show_all(wAct);
	    gtk_widget_queue_resize(dlg->wDialog);

		if(dlg->itmFocus){
			itm = (struct DIALOG_ITEM*)dlg->itmFocus;
			if(itm->wItem && GTK_IS_WIDGET(itm->wItem) && gtk_widget_get_can_focus(itm->wItem))
				gtk_window_set_focus(GTK_WINDOW(dlg->wDialog), itm->wItem);
		}
	}
}

char* __impl_dgl_initString_number(char *initstring, int *out){
	char 	tmp[16],
			*head = tmp,
			*tail = initstring;

	while(tail && *tail){
		if(*tail == ';'){
			tail++;
			break;
		}
		else if(*tail >= '0' && *tail <= '9'){
			*head = *tail;
			head++;
		}
		tail++;
	}
	*head = 0;
	*out = atoi(tmp);

	return tail;
}

DIALOGBOX dlg_add_item(DIALOGBOX dialog, IPLAT ctlType, unsigned int id, char *initstring){
	int nx;
	struct DIALOG_ITEM *itm;
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	char tmp[TXTOUT_MAX], *head, *tail;
	int len, wdt, hgt;

	GtkTreeViewColumn	*col;
	GtkCellRenderer		*renderer;

	if(!dlg){
		dlg = (struct DIALOG*) malloc(sizeof(struct DIALOG));
		if(!dlg) return 0;
		memset(dlg, 0, sizeof(struct DIALOG));
		dlg->wDialog =  gtk_dialog_new_with_buttons(
							"Dialog Box",
							GTK_WINDOW(wMain),
							GTK_DIALOG_DESTROY_WITH_PARENT,
							NULL);
	    if(!dlg->wDialog){
	    	free(dlg);
	    	return 0;
	    }

	    //prevent destruction when close button is clicked
	    g_signal_connect (dlg->wDialog, "delete-event", G_CALLBACK (gtk_widget_hide_on_delete), NULL);

	    dlg->lstCtlItems = list_new(LIST_MIN_BLOCK);
	}

	itm = (struct DIALOG_ITEM*)malloc(sizeof(struct DIALOG_ITEM));
	if(itm){
		memset(itm, 0, sizeof(struct DIALOG_ITEM));
		switch(ctlType){
		case CTL_LABEL:				/*initstring label text*/
			itm->wItem = gtk_label_new(initstring);
		    break;

		case CTL_BUTTON:			/*initstring button text*/
			itm->wItem = gtk_button_new_with_label(initstring);
			g_signal_connect(G_OBJECT(itm->wItem), "clicked", G_CALLBACK(dlg_runtime_btn_clicked), dlg);
			break;
		case CTL_CHECK_BUTTON:		/*initstring check box text*/
			itm->wItem = gtk_check_button_new_with_label(initstring);
			g_signal_connect(G_OBJECT(itm->wItem), "clicked", G_CALLBACK(dlg_runtime_btn_clicked), dlg);
			break;
		case CTL_PUSH_BUTTON:		/*initstring button text*/
			itm->wItem = gtk_toggle_button_new_with_label(initstring);
			g_signal_connect(G_OBJECT(itm->wItem), "clicked", G_CALLBACK(dlg_runtime_btn_clicked), dlg);
			break;

		case CTL_TEXT:				/*initstring x; where x = width in characters*/
			itm->wItem = gtk_entry_new();

			initstring = __impl_dgl_initString_number(initstring, &wdt);
			if(wdt) g_object_set(GTK_ENTRY(itm->wItem), "width-chars", wdt, NULL);

			g_signal_connect(G_OBJECT(itm->wItem), "key-release-event", G_CALLBACK(dlg_runtime_txtbox_onkey), dlg);
			break;

		case CTL_TEXT_MULTI_LINE:	/*initstring x;y; where x = width in characters and y = amount of lines*/
			itm->wItem = gtk_text_view_new();
			itm->sw = gtk_scrolled_window_new (NULL, NULL);

			initstring = __impl_dgl_initString_number(initstring, &wdt);
			if(wdt) g_object_set(itm->wItem, "width-request", wdt * fwplat->gui.fontwdt(), NULL);

			initstring = __impl_dgl_initString_number(initstring, &hgt);
			if(hgt) g_object_set(itm->wItem, "height-request", hgt * fwplat->gui.fonthgt(), NULL);

			gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(itm->sw), GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);
			gtk_container_add(GTK_CONTAINER(itm->sw), GTK_WIDGET(itm->wItem));
			g_signal_connect(G_OBJECT(itm->wItem), "key-release-event", G_CALLBACK(dlg_runtime_txtbox_onkey), dlg);
			break;

		case CTL_DROPBOX:			/*initstring x; where x = width in characters*/
			itm->wItem = gtk_combo_box_new_text();

			initstring = __impl_dgl_initString_number(initstring, &wdt);
			if(wdt) g_object_set(itm->wItem, "width-request", wdt * fwplat->gui.fontwdt(), NULL);

			g_signal_connect(G_OBJECT(itm->wItem), "changed", G_CALLBACK(dlg_runtime_combo_selchanged), dlg);
			break;

		case CTL_LIST_BOX:			/*initstring title "x;y;column 1;title column 2;...;" where x = width in characters and y = height in lines of the box*/
			itm->wItem = gtk_tree_view_new();

			initstring = __impl_dgl_initString_number(initstring, &wdt);
			if(wdt) g_object_set(itm->wItem, "width-request", wdt * fwplat->gui.fontwdt(), NULL);

			initstring = __impl_dgl_initString_number(initstring, &hgt);
			if(hgt) g_object_set(itm->wItem, "height-request", hgt * fwplat->gui.fonthgt(), NULL);

			itm->sw = gtk_scrolled_window_new(NULL, NULL);
			gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(itm->sw), GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);
			gtk_container_add(GTK_CONTAINER(itm->sw), GTK_WIDGET(itm->wItem));

			/*create the 1st invisible column for the ID*/
			renderer = gtk_cell_renderer_text_new();
			col = gtk_tree_view_column_new();
			gtk_tree_view_column_pack_start(col, renderer, TRUE);
			gtk_tree_view_append_column(GTK_TREE_VIEW(itm->wItem), col);
			gtk_tree_view_column_set_title(col, "ID");
			gtk_tree_view_column_set_visible(col, FALSE);
			gtk_tree_view_column_set_alignment(col, 0);
			gtk_cell_renderer_set_visible(renderer, FALSE);

			/*create the columns and column titles from the initstring*/
			head = (char*)initstring;
			len = strlen(initstring);
			while(head < (char*)initstring + len){
				tail = strchr(head, ';');
				nx = tail - head;
				if(nx > TXTOUT_MAX)	nx = TXTOUT_MAX;
				if(!nx) break;
				renderer = gtk_cell_renderer_text_new();
				col = gtk_tree_view_column_new();
				itm->nColumns++;
				memset(&tmp, 0, TXTOUT_MAX);
				memcpy(tmp, head, nx);
				gtk_tree_view_column_set_title(col, tmp);
				gtk_tree_view_column_pack_start(col, renderer, TRUE);
				gtk_tree_view_append_column(GTK_TREE_VIEW(itm->wItem), col);
				gtk_tree_view_column_add_attribute (col, renderer, "text", itm->nColumns);

				gtk_tree_view_column_set_sizing(col, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
				gtk_tree_view_column_set_resizable(col, TRUE);

				head = tail + 1;
			}

			gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(itm->wItem)), GTK_SELECTION_SINGLE);
			g_signal_connect(G_OBJECT(itm->wItem), "cursor-changed", G_CALLBACK(dlg_runtime_listbox_selchanged), itm);
			gtk_widget_show(GTK_WIDGET(itm->wItem));
			break;

		case CTL_NEWLINE_DRAW:
			itm->wItem = gtk_hseparator_new();
			itm->wExtra = gtk_label_new(initstring);
			g_object_ref(itm->wExtra);
			break;

		case CTL_NEWLINE:
			break;

		default:
			free(itm);
			itm = 0;
			break;
		}
		if(itm){
			itm->controltype = ctlType;
			itm->id = id;
			if(itm->wItem){
				g_object_ref(itm->wItem); /* this will prevent GTK from destroying the widget during the dialog's life cycle*/
				gtk_widget_show(itm->wItem);
			}
			dlg->lstCtlItems = list_add(dlg->lstCtlItems, LIST_ELEMENT_CAST(itm));
		}
	}

	return dlg;
}

IPLAT dlg_domodal(DIALOGBOX dialog, char *title, dialog_routine fct){
	IPLAT res = IPLAT_NULL;
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	struct DIALOG *CurrentDialogCopy = CurrentDialog;

	CurrentDialog = dlg;

	if(dlg && dlg->wDialog){
		dlg->cb_resp = IPLAT_NULL;
		dlg->on_altered = fct;
		dlg_runtime_draw_dialog(dialog, title);

		gtk_dialog_run(GTK_DIALOG(dlg->wDialog));
		res = dlg->cb_resp;
	}

	CurrentDialog = CurrentDialogCopy;
	return res;
}

void dlg_set_focus(DIALOGBOX dialog, DIALOGCONTROL item){
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	struct DIALOG_ITEM *itg = (struct DIALOG_ITEM*)(item);
	if(dlg){
		dlg->itmFocus = item;
		if(itg && itg->wItem && GTK_IS_WIDGET(itg->wItem) && gtk_widget_get_can_focus(itg->wItem))
			gtk_window_set_focus(GTK_WINDOW(dlg->wDialog), itg->wItem);
	}
}

void dlg_set_response(DIALOGBOX dialog, IPLAT resp){
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	if(dlg) dlg->cb_resp = resp;
}

DIALOGBOX dlg_destroy(DIALOGBOX dialog){
	struct DIALOG_ITEM *item;
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	if(dlg){
		while(dlg->lstCtlItems && dlg->lstCtlItems->cnt){
			item = (struct DIALOG_ITEM*) list_pop(dlg->lstCtlItems);
			if(item){
				if(item->wItem){
					g_object_unref(item->wItem);	/*release the reference created at the controls creation time*/
					if(GTK_IS_WIDGET(item->wItem))
						gtk_widget_destroy(item->wItem);
				}
				if(item->wExtra){
					g_object_unref(item->wExtra);
					if(GTK_IS_WIDGET(item->wExtra))
						gtk_widget_destroy(item->wExtra);
				}
				free(item);
			}
		}
		dlg->lstCtlItems = list_destroy(dlg->lstCtlItems);
		if(dlg->wDialog){
			if(GTK_IS_WIDGET(dlg->wDialog))
				gtk_widget_destroy(dlg->wDialog);
		}
		free(dlg);
	}
	return 0;
}

char* get_user(void){
	return getenv("USER");
}
char* get_home(void){
	return getenv("HOME");
}

char get_path_seperator(void){
	return '/';
}

int makedir(char *path){
	return mkdir(path, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH | S_ISVTX);
}

int ls_exists(char *path){
	struct stat	 st;
	return (path && (0 == stat(path, &st)) ? 1 : 0);
}
/***********************************************************************************************************
 * PROGRAM ENTRY GTK/GDK
 ***********************************************************************************************************/
int main(int argc, char *argv[]){
	int i, screenw, screenh;
	__sw_platform_init = 0;
	gtk_init(&argc, &argv);
	__sw_platform_init = 1;

	selectedFont[0] = 0;

	fwplat = (PLATFORM_APP*)malloc(sizeof(PLATFORM_APP));
	if(fwplat){
		memset(fwplat, 0, sizeof(PLATFORM_APP));
	} else
		return 1;

	platform_app_init(fwplat);
	fwplat->sys.lstArgv = list_new(LIST_MIN_BLOCK);
	for(i = 1; i < argc; i++)
		fwplat->sys.lstArgv = list_add(fwplat->sys.lstArgv, LIST_ELEMENT_CAST(argv[i]));

	load_systems_defaults();
	load_system_color_sizes();

	fwplat->gui.line = line;
	fwplat->gui.dashline = dashline;
	fwplat->gui.putpixel = putpixel;
	fwplat->gui.clrscr = clrscr;
	fwplat->gui.rectangle = rectangle;
	fwplat->gui.circle = circle;
	fwplat->gui.polygon = polygon;
	fwplat->gui.linepath = linepath;
	fwplat->gui.font_use = font_use;
	fwplat->gui.font_face = font_face;
	fwplat->gui.fonthgt = fonthgt;
	fwplat->gui.fontwdt = fontwdt;

	fwplat->gui.textmode = textmode;
	fwplat->gui.textout = textout;
	fwplat->gui.lenpixels = lenpixels;

	fwplat->gui.title = set_title;
	fwplat->gui.icon = set_icon;

	fwplat->gui.set_intensity = set_intensity;
	fwplat->gui.set_bcolor = set_bcolor;
	fwplat->gui.set_fcolor = set_fcolor;
	fwplat->gui.set_sys_color = set_sys_color;
	fwplat->gui.set_alpha = set_alpha;
	fwplat->gui.swap_colors = swap_colors;

	fwplat->gui.maxx = maxx;
	fwplat->gui.maxy = maxy;
	fwplat->gui.resolutionx = resolutionx;
	fwplat->gui.resolutiony = resolutiony;
	fwplat->gui.caret_set = caret_set;
	fwplat->gui.redraw = gtk_send_redraw;

	fwplat->gui.screencopy = screen_copy;
	fwplat->gui.screenput = screen_put;
	fwplat->gui.screendelete = screen_destroy;

	fwplat->ctl.set_mouse_cursor = set_mouse_cursor;
	fwplat->ctl.get_mouse_cursor = get_mouse_cursor;
	fwplat->ctl.error_msg = error_msg;
	fwplat->ctl.yesno_msg = yesno_msg;
	fwplat->ctl.msgbox = msgbox;
	fwplat->ctl.sbar_set = sbar_set;
	fwplat->ctl.sbar_show = sbar_show;
	fwplat->ctl.tabwnd_create = tabwnd_create;
	fwplat->ctl.tabwnd_set_title = tabwnd_set_title;
	fwplat->ctl.tabwnd_set_lock = tabwnd_set_lock;
	fwplat->ctl.toolbar_add = toolbar_add;
	fwplat->ctl.toolbar_refresh = toolbar_refresh;
	fwplat->ctl.statusbar_add = statusbar_add;
	fwplat->ctl.statusbar_refresh = statusbar_refresh;

	fwplat->ctl.dlg_add_item = dlg_add_item;
	fwplat->ctl.dlg_set_focus = dlg_set_focus;
	fwplat->ctl.dlg_item_type = dlg_item_type;
	fwplat->ctl.dlg_item_show = dlg_item_show;
	fwplat->ctl.dlg_item_shown =dlg_item_shown;
	fwplat->ctl.dlg_item_enable = dlg_item_enable;
	fwplat->ctl.dlg_item_enabled = dlg_item_enabled;
	fwplat->ctl.dlg_item_set_value = dlg_item_set_value;
	fwplat->ctl.dlg_item_get_value = dlg_item_get_value;
	fwplat->ctl.dlg_item_get_value_size = dlg_item_get_value_size;
	fwplat->ctl.dlg_item_get_id = dlg_item_get_id;
	fwplat->ctl.dlg_item_find = dlg_find_item;
	fwplat->ctl.dlg_item_next = dlg_next_item;
	fwplat->ctl.dlg_update = dlg_runtime_draw_dialog;
	fwplat->ctl.dlg_set_response = dlg_set_response;
	fwplat->ctl.dlg_domodal = dlg_domodal;
	fwplat->ctl.dlg_get_userdata = dlg_get_userdata;
	fwplat->ctl.dlg_set_userdata = dlg_set_userdata;
	fwplat->ctl.dlg_destroy = dlg_destroy;

	fwplat->sys.doevents = doevents;
	fwplat->sys.exit = gtk_impl_exit;
	fwplat->sys.total_memory = total_memory;
	fwplat->sys.cboard_text_put = cboard_text_put;	/* clipboard */
	fwplat->sys.cboard_text_sze = cboard_text_sze;
	fwplat->sys.cboard_text_ptr = cboard_text_ptr;
	fwplat->sys.cboard_text_free = cboard_text_free;
	fwplat->sys.timed_call = __impl_gtk_timer_call;
	fwplat->sys.time = __impl_time;
	fwplat->sys.time_delta = __impl_time_delta;
	fwplat->sys.get_home = get_home;
	fwplat->sys.get_user = get_user;
	fwplat->sys.get_path_seperator = get_path_seperator;
	fwplat->sys.mkdir = makedir;
	fwplat->sys.ls_exists = ls_exists;

	screenw = gdk_screen_width();
	screenh = gdk_screen_height();

	/**
	 * Create the main widget layout, confer layout schema at the beginning of this file...
	 **/
	wMain = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	if(wMain){
		if(fwplat->wShapeHint == WS_SQUARE){
			screenw = MIN(screenw, screenh);
			screenh = screenw;
		}
		screenw *= DEFAULT_WINDOW_SIZE;
		screenh *= DEFAULT_WINDOW_SIZE;

		gtk_window_resize((GtkWindow*)wMain, screenw, screenh);

		vBoxMain = gtk_vbox_new(FALSE, 0);
		vBoxMainCenter = gtk_vbox_new(FALSE, 0);
		vBoxDrawContainer = gtk_vbox_new(FALSE, 0);

		vBoxInfoBox = gtk_vbox_new(FALSE, 0);

		wDrawingArea = new_draw_area();
		hBoxDrawingArea = gtk_hbox_new(FALSE, 0);
		vBoxDrawingArea = gtk_vbox_new(FALSE, 0);

		lstStatusBar = list_new(LIST_MIN_BLOCK);
		hBoxStatus = gtk_hbox_new(FALSE, 0);

		lstToolBar = list_new(LIST_MIN_BLOCK);
		wToolBar = GTK_TOOLBAR(gtk_toolbar_new());
		gtk_toolbar_set_style(wToolBar, GTK_TOOLBAR_BOTH);
		gtk_toolbar_set_icon_size(wToolBar, GTK_ICON_SIZE_MENU);

		wNoteBook = GTK_NOTEBOOK(gtk_notebook_new());
		gtk_notebook_set_scrollable(wNoteBook, TRUE);	/*scroll means to scroll the tab buttons inside the notebook*/
		gtk_notebook_set_show_border(wNoteBook, FALSE);
		g_signal_connect(G_OBJECT(wNoteBook), "switch-page", G_CALLBACK(tabwnd_selectpage), 0);
		gtk_box_pack_start(GTK_BOX(vBoxMainCenter), GTK_WIDGET(wNoteBook), TRUE, TRUE, 0);
		gtk_box_pack_start(GTK_BOX(vBoxMainCenter), GTK_WIDGET(vBoxDrawContainer), TRUE, TRUE, 0);

		vadj = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 0, 0, 0, 0));
		vscroll = GTK_VSCROLLBAR(gtk_vscrollbar_new(vadj));
		hadj = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 0, 0, 0, 0));
		hscroll = GTK_HSCROLLBAR(gtk_hscrollbar_new(hadj));

		gtk_container_add(GTK_CONTAINER(wMain), vBoxMain);
		gtk_box_pack_start(GTK_BOX(vBoxMain), GTK_WIDGET(wToolBar), FALSE, FALSE, 0);
		gtk_container_add(GTK_CONTAINER(vBoxMain), vBoxMainCenter);
		gtk_box_pack_end(GTK_BOX(vBoxMain), GTK_WIDGET(hBoxStatus), FALSE, FALSE, 0);

		gtk_box_pack_start(GTK_BOX(vBoxDrawContainer), GTK_WIDGET(vBoxInfoBox), FALSE, FALSE, 0);
		gtk_container_add(GTK_CONTAINER(vBoxDrawContainer), hBoxDrawingArea);

		gtk_container_add(GTK_CONTAINER(hBoxDrawingArea), GTK_WIDGET(vBoxDrawingArea));
		gtk_box_pack_end(GTK_BOX(hBoxDrawingArea), GTK_WIDGET(vscroll), FALSE, FALSE, 0);

		gtk_container_add(GTK_CONTAINER(vBoxDrawingArea), GTK_WIDGET(wDrawingArea));
		gtk_box_pack_start(GTK_BOX(vBoxDrawingArea), GTK_WIDGET(hscroll), FALSE, FALSE, 0);

		gtk_widget_set_events(wMain, GDK_ALL_EVENTS_MASK);
		gtk_window_set_default_size(GTK_WINDOW(wMain), screenw, screenh);
		g_signal_connect(G_OBJECT(wMain), "key-press-event", G_CALLBACK(on_key), NULL);
		g_signal_connect(G_OBJECT(wMain), "key-release-event", G_CALLBACK(on_key_released), NULL);
		g_signal_connect(G_OBJECT(wMain), "focus-out-event", G_CALLBACK(on_focus_out), NULL);
		g_signal_connect(G_OBJECT(wMain), "focus-in-event", G_CALLBACK(on_focus_in), NULL);
		g_signal_connect(G_OBJECT(wMain), "scroll_event", G_CALLBACK(on_mousewheel), NULL);
		g_signal_connect(GTK_OBJECT(wMain), "delete_event", GTK_SIGNAL_FUNC(on_destroy), NULL);
		g_signal_connect(G_OBJECT(wMain), "event-after", G_CALLBACK(on_wmain_event_after), NULL);

		g_signal_connect(G_OBJECT(vadj), "value-changed", G_CALLBACK(on_srollbar_changed), NULL);
		g_signal_connect(G_OBJECT(hadj), "value-changed", G_CALLBACK(on_srollbar_changed), NULL);

		reparent_draw_container(GTK_BOX(vBoxMainCenter));

		gtk_widget_show_all(wMain);
		gtk_widget_hide(GTK_WIDGET(wNoteBook));
		gtk_widget_hide(GTK_WIDGET(hscroll));
		gtk_widget_hide(GTK_WIDGET(vscroll));
		gtk_widget_hide(GTK_WIDGET(hBoxStatus));

		if(lstToolBar->cnt == 0)
			gtk_widget_hide(GTK_WIDGET(wToolBar));

		CairoCtx = local_create_dc_from_window(wMain->window);
		fo = cairo_font_options_create();
		if(fo){
			cairo_font_options_set_hint_style(fo, CAIRO_HINT_STYLE_MEDIUM);
			cairo_font_options_set_hint_metrics(fo, CAIRO_HINT_METRICS_ON);
			cairo_font_options_set_antialias(fo, CAIRO_ANTIALIAS_SUBPIXEL);
			cairo_font_options_set_subpixel_order(fo, CAIRO_SUBPIXEL_ORDER_DEFAULT);
		}

		caret_set(CARET_OFF, CARET_NORMAL, 0, 0, 0);
		__caret_blink_time = int_get_setting(wMain, CARET_BLINK_TIME);
		if(!__caret_blink_time) __caret_blink_time = 1000;
		__caret_blink_time /= 2;
		g_timeout_add(__caret_blink_time, __caret_blink_routine, 0);


		cursors[0] = 0;
		cursors[1] = gdk_cursor_new_for_display(gtk_widget_get_display(wMain), GDK_XTERM);
		cursors[2] = gdk_cursor_new_for_display(gtk_widget_get_display(wMain), GDK_WATCH);

		font_use(FONT_SYSTEM);
		set_sys_color(SYSCOL_TEXT_F, SYSCOL_TEXT_B);
		ongui_begin(fwplat);


		CairoCtx = local_destroy_dc(CairoCtx);

		gtk_main();

		__caret_img = screen_destroy(__caret_img);

		fwplat->sys.lstArgv = list_destroy(fwplat->sys.lstArgv);
		platform_app_destroy(fwplat);

		if(CairoCtx) cairo_destroy(CairoCtx);
		if(fo) cairo_font_options_destroy(fo);

		lstToolBar = list_destroy(lstToolBar);
		lstStatusBar = list_destroy(lstStatusBar);

		if(cursors[1]) gdk_cursor_unref(cursors[1]);

		free(fwplat);
		return __exit_code;
	}

	return -1;
}
