/*
    Copyright (C) 2000, 2009 Vansteenhuyse Hubert <hubert.vansteenhuyse@freecode.be>

	This file is part of MathCandy.

    MathCandy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    MathCandy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MathCandy. If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * Implementatie van platform.h voor GDK/GTK.
 */
#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <glib.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include "platform.h"
#define TXTOUT_MAX (1024 * 16)

//http://library.gnome.org/devel/gtk-tutorial/2.13/x2928.html
//`pkg-config --cflags gdk-2.0` `pkg-config --libs gdk-2.0` `pkg-config --cflags gtk+-2.0` `pkg-config --libs gtk+-2.0`

int __exit_code = 0;

PangoContext	*ctxPango = 0;
GtkWidget	*wMain = 0,  *drawing_area = 0, *vbox = 0;
PLATFORM_APP *fwplat = 0;
static GdkPixmap *pixmap = 0;

GdkColor colf, colb;
GdkModifierType mouseB = 0;
PangoLayout *layout = 0;

volatile int __is_drawing = 0;
int __got_focus = 0, kbd = 0, mouseX = 0, mouseY = 0;

/**
 * timers
 */
volatile int __in_timer_thread = 0;
guint __gsource = 0, __gcaret = 0;

/**
 * caret
 */
void __caret_update(void);
GdkImage 	*imgCaret = 0;
GdkRectangle CaretPos;
int __sw_showCaret = 0, __sw_caret_on = 0;
int __sw_opaque = 0;

/*http://library.gnome.org/devel/gtk/stable/GtkSettings.html*/
#define CARET_BLINK "gtk-cursor-blink"
#define CARET_BLINK_TIME "gtk-cursor-blink-time"
#define CARET_BLINK_TIMEOUT "gtk-cursor-blink-timeout"
gint int_get_setting(const char *name){
	gint n = 0;
	GtkSettings *settings;
	if(wMain){
		settings = gtk_widget_get_settings(GTK_WIDGET(wMain));
		g_object_get (settings, name, &n, NULL);
	}
	return n;
}

void __redraw(){
	if( GTK_IS_WIDGET(drawing_area) )
		gtk_widget_queue_draw(drawing_area);
}


int __pango_font_metrix(int swHeight){
	int wc = 0, wd = 0;
	PangoFontDescription *pdesc = 0;
	PangoFontMetrics *metric = 0;
	if(layout){
		pdesc = (PangoFontDescription*)pango_layout_get_font_description(layout);
		metric = pango_context_get_metrics(pango_layout_get_context(layout), pdesc, 0 );
		if(metric){
			if(!swHeight){
				wc = pango_font_metrics_get_approximate_char_width(metric) / PANGO_SCALE;
				wd = pango_font_metrics_get_approximate_digit_width(metric) / PANGO_SCALE;
				if(wc < wd) wc = wd;
			} else {
				wc = (pango_font_metrics_get_descent(metric) + pango_font_metrics_get_ascent(metric) )/ PANGO_SCALE;
			}
			pango_font_metrics_unref(metric);
		}
	}
	return wc;
}

void clrscr(void){
	if(pixmap)
		gdk_draw_rectangle(pixmap, drawing_area->style->white_gc, TRUE, 0, 0, drawing_area->allocation.width, drawing_area->allocation.height);
}
int sys_fontmetric(char *face, int hgt, int wdt){
	PangoFontDescription *fdesc = 0;
	int res = 0;
	double sze = (double)((hgt + wdt) /2.f * PANGO_SCALE);
	if(layout){
		if(face && strlen(face)) fdesc = pango_font_description_from_string(face);
		//else fdesc = pango_font_description_from_string(def_font_name);
		if(!fdesc) fdesc = pango_context_get_font_description(pango_layout_get_context(layout));
		if(sze) pango_font_description_set_absolute_size(fdesc, sze);
		pango_layout_set_font_description(layout, fdesc);
		pango_layout_context_changed(layout);
	}
	return res;
}
int sys_fontwdt(void){
	return __pango_font_metrix(0);
}
int sys_fonthgt(void){
	return __pango_font_metrix(1);
}

/* foreground color of white_black acts as foreground??? */
void set_fcolor(int r, int g, int b){
	colf.red = r;
	colf.green = g;
	colf.blue = b;
	gdk_gc_set_rgb_fg_color(drawing_area->style->black_gc, &colf);
}
/* foreground color of white_gc acts as background??? */
void set_bcolor(int r, int g, int b){
	colb.red = r;
	colb.green = g;
	colb.blue = b;
	gdk_gc_set_rgb_bg_color(drawing_area->style->white_gc, &colb);
}
void putpixel(int x, int y){
	//gdk_draw_point(drawing_area->window, drawing_area->style->black_gc, x, y);
	gdk_draw_point(pixmap, drawing_area->style->black_gc, x, y);
}
void line(int x1, int y1, int x2, int y2){
	gdk_draw_line(pixmap, drawing_area->style->black_gc, x1, y1, x2, y2);
}
void dashline(int x1, int y1, int x2, int y2){
	gdk_gc_set_line_attributes(drawing_area->style->black_gc, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_NOT_LAST, GDK_JOIN_BEVEL);
	gdk_draw_line(pixmap, drawing_area->style->black_gc, x1, y1, x2, y2);
	gdk_gc_set_line_attributes(drawing_area->style->black_gc, 1, GDK_LINE_SOLID, GDK_CAP_NOT_LAST, GDK_JOIN_BEVEL);
}

void caret_set(IPLAT caret_onoff, IPLAT caret_mode, int x, int y){
	if((x != CaretPos.x || y != CaretPos.y) && caret_onoff == CARET_ON)	__sw_caret_on = 1;
	CaretPos.x = x;
	CaretPos.y = y;
	CaretPos.width = 0;
	CaretPos.height = sys_fonthgt();
	if(caret_mode == CARET_OVW){
		CaretPos.width = sys_fontwdt();
	} else {
		CaretPos.width = 2;
	}

	if(caret_onoff == CARET_OFF)
		__sw_showCaret = 0;
	else{
		if(!__sw_showCaret){
			if(!__sw_showCaret) __sw_caret_on = 1;
			__sw_showCaret = 1;
		}
	}
	__caret_update();
}

void textmode(IPLAT text_mode){
	if(text_mode == TEXT_OPAQUE)
		__sw_opaque = 1;
	else
		__sw_opaque = 0;
}

void textoutva(int x, int y, char *fmt, va_list args){
	char buf[TXTOUT_MAX];
	cairo_t *cr = 0;
	GdkRectangle CairoRectPos;
	if(layout && fmt){
		vsnprintf(buf, TXTOUT_MAX, fmt, args);
		pango_layout_set_text(layout, buf, -1);
		if(__sw_opaque){
			cr = gdk_cairo_create(pixmap);
			if(cr){
				CairoRectPos.x = x;
				CairoRectPos.y = y;
				pango_layout_get_pixel_size(layout, &CairoRectPos.width, &CairoRectPos.height);
				cairo_set_source_rgba(
						cr,
						(double)(colb.red)   / (double)(fwplat->gui.color_maxval),
						(double)(colb.green) / (double)(fwplat->gui.color_maxval),
						(double)(colb.blue)  / (double)(fwplat->gui.color_maxval),
						1.f
				);
				gdk_cairo_rectangle(cr, &CairoRectPos);
				cairo_fill(cr);
				cairo_destroy(cr);
			}
		}
		gdk_draw_layout(pixmap, drawing_area->style->black_gc, x, y, layout);
	}
}

void textout(int x, int y, char *fmt, ...){
	va_list args;
	if(layout && fmt){
		va_start(args, fmt);
		textoutva(x, y, fmt, args);
		va_end(args);
	}
}

int lenpixelsva(int *wdt, int *hgt, char *fmt, va_list args){
	char buf[TXTOUT_MAX];
	int w = 0, h = 0;
	int stat = 0;
	if(layout && fmt){
		vsnprintf(buf, TXTOUT_MAX, fmt, args);
		pango_layout_set_text(layout, buf, -1);
		pango_layout_get_pixel_size(layout, &w, &h);
		stat = 1;
		if(wdt) *wdt = w;
		if(hgt) *hgt = h;
	}
	return stat;
}
int lenpixels(int *wdt, int *hgt, char *fmt, ...){
	int stat = 0;
	va_list args;
	if(layout && fmt){
		va_start(args, fmt);
		stat = lenpixelsva(wdt, hgt, fmt, args);
		va_end(args);
	}
	return stat;
}

int sys_maxx(void){
	return drawing_area->allocation.width;
}
int sys_maxy(void){
	return drawing_area->allocation.height;
}
int sys_resolutionx(void){
	int w=0, h=0;
	GdkWindow *rw = gtk_widget_get_root_window(wMain);
	gdk_drawable_get_size(rw, &w, &h);
	return w;
}
int sys_resolutiony(void){
	int w=0, h=0;
	GdkWindow *rw = gtk_widget_get_root_window(wMain);
	gdk_drawable_get_size(rw, &w, &h);
	return h;
}

int sys_redraw(void){
	if(wMain){
		__redraw();
		return 1;
	}
	return 0;
}

void* screencopy(int x, int y, int wdt, int hgt){
	GdkGC *gc = 0;
	GdkPixmap *pmap = 0;

	if(GTK_IS_WIDGET(drawing_area) && pixmap){
		gc = gdk_gc_new(pixmap);
		if(gc){
			pmap = gdk_pixmap_new(pixmap, sys_maxx(), sys_maxy(), -1);
			gdk_draw_drawable(
					pmap,
					gc,
					pixmap,
					x,
					y,
					0,
					0,
					wdt,
					hgt
			);
			g_object_unref(gc);
		}
	}
	return pmap;
}

void screenput(void *img, int scrx, int scry, int imgx, int imgy, int wdt, int hgt){
	GdkGC *gc = 0;
	if(GTK_IS_WIDGET(drawing_area) && img){
		gc = gdk_gc_new(pixmap);
		if(gc){
			gdk_draw_drawable(
					pixmap,
					gc,
					(GdkPixmap*)img,
					imgx,
					imgy,
					scrx,
					scry,
					wdt,
					hgt
			);
			g_object_unref(gc);
		}
	}
}

void screendelete(void *img){
	if(img) g_object_unref(img);
}

/* Create a new backing pixmap of the appropriate size */
static gboolean on_configure(GtkWidget *widget, GdkEventConfigure *event){
	GdkVisual *gv = 0;
	if(widget == drawing_area){
		gv = gtk_widget_get_visual(widget);
		if(gv){
			if(pixmap) g_object_unref(pixmap);
			pixmap = gdk_pixmap_new(widget->window, widget->allocation.width, widget->allocation.height, gv->depth);
		}
	}
	return TRUE;
}

static void on_destroy(GtkWidget *widget, gpointer data){
	gtk_main_quit();
	wMain = 0;
}

static gboolean on_draw(GtkWidget *widget, GdkEventExpose *event){
	if(widget == drawing_area && GTK_IS_WIDGET(drawing_area) && event){
		if(!__is_drawing){
			__is_drawing = 1;
			if(pixmap){
				fwplat->gui.color_bg.r = wMain->style->base[GTK_WIDGET_STATE (widget)].red;
				fwplat->gui.color_bg.g = wMain->style->base[GTK_WIDGET_STATE (widget)].green;
				fwplat->gui.color_bg.b = wMain->style->base[GTK_WIDGET_STATE (widget)].blue;

				fwplat->gui.color_fg.r = wMain->style->fg[GTK_WIDGET_STATE (widget)].red;
				fwplat->gui.color_fg.g = wMain->style->fg[GTK_WIDGET_STATE (widget)].green;
				fwplat->gui.color_fg.b = wMain->style->fg[GTK_WIDGET_STATE (widget)].blue;

				fwplat->gui.color_sel_bg.r = wMain->style->base[GTK_STATE_SELECTED].red;
				fwplat->gui.color_sel_bg.g = wMain->style->base[GTK_STATE_SELECTED].green;
				fwplat->gui.color_sel_bg.b = wMain->style->base[GTK_STATE_SELECTED].blue;

				fwplat->gui.color_sel_fg.r = wMain->style->fg[GTK_STATE_SELECTED].red;
				fwplat->gui.color_sel_fg.g = wMain->style->fg[GTK_STATE_SELECTED].green;
				fwplat->gui.color_sel_fg.b = wMain->style->fg[GTK_STATE_SELECTED].blue;

				if(layout) g_object_unref(layout);
				//ctxPango = gdk_pango_context_get(); 			//<-----<< context has to be freed = memory leak
				ctxPango= gtk_widget_get_pango_context(wMain);	//<-----<< 'correct' way to get a context, owned by the widget
				if(ctxPango){
					layout = pango_layout_new(ctxPango);
					if(layout){
						//redraw to the pixmap
						fwplat->onredraw(fwplat);

						//copy pixmap to widget
						gdk_draw_drawable(drawing_area->window, drawing_area->style->white_gc, pixmap, event->area.x, event->area.y, event->area.x, event->area.y, event->area.width, event->area.height);
						__caret_update();
					}
				}
			}
		  __is_drawing = 0;
		}
	}
  return TRUE;
}
static void on_mouse(GtkWidget *widget, GdkEventMotion *event){
	IPLAT mB = IPLAT_NULL;
	if(widget == drawing_area && GTK_IS_WIDGET(widget)){
		gdk_window_get_pointer(event->window, &mouseX, &mouseY, &mouseB);
		if(mouseB & GDK_RELEASE_MASK){
			//if(mouseB & GDK_BUTTON2_MASK) mB = MBTN_RIGHT_UP;
			if(mouseB & GDK_BUTTON1_MASK) mB = MBTN_LEFT_UP;
		} else {
			//if(mouseB & GDK_BUTTON2_MASK) mB = MBTN_RIGHT_IN;
			if(mouseB & GDK_BUTTON1_MASK) mB = MBTN_LEFT_IN;
		}
		fwplat->onmouse(fwplat, mouseX, mouseY, mB);
	}
}

static gboolean on_key(GtkWidget *widget, GdkEventKey *event, gpointer user_data){
	PLATKEY pk = PKAN;
	char *kname;

	if(GTK_IS_WIDGET(widget) && fwplat && fwplat->onkey && event){
		kname = gdk_keyval_name(event->keyval);
		if(kname){
			if(0 == strcasecmp(kname, "down"))pk = PKDN;
			else if(0 == strcasecmp(kname, "up"))pk = PKUP;
			else if(0 == strcasecmp(kname, "left"))pk = PKLE;
			else if(0 == strcasecmp(kname, "right"))pk = PKRI;
			else if(0 == strcasecmp(kname, "page_down"))pk = PKPD;
			else if(0 == strcasecmp(kname, "page_up"))pk = PKPU;
			else if(0 == strcasecmp(kname, "insert"))pk = PKIN;
			else if(0 == strcasecmp(kname, "home"))pk = PKHM;
			else if(0 == strcasecmp(kname, "end"))pk = PKED;
			else if(0 == strcasecmp(kname, "delete"))pk = PKDL;
			else if(0 == strcasecmp(kname, "backspace"))pk = PKBK;
		}
		if(pk == PKAN)
			fwplat->onkey(fwplat, event->string[(event->length ? event->length - 1 : 0)]);
		else
			fwplat->onskey(fwplat, pk);
	}
	return TRUE;
}

static gboolean on_focus_out(GtkWidget *widget, GdkEventMotion *event){
	if(widget == wMain){
		__got_focus = 0;
		__sw_caret_on = 0;
		//__redraw();
		__caret_update();
	}
	return FALSE;
}
static gboolean on_focus_in(GtkWidget *widget, GdkEventMotion *event){
	if(widget == wMain){
		__got_focus = 1;
		__sw_caret_on = 1;
		//__redraw();
		__caret_update();
	}
	return FALSE;
}

void DoEvents(){
	while(gtk_events_pending())
		gtk_main_iteration();
}

/**
 * Caret routine
 */
int __caret_is_updating = 0, sw_caret_mem_inv = 0;
void __caret_update(){
	GdkImage *imgScreen;
	int swAan = 0,swI = 0;
	int i, j, pc, ps;

	if(!__caret_is_updating && GTK_IS_WIDGET(drawing_area)){
		__caret_is_updating = 1;
		//de caret in geheugen
		if(!imgCaret){
			imgCaret = gdk_drawable_copy_to_image(drawing_area->window, 0, CaretPos.x, CaretPos.y, 0, 0, CaretPos.width, CaretPos.height);
			if(!imgCaret) goto caret_done;
		}

		//moet hij aan of uit?
		if(__sw_showCaret && __sw_caret_on && __got_focus) swAan = 1;

		if(imgCaret->width != CaretPos.width || imgCaret->height != CaretPos.height){
			if(imgCaret) g_object_unref(imgCaret);
			imgCaret = gdk_drawable_copy_to_image(drawing_area->window, 0, CaretPos.x, CaretPos.y, 0, 0, CaretPos.width, CaretPos.height);
			if(!imgCaret) goto caret_done;
		}

		imgScreen = gdk_drawable_copy_to_image(drawing_area->window, 0, CaretPos.x, CaretPos.y, 0, 0, CaretPos.width, CaretPos.height);
		if(imgScreen){
			if(!swAan && sw_caret_mem_inv) swI = 1;
			for(i = 0; i < imgCaret->height; i++){
				for(j = 0; j < CaretPos.width; j++){
					ps = gdk_image_get_pixel(imgScreen, j, i);
					pc = gdk_image_get_pixel(imgCaret, j, i);
					if(pc != ps) pc = ps;
					pc = (swAan || swI ? 0xFFFFFF - pc : pc);
					gdk_image_put_pixel(imgCaret, j, i, pc);
				}
			}
			sw_caret_mem_inv = swAan;
			gdk_draw_image(drawing_area->window, drawing_area->style->white_gc, imgCaret, 0, 0, CaretPos.x, CaretPos.y, CaretPos.width, CaretPos.height);
			g_object_unref(imgScreen);
		}

	caret_done:
		__caret_is_updating = 0;
	}

}

static gint __caret_blink(gpointer arg){
	if(__sw_showCaret)__sw_caret_on = (__sw_caret_on ? 0 : 1);
	__caret_update();
	return TRUE;
}

gboolean __thread_timer(gpointer arg){
	DoEvents();
	if(!__in_timer_thread){
		__in_timer_thread = 1;
		__redraw();
		__in_timer_thread = 0;
	}
	return TRUE;
}

void sys_timer_redraw(int msec){ /*msec > 1 timer runs, msec = 0 timer stops*/
	if(__gsource){
		g_source_remove(__gsource);
		__gsource = 0;
		__in_timer_thread = 0;
	}

	if(msec < 0) msec *= -1;

	if(msec){
		//if(msec < 50) msec = 50;
		if(msec < 25) msec = 25;
		__gsource = g_timeout_add(msec, __thread_timer, 0);
	}
}

void sys_exit(int exitcode){
	__exit_code = exitcode;
	gtk_main_quit();
}

/**
 * that's one for the list of stupid error messages...
 * when expose events are still being processed GTK displays an empty error box :)
 * nice... there is no way of preventing this!
 */
char msgbuf[1024 * 16];
void sys_error_msg(char *fmt, ...){
	va_list args;
	GtkWidget *werr;

	DoEvents();

	va_start(args, fmt);
	vsprintf(msgbuf, fmt, args);
	va_end(args);

	werr = gtk_message_dialog_new(GTK_WINDOW(wMain), GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, "%s", msgbuf);
	if(werr){
		gtk_window_set_title(GTK_WINDOW(werr), "Error");
		gtk_dialog_run(GTK_DIALOG(werr));
		gtk_widget_destroy(werr);
	}
}

int main(int argc, char *argv[]){
	int screenw, screenh, msecblink;

	fwplat = platform_app_init();
	if(fwplat){
		fwplat->gui.color_maxval = 65535;

		fwplat->gui.DoEvents = DoEvents;
		fwplat->gui.line = line;
		fwplat->gui.dashline = dashline;
		fwplat->gui.putpixel = putpixel;
		fwplat->gui.set_bcolor = set_bcolor;
		fwplat->gui.set_fcolor = set_fcolor;
		fwplat->gui.sys_maxx = sys_maxx;
		fwplat->gui.sys_maxy = sys_maxy;
		fwplat->gui.sys_resolutionx = sys_resolutionx;
		fwplat->gui.sys_resolutiony = sys_resolutiony;
		fwplat->gui.textoutva = textoutva;
		fwplat->gui.textout = textout;
		fwplat->gui.textout_uni = textout; //<-----<< textout_uni is enkel nodig onder windows
		fwplat->gui.caret_set = caret_set;
		fwplat->gui.textmode = textmode;
		fwplat->gui.lenpixelsva = lenpixelsva;
		fwplat->gui.lenpixels = lenpixels;
		fwplat->gui.lenpixels_uni = lenpixels; //<-----<< lenpixels_uni is enkel nodig onder windows
		fwplat->gui.clrscr = clrscr;
		fwplat->gui.sys_fontmetric = sys_fontmetric;
		fwplat->gui.sys_fonthgt = sys_fonthgt;
		fwplat->gui.sys_fontwdt = sys_fontwdt;
		fwplat->gui.sys_redraw = sys_redraw;
		fwplat->gui.sys_timer_redraw = sys_timer_redraw;
		fwplat->gui.sys_exit = sys_exit;
		fwplat->gui.sys_error_msg = sys_error_msg;
		fwplat->gui.screencopy = screencopy;
		fwplat->gui.screenput = screenput;
		fwplat->gui.screendelete = screendelete;


		gdk_init(&argc, &argv);
		gtk_init(&argc, &argv);
		screenw = gdk_screen_width();
		screenh = gdk_screen_height();
		screenw = screenw * 4 / 10;
		screenh = screenh * 5 / 10;

		wMain = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		if(wMain){
			vbox = gtk_vbox_new(FALSE, 0);
			gtk_container_add(GTK_CONTAINER(wMain), vbox);

			drawing_area = gtk_drawing_area_new();
			gtk_widget_set_size_request(GTK_WIDGET(drawing_area), 200, 200);
			gtk_box_pack_start(GTK_BOX(vbox), drawing_area, TRUE, TRUE, 0);

			gtk_window_set_title(GTK_WINDOW(wMain), (fwplat->title ? fwplat->title : "GtkApp"));
			gtk_window_set_default_size(GTK_WINDOW(wMain), screenw, screenh);

			gtk_widget_set_events(
					drawing_area,
					GDK_EXPOSURE_MASK 				|
					GDK_CONFIGURE					|
					GDK_POINTER_MOTION_MASK			|
					GDK_BUTTON_PRESS_MASK			|
					GDK_BUTTON_RELEASE_MASK			|
					GDK_POINTER_MOTION_HINT_MASK	|
					GDK_ENTER_NOTIFY_MASK			|
					GDK_LEAVE_NOTIFY_MASK
			);
			g_signal_connect(G_OBJECT(drawing_area), "motion_notify_event", G_CALLBACK(on_mouse), NULL);
			g_signal_connect(G_OBJECT(drawing_area), "button-press-event", G_CALLBACK(on_mouse), NULL);
			g_signal_connect(G_OBJECT(drawing_area), "button-release-event", G_CALLBACK(on_mouse), NULL);
			g_signal_connect(G_OBJECT(drawing_area), "expose_event", G_CALLBACK(on_draw), NULL);
			g_signal_connect(G_OBJECT(drawing_area), "configure_event", G_CALLBACK(on_configure), NULL);
			g_signal_connect(G_OBJECT(wMain), "key-press-event", G_CALLBACK(on_key), NULL);
			g_signal_connect(G_OBJECT(wMain), "focus-out-event", G_CALLBACK(on_focus_out), NULL);
			g_signal_connect(G_OBJECT(wMain), "focus-in-event", G_CALLBACK(on_focus_in), NULL);
			g_signal_connect(G_OBJECT(wMain), "destroy", G_CALLBACK(on_destroy), NULL);
			gtk_widget_show(vbox);
			gtk_widget_show(drawing_area);
			gtk_widget_show(wMain);

			caret_set(CARET_OFF, CARET_NORMAL, 0, 0);
			msecblink = int_get_setting(CARET_BLINK_TIME);
			if(msecblink)
				__gcaret = g_timeout_add(msecblink / 2, __caret_blink, 0);

			if(fwplat->ongui_begin) fwplat->ongui_begin(fwplat);
			gtk_main();

			if(__gcaret) g_source_remove(__gcaret);
			__gcaret = 0;

			fwplat->onexit(fwplat);
			fwplat = platform_app_destroy(fwplat);

			if(layout) g_object_unref(layout);
			if(imgCaret) g_object_unref(imgCaret);

			return __exit_code;
		}
	}
	return -1;
}

