/*
    Copyright (C) 2000, 2009 Vansteenhuyse Hubert <hubert.vansteenhuyse@freecode.be>

	This file is part of MathCandy.

    MathCandy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    MathCandy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MathCandy. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __PLATFORM_FRAMEWORK_H__
#define __PLATFORM_FRAMEWORK_H__
#include <stdarg.h>
#include <wchar.h>
#include <math.h>
#include <float.h>
#define FP_TYPE double
#define FP_MAX DBL_MAX

typedef enum IPLAT{
	IPLAT_NULL = 0,
	E_ERR,
	E_OK,
	MBTN_LEFT_IN,
	MBTN_LEFT_UP,
	CARET_ON,
	CARET_OFF,
	CARET_NORMAL,
	CARET_OVW,
	TEXT_TRANSPARENT,
	TEXT_OPAQUE,
}IPLAT;

typedef enum PLATKEY{
	PKAN, /*alphanumeric key*/
	PKUP, PKDN, PKLE, PKRI, PKPU, PKPD,
	PKHM, PKED, PKIN,	/*home, end, insert*/
	PKDL, PKBK,		/*delete, backspace*/
}PLATKEY;

typedef struct C_RGBA{
	int r,g,b,a;
}C_RGBA;

typedef struct PLATFORM_GUI{
	int		color_maxval;
	C_RGBA	color_fg,
			color_bg,
			color_sel_bg,
			color_sel_fg;
	void *sysdata;

	void (*clrscr)(void);
	void (*DoEvents)(void);
	void (*caret_set)(IPLAT caret_onoff, IPLAT caret_mode, int x, int y);

	void (*set_fcolor)(int r, int g, int b);
	void (*set_bcolor)(int r, int g, int b);
	void (*putpixel)(int x, int y);
	void (*line)(int x1, int y1, int x2, int y2);
	void (*dashline)(int x1, int y1, int x2, int y2);

	void (*textmode)(IPLAT text_mode);

	/**
	 * Reason for splitting routines into unicode and charachter bytes is the mess between interpretation of
	 * different platforms and compilers :(
	 *
	 * UNICODE GLYPHS version of lenpixels and textout.
	 * Unicode output format: "%lc", UNICODE_VALUE
	 * !!! A "%s" must also be in wide chars "%ls" is not supported by mingw
	 */
	int (*lenpixels_uni)(int *wdt, int *hgt, char *fmt, ...);
	void (*textout_uni)(int x, int y, char *fmt, ...);

	/**
	 * CHARACHTER BYTES versionsof lenpixels and textout
	 * returns: width of string in pixels
	 */
	void (*textout)(int x, int y, char *fmt, ...);
	void (*textoutva)(int x, int y, char *fmt, va_list);
	int (*lenpixels)(int *wdt, int *hgt, char *fmt, ...);
	int (*lenpixelsva)(int *wdt, int *hgt, char *fmt, va_list);

	int (*sys_maxx)(void);
	int (*sys_maxy)(void);
	int (*sys_resolutionx)(void);
	int (*sys_resolutiony)(void);

	int (*sys_fontmetric)(char *face, int hgt, int wdt);
	int (*sys_fontwdt)(void);
	int (*sys_fonthgt)(void);
	int (*sys_redraw)(void);

	void (*sys_timer_redraw)(int msec); /*msec > 1 timer runs, msec = 0 timer stops*/

	void (*sys_exit)(int exitcode);
	void (*sys_error_msg)(char *fmt, ...);

	void* (*screencopy)(int x, int y, int wdt, int hgt);
	void (*screenput)(void*, int scrx, int scry, int imgx, int imgy, int wdt, int hgt);
	void (*screendelete)(void*);

}PLATFORM_GUI;

typedef struct PLATFORM_APP {
	/*fsys part to be assigned by the platform dependent GUI code*/
	PLATFORM_GUI gui;

	/*part to be assigned by the platform independent LOGIC code*/
	const char *title;
	void (*onkey)(struct PLATFORM_APP*, int);
	void (*onskey)(struct PLATFORM_APP*, PLATKEY);
	void (*ongui_begin)(struct PLATFORM_APP*);
	void (*onexit)(struct PLATFORM_APP*);
	void (*onredraw)(struct PLATFORM_APP*);
	void (*onmouse)(struct PLATFORM_APP*, int x, int y, IPLAT b);

} PLATFORM_APP;

PLATFORM_APP* platform_app_init();
PLATFORM_APP* platform_app_destroy(PLATFORM_APP*);

/******************************************************************************************************
 * platform commons
 * implementation independent routines
 */
void __do_log(char *fname, char *str);

/*drawables*/
void drawArc(PLATFORM_APP *fwplat, int x, int y, FP_TYPE r, FP_TYPE anglefrom, FP_TYPE angletill);
void drawCircle(PLATFORM_APP *fwplat, int x, int y, FP_TYPE r);

/*Time*/
typedef struct uT { long int sec, usec; } uT;
uT utnow();
uT utdelta(uT);
double utsec(uT);


#endif /*__PLATFORM_FRAMEWORK_H__*/
