/*
    Copyright (C) 2000, 2009 Vansteenhuyse Hubert <hubert.vansteenhuyse@freecode.be>

	This file is part of MathCandy.

    MathCandy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    MathCandy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MathCandy. If not, see <http://www.gnu.org/licenses/>.
*/

#include <sys/time.h>
#include "platform.h"

#define USEC 1000000

#include <stdio.h>
void __do_log(char *fname, char *str){
	FILE *f;
	if(!fname || !str) return;
	f  = fopen(fname, "a");
	if(f){
		fprintf(f, "%s\n", str);
		fclose(f);
	}
}


/*
 * returns the milliseconds since 1970
 */
uT utnow(){
	struct timeval tv;
	uT ut;
	if(-1 == gettimeofday(&tv, 0)){
		ut.sec = ut.usec = 0;
	}else{
		ut.sec = (long int)tv.tv_sec;
		ut.usec = (long int)tv.tv_usec;
	}
	return ut;
}
/*
 * returns the difference between now and time taken in structure uT ut
 */
uT utdelta(uT ut){
	long int us;
	uT now = utnow();
		us = now.usec - ut.usec;
		now.usec = us % USEC;
		now.sec = (us / USEC) + now.sec - ut.sec;
	return now;
}
/*
 * converts structure uT ut into a double scalar
 */
double utsec(uT ut){
	double d;
		d = (double)(ut.usec) / (double)(USEC);
		d += (double)(ut.sec);
	return d;
}


void drawArc(PLATFORM_APP *fwplat, int x, int y, FP_TYPE r, FP_TYPE anglefrom, FP_TYPE angletill){
	FP_TYPE i, c, s;
	if(fwplat && r){
		if(anglefrom > angletill){
			i = anglefrom;
			anglefrom = angletill;
			angletill = i;
		}
		i = (angletill - anglefrom) / (2.f * M_PI * r);
		if(!i) return;
		while(anglefrom <= angletill){
			s = -sin(anglefrom) * r + y;
			c = cos(anglefrom) * r + x;
			fwplat->gui.putpixel( c, s);
			anglefrom += i;
		}
	}
}


/**
 * cirkel tekenen (niet de elegantste noch meest effici�nte manier)
 */
void drawCircle(PLATFORM_APP *fwplat, int x, int y, FP_TYPE r){
	FP_TYPE t, i;
	int c, s;
	if(fwplat && r){
		/*i = interval voor hoek theta om de cirkelomtrek volledig te kunnen dekken */
		i = 1.f / (2.f * M_PI * r);
		if(!i) return;
		/* doorloop een halve cirkel per interval i en teken pixel op hoek theta en hoek -theta */
		t = 0;
		while(t < M_PI){
			c = (int)(cos(t) * r);
			s = (int)(sin(t) * r);
			fwplat->gui.putpixel( x + c, y + s);
			fwplat->gui.putpixel( x - c, y - s);
			t += i;
		}
	}
}

