/*
    Copyright (C) 2000, 2009 Vansteenhuyse Hubert <hubert.vansteenhuyse@freecode.be>

	This file is part of MathCandy.

    MathCandy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    MathCandy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MathCandy. If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * http://msdn.microsoft.com/en-us/library/ms648409%28VS.85%29.aspx
 * WIN7 :(
 * Wdm.h
 * Ntdll.dll
 * RtlUnicodeToUTF8N
 * RtlUTF8ToUnicodeN
 *
 * win2000
 * Winternl.h
 * RtlUnicodeStringToAnsiString
 * RtlAnsiStringToUnicodeString
 * Ntdll.dll
 * Winternl.h
 *
 * Implementatie van platform.h voor Windows GDI.
 * Met anti-flickering bij het updaten van het venster.
 */
#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include <stdlib.h>
#include "platform.h"
#define TXTOUT_MAX (1024 * 16)

int sys_redraw(void);

int sys_maxx(void);
int sys_maxy(void);

typedef struct WINDC{
	HDC		windc, bmpdc, oldbmpdc;
	HBITMAP	bmp;
}WINDC;

WINDC mdc;

HMODULE hWinGdi = 0;
int kbd = 0, mouseX = 0, mouseY = 0;
IPLAT mouseB = IPLAT_NULL;
COLORREF fcol, bcol;

HGDIOBJ OldPen = 0, OldBrush = 0, OldFont = 0;
HPEN Pen = 0;
HBRUSH Brush = 0;
HFONT Font = 0;

int __got_focus = 0, __sw_showCaret = 0, __sw_caret_insert = 0;
int __sw_opaque = 0;

BOOL WINAPI (CALLBACK* dllCreateFont)(int nHeight, int nWidth, int nEscapement, int nOrientation, int fnWeight, DWORD fdwItalic, DWORD fdwUnderline, DWORD fdwStrikeOut, DWORD fdwCharSet, DWORD fdwOutputPrecision, DWORD fdwClipPrecision, DWORD fdwQuality, DWORD fdwPitchAndFamily, LPCTSTR lpszFace) = 0;
BOOL WINAPI (CALLBACK* dllGetTextFace)(HDC hdc, int n, LPTSTR name) = 0;
BOOL WINAPI (CALLBACK* dllLineTo)(HDC hdc, int nXEnd, int nYEnd) = 0;
BOOL WINAPI (CALLBACK* dllMoveToEx)(HDC hdc, int X, int Y, LPPOINT lpPoint) = 0;
BOOL WINAPI (CALLBACK* dllRectangle)(HDC hdc, int nLeftRect, int nTopRect, int nRightRect, int nBottomRect) = 0;

BOOL WINAPI (CALLBACK* dllTextOut)(HDC hdc, int x, int y, char*, int len) = 0;
BOOL WINAPI (CALLBACK* dllTextOutW)(HDC hdc, int x, int y, wchar_t*, int len) = 0;
BOOL WINAPI (CALLBACK* dllGetTextExtentPoint)(HDC hdc, char *str, int c, LPSIZE lpSize);
BOOL WINAPI (CALLBACK* dllGetTextExtentPointW)(HDC hdc, wchar_t* lpString, int c, LPSIZE lpSize);

BOOL WINAPI (CALLBACK* dllSetPixel)(HDC hdc, int X, int Y, COLORREF crColor) = 0;
BOOL WINAPI (CALLBACK* dllCreatePen)(int fnPenStyle, int nWidth, COLORREF crColor) = 0;
BOOL WINAPI (CALLBACK* dllCreateSolidBrush)(COLORREF crColor) = 0;
BOOL WINAPI (CALLBACK* dllSelectObject)(HDC hdc, HGDIOBJ hgdiobj) = 0;
BOOL WINAPI (CALLBACK* dllDeleteObject)(HGDIOBJ hObject) = 0;
BOOL WINAPI (CALLBACK* dllSetTextColor)(HDC hdc, COLORREF crColor) = 0;
BOOL WINAPI (CALLBACK* dllSetBkColor)(HDC hdc, COLORREF crColor) = 0;
BOOL WINAPI (CALLBACK* dllCreateCompatibleDC)(HDC hdc) = 0;
BOOL WINAPI (CALLBACK* dllBitBlt)(HDC hdcDest, int nXDest, int nYDest, int nWidth, int nHeight,	HDC hdcSrc, int nXSrc, int nYSrc, DWORD dwRop) = 0;
BOOL WINAPI (CALLBACK* dllCreateCompatibleBitmap)(HDC hdc, int nWidth, int nHeight) = 0;
BOOL WINAPI (CALLBACK* dllGetTextMetrics)(HDC hdc, LPTEXTMETRIC lptm) = 0;
BOOL WINAPI (CALLBACK* dllSetBkMode)(HDC hdc, int mode) = 0;

int swExit = 0;

HWND hwndMain;
WNDCLASSEX wndclass;
char *szMainWndClass = "WinApp";

UINT_PTR	__timer_handler = 0;
volatile int __is_drawing = 0;


PLATFORM_APP *fwplat = 0;

void __begin_draw(){
	DWORD ncolor;

	if(!__is_drawing && !fwplat->gui.sysdata){
		__is_drawing = 1;

		//http://msdn.microsoft.com/en-us/library/dd144927%28VS.85%29.aspx
		//GetSysColor
		fwplat->gui.color_maxval = 255;
		ncolor = GetSysColor(COLOR_WINDOW);
		fwplat->gui.color_bg.r = GetRValue(ncolor);
		fwplat->gui.color_bg.g = GetGValue(ncolor);
		fwplat->gui.color_bg.b = GetBValue(ncolor);

		ncolor = GetSysColor(COLOR_WINDOWTEXT);
		fwplat->gui.color_fg.r = GetRValue(ncolor);
		fwplat->gui.color_fg.g = GetGValue(ncolor);
		fwplat->gui.color_fg.b = GetBValue(ncolor);

		ncolor = GetSysColor(COLOR_HIGHLIGHT);
		fwplat->gui.color_sel_bg.r = GetRValue(ncolor);
		fwplat->gui.color_sel_bg.g = GetGValue(ncolor);
		fwplat->gui.color_sel_bg.b = GetBValue(ncolor);

		ncolor = GetSysColor(COLOR_HIGHLIGHTTEXT);
		fwplat->gui.color_sel_fg.r = GetRValue(ncolor);
		fwplat->gui.color_sel_fg.g = GetGValue(ncolor);
		fwplat->gui.color_sel_fg.b = GetBValue(ncolor);

    	mdc.windc = GetDC(hwndMain);
    	mdc.bmpdc = (HDC)dllCreateCompatibleDC(mdc.windc);
    	mdc.bmp = (HBITMAP)dllCreateCompatibleBitmap(mdc.windc, sys_maxx(), sys_maxy());
    	mdc.oldbmpdc = (HDC)dllSelectObject(mdc.bmpdc, mdc.bmp);
		fwplat->gui.sysdata = &mdc;
	}
}
void __end_draw(){
	if(__is_drawing && fwplat->gui.sysdata){
		dllBitBlt(mdc.windc, 0, 0, sys_maxx(), sys_maxy(), mdc.bmpdc, 0, 0, SRCCOPY);
    	fwplat->gui.sysdata = 0;

		if(OldPen) dllSelectObject(mdc.bmpdc, OldPen);
		OldPen = 0;
		if(Pen) dllDeleteObject(Pen);
		Pen = 0;

		if(OldBrush) dllSelectObject(mdc.bmpdc, OldBrush);
		OldBrush = 0;
		if(Brush) dllDeleteObject(Brush);
		Brush = 0;

		if(OldFont) dllSelectObject(mdc.bmpdc, OldFont);
		OldFont = 0;
		if(Font) dllDeleteObject(Font);
		Font = 0;

		dllDeleteObject(mdc.oldbmpdc);
		dllDeleteObject(mdc.bmp);
    	dllDeleteObject(mdc.bmpdc);
    	ReleaseDC(hwndMain, mdc.windc);
    	mdc.bmpdc = mdc.windc = 0;
		__is_drawing = 0;
	}
}

void clrscr(void){
	RECT r;
	if(fwplat->gui.sysdata && dllRectangle){
		GetClientRect(hwndMain, &r);
		dllRectangle(mdc.bmpdc, r.left-1, r.top-1, r.right+1, r.bottom+1);
	}
}

int sys_fontmetric(char *face, int hgt, int wdt){
	char fface[64 * 4];
	if(!fwplat->gui.sysdata) return 0;

	if(!face || !strlen(face))
		dllGetTextFace(mdc.windc, 64*4, fface);
	else
		sprintf(fface, "%s", face);

	if(OldFont) dllSelectObject(mdc.bmpdc, OldFont);
	OldFont = 0;
	if(Font) dllDeleteObject(Font);
	Font = (HFONT)dllCreateFont(hgt, wdt, 0, 0, FW_DONTCARE, 0, 0, 0, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, fface);
	if(Font) OldFont = (HFONT)dllSelectObject(mdc.bmpdc, Font);

	return hgt * wdt;
}

int __txt_metric(int swHgt){
	int n = 0;
	TEXTMETRIC tm;
	if(fwplat->gui.sysdata){
		tm.tmHeight = 0;
		tm.tmAveCharWidth = 0;
		dllGetTextMetrics(mdc.bmpdc, &tm);
		if(swHgt)
			n = tm.tmHeight;
		else
			n = tm.tmAveCharWidth;
	}
	return n;
}

int sys_fontwdt(void){
	return __txt_metric(0);
}
int sys_fonthgt(void){
	return __txt_metric(1);
}
void set_fcolor(int r, int g, int b){
	if(fwplat->gui.sysdata){
		fcol = RGB(r, g, b);

		if(OldPen) dllSelectObject(mdc.bmpdc, OldPen);
		OldPen = 0;
		if(Pen) dllDeleteObject(Pen);

		Pen = (HPEN)dllCreatePen(PS_SOLID, 0, fcol);
		if(Pen) OldPen = (HPEN)dllSelectObject(mdc.bmpdc, Pen);

		dllSetTextColor(mdc.bmpdc, fcol);
	}
}
void set_bcolor(int r, int g, int b){
	if(fwplat->gui.sysdata){
		bcol = RGB(r, g, b);

		if(OldBrush) dllSelectObject(mdc.bmpdc, OldBrush);
		OldBrush = 0;
		if(Brush) dllDeleteObject(Brush);

		Brush = (HBRUSH)dllCreateSolidBrush(bcol);
		if(Brush) OldBrush = (HBRUSH)dllSelectObject(mdc.bmpdc, Brush);

		dllSetBkColor(mdc.bmpdc, bcol);
	}
}
void putpixel(int x, int y){
	if(fwplat->gui.sysdata && dllSetPixel){
		dllSetPixel(mdc.bmpdc, x, y, fcol);
	}
}
void line(int x1, int y1, int x2, int y2){
	if(fwplat->gui.sysdata && dllLineTo && dllMoveToEx){
		dllMoveToEx(mdc.bmpdc, x1, y1, 0);
		dllLineTo(mdc.bmpdc, x2, y2);
	}
}
void dashline(int x1, int y1, int x2, int y2){
	HPEN DashPen = 0, CurPen = 0;
	if(fwplat->gui.sysdata && dllLineTo && dllMoveToEx){
		DashPen = (HPEN)dllCreatePen(PS_DASH, 0, fcol);
		if(DashPen) CurPen = (HPEN)dllSelectObject(mdc.bmpdc, DashPen);
		line(x1, y1, x2, y2);
		if(CurPen) dllSelectObject(mdc.bmpdc, CurPen);
		if(DashPen) dllDeleteObject(DashPen);
	}
}

void textmode(IPLAT text_mode){
	if(text_mode == TEXT_OPAQUE)
		__sw_opaque = 1;
	else
		__sw_opaque = 0;
}


/***************************************************************************
 * KARAKTER VERSIE VAN TEXTOUT, LENPIXELS en METRICS
 ***************************************************************************/
void __string_metrics(char *str, int *wdt, int *hgt){
	SIZE metric;
	if(fwplat->gui.sysdata && str){
		metric.cx = metric.cy = 0;
		dllGetTextExtentPoint(mdc.bmpdc, str, strlen(str), &metric);
		if(wdt) *wdt = metric.cx;
		if(hgt) *hgt = metric.cy;
	}
}

void textoutva(int x, int y, char *fmt, va_list args){
	char buf[TXTOUT_MAX];
	int i;
	if(fwplat->gui.sysdata && fmt){
		vsnprintf(buf, TXTOUT_MAX, fmt, args);
		i = strlen(buf);
		dllSetBkMode( mdc.bmpdc, (__sw_opaque ? OPAQUE : TRANSPARENT) );
		dllTextOut(mdc.bmpdc, x, y, buf, i);
	}
}

void textout(int x, int y, char *fmt, ...){
	va_list args;
	if(fwplat->gui.sysdata && fmt){
    	va_start(args, fmt);
    	textoutva(x, y, fmt, args);
    	va_end(args);
	}
}

int lenpixelsva(int *wdt, int *hgt, char *fmt, va_list args){
	char buf[TXTOUT_MAX];
	if(fwplat->gui.sysdata && fmt){
		vsnprintf(buf, TXTOUT_MAX, fmt, args);
		__string_metrics(buf, wdt, hgt);
	}
	return 0;
}

int lenpixels(int *wdt, int *hgt, char *fmt, ...){
	va_list args;
	int stat = 0;
	if(fwplat->gui.sysdata && fmt){
    	va_start(args, fmt);
    	stat = lenpixelsva(wdt, hgt, fmt, args);
    	va_end(args);
	}
	return stat;
}

/***************************************************************************
 * UNICODE VERSIE VAN TEXTOUT, LENPIXELS en METRICS
 ***************************************************************************/
void __string_metrics_uni(wchar_t *str, int *wdt, int *hgt){
	SIZE metric;
	if(fwplat->gui.sysdata && str){
		metric.cx = metric.cy = 0;
		dllGetTextExtentPointW(mdc.bmpdc, str, wcslen(str), &metric);
		if(wdt) *wdt = metric.cx;
		if(hgt) *hgt = metric.cy;
	}
}

void textout_uni(int x, int y, char *fmt, ...){
	va_list args;
	wchar_t wbuf[TXTOUT_MAX],
			wfmt[TXTOUT_MAX];
	int i = 0, j, l;

	if(fwplat->gui.sysdata && fmt){
		l = strlen(fmt);
		i = 0;
		for(j = 0; j < l; j++) i += mbtowc(wfmt + i, fmt + j, MB_CUR_MAX);
		wfmt[i] = 0;
		va_start(args, fmt);
		_vsnwprintf(wbuf, TXTOUT_MAX, wfmt, args);
		va_end(args);
		i = wcslen(wbuf);
		dllSetBkMode( mdc.bmpdc, (__sw_opaque ? OPAQUE : TRANSPARENT) );
		dllTextOutW(mdc.bmpdc, x, y, wbuf, i);
	}
}

int lenpixels_uni(int *wdt, int *hgt, char *fmt, ...){
	va_list args;
	wchar_t wbuf[TXTOUT_MAX],
			wfmt[TXTOUT_MAX];
	int stat = 0, i = 0, j, l;
	if(fwplat->gui.sysdata && fmt){
		l = strlen(fmt);
		i = 0;
		for(j = 0; j < l; j++) i += mbtowc(wfmt + i, fmt + j, MB_CUR_MAX);
		wfmt[i] = 0;
		va_start(args, fmt);
		_vsnwprintf(wbuf, TXTOUT_MAX, wfmt, args);
		va_end(args);
		stat = 1;
		__string_metrics_uni(wbuf, wdt, hgt);
	}
	return stat;
}

void caret_set(IPLAT caret_onoff, IPLAT caret_mode, int x, int y){
	int w, h;
	if(hwndMain){
		if(caret_onoff == CARET_OFF){
			//HideCaret(hwndMain);
			DestroyCaret();
			__sw_showCaret = 0;
		} else {
			h = sys_fonthgt();
			if(caret_mode == CARET_OVW)
				w = sys_fontwdt();
			else
				w = 2;
			CreateCaret(hwndMain, (HBITMAP)0, w, h);
			SetCaretPos(x, y);
			if(__got_focus) ShowCaret(hwndMain);
			__sw_showCaret = 1;
		}
	}
}

int sys_maxx(void){
	RECT r;
	GetClientRect(hwndMain, &r);
	return (int)(r.right - r.left);
}
int sys_maxy(void){
	RECT r;
	GetClientRect(hwndMain, &r);
	return (int)(r.bottom - r.top);
}

int sys_resolutionx(void){
	RECT r;
	GetClientRect(GetDesktopWindow(), &r);
	return (int)(r.right - r.left);
}
int sys_resolutiony(void){
	RECT r;
	GetClientRect(GetDesktopWindow(), &r);
	return (int)(r.bottom - r.top);
}


int sys_redraw(void){
	if(hwndMain){
		return (int)RedrawWindow(hwndMain, 0, 0, RDW_INTERNALPAINT);
	}
	return 0;
}

void DoEvents(){
	MSG msg;
	while(!swExit && PeekMessage(&msg, hwndMain, 0, 0, PM_REMOVE)){
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}

char msgbuf[1024 * 16];
void sys_error_msg(char *fmt, ...){
	va_list args;

	va_start(args, fmt);
	vsprintf(msgbuf, fmt, args);
	va_end(args);

	MessageBoxA(0, msgbuf, "Error", MB_ICONERROR | MB_OK);
}

void* screencopy(int x, int y, int wdt, int hgt){
	HBITMAP	bmpshot = 0;
	HDC dcshot = 0;

	if(fwplat->gui.sysdata){
    	dcshot = (HDC)dllCreateCompatibleDC(mdc.windc);
    	if(dcshot){
    		bmpshot = (HBITMAP)dllCreateCompatibleBitmap(mdc.bmpdc, wdt, hgt);
    		if(bmpshot){
    	    	dllSelectObject(dcshot, bmpshot);
    			if( !dllBitBlt(dcshot, 0, 0, wdt, hgt, mdc.bmpdc, x, y, SRCCOPY) ){
    				dllDeleteObject((HBITMAP)bmpshot);
    				bmpshot = 0;
    			}
    		}
    	   	dllDeleteObject(dcshot);
    	}
	}

	return bmpshot;
}

void screenput(void *bmpshot, int scrx, int scry, int imgx, int imgy, int wdt, int hgt){
	HDC dcshot = 0;
	if(bmpshot && fwplat->gui.sysdata){
		dcshot = (HDC)dllCreateCompatibleDC(mdc.windc);
    	if(dcshot){
        	dllSelectObject(dcshot, (HBITMAP)bmpshot);
    		dllBitBlt(mdc.bmpdc, scrx, scry, wdt, hgt, dcshot, imgx, imgy, SRCCOPY);
    	}
	   	dllDeleteObject(dcshot);
	}
}
void screendelete(void *bmpshot){
	if(bmpshot) dllDeleteObject((HBITMAP)bmpshot);
}


void CALLBACK __thread_timer(HWND wnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime){
	sys_redraw();
}

void sys_timer_redraw(int msec){ /*msec > 1 timer runs, msec = 0 timer stops*/
	if(hwndMain){
		if(__timer_handler){
			KillTimer(hwndMain, __timer_handler);
			__timer_handler = 0;
		}
		if(msec < 0) msec *= -1;
		if(msec){
			//if(msec < 50) msec = 50;
			if(msec < 25) msec = 25;
			__timer_handler = SetTimer(hwndMain, 0, msec, __thread_timer);
		}
	}
}
void sys_exit(int exitcode){
	PostQuitMessage(exitcode);
}

int MB_LEFT_IN = 0;
LRESULT CALLBACK MainWndProc(HWND hwnd, UINT nMsg, WPARAM wParam, LPARAM lParam){
	PLATKEY pk = PKAN;

    switch (nMsg) {
		case WM_LBUTTONUP:
			mouseX = LOWORD(lParam);
			mouseY = HIWORD(lParam);
			MB_LEFT_IN = 0;
			fwplat->onmouse(fwplat, mouseX, mouseY, MBTN_LEFT_UP);
			break;
		case WM_LBUTTONDOWN:
			mouseX = LOWORD(lParam);
			mouseY = HIWORD(lParam);
			MB_LEFT_IN = 1;
			fwplat->onmouse(fwplat, mouseX, mouseY, MBTN_LEFT_IN);
			break;
		case WM_MOUSEMOVE:
			mouseX = LOWORD(lParam);
			mouseY = HIWORD(lParam);
			mouseB = IPLAT_NULL;

			if(MB_LEFT_IN == 1) mouseB = MBTN_LEFT_IN;
			else mouseB = MBTN_LEFT_UP;

			fwplat->onmouse(fwplat, mouseX, mouseY, mouseB);
			break;
        case WM_SETFOCUS:
			if(__sw_showCaret) ShowCaret(hwndMain);
			__got_focus = 1;
			sys_redraw();
            break;

        case WM_KILLFOCUS:
        	if(__sw_showCaret) HideCaret(hwndMain);
        	__got_focus = 0;
        	sys_redraw();
            break;

        case WM_DESTROY:
        	swExit = 1;
        	fwplat = platform_app_destroy(fwplat);
        	PostQuitMessage(0);
        	break;

        case WM_PAINT:
        	//update the screen
        	if(!__is_drawing){
				__begin_draw();
					fwplat->onredraw(fwplat);
				__end_draw();
        	}
    	   break;

        case WM_KEYDOWN:
			kbd = (int)wParam;
			if(kbd == VK_DOWN)pk = PKDN;
			else if(kbd == VK_UP)pk = PKUP;
			else if(kbd == VK_LEFT)pk = PKLE;
			else if(kbd == VK_RIGHT)pk = PKRI;
			else if(kbd == VK_NEXT)pk = PKPD;
			else if(kbd == VK_PRIOR)pk = PKPU;
			else if(kbd == VK_INSERT)pk = PKIN;
			else if(kbd == VK_HOME)pk = PKHM;
			else if(kbd == VK_END)pk = PKED;
			else if(kbd == VK_DELETE)pk = PKDL;
			else if(kbd == VK_BACK)pk = PKBK;
			if(pk != PKAN) fwplat->onskey(fwplat, pk);
			break;

        case WM_CHAR:
			kbd = (int)wParam;
			fwplat->onkey(fwplat, kbd);
			break;
    }
    return DefWindowProc(hwnd, nMsg, wParam, lParam);
}

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow){
	int posx = CW_USEDEFAULT, posy = CW_USEDEFAULT, wdt = CW_USEDEFAULT, hgt = CW_USEDEFAULT;
	BOOL bRet;
	MSG msg;

	fwplat = platform_app_init();
	if(!fwplat) return -1;

	fwplat->gui.DoEvents = DoEvents;
	fwplat->gui.line = line;
	fwplat->gui.dashline = dashline;
	fwplat->gui.putpixel = putpixel;
	fwplat->gui.set_bcolor = set_bcolor;
	fwplat->gui.set_fcolor = set_fcolor;
	fwplat->gui.sys_maxx = sys_maxx;
	fwplat->gui.sys_maxy = sys_maxy;
	fwplat->gui.sys_resolutionx = sys_resolutionx;
	fwplat->gui.sys_resolutiony = sys_resolutiony;
	fwplat->gui.textmode = textmode;
	fwplat->gui.textoutva = textoutva;
	fwplat->gui.textout = textout;
	fwplat->gui.textout_uni = textout_uni;
	fwplat->gui.caret_set = caret_set;
	fwplat->gui.lenpixelsva = lenpixelsva;
	fwplat->gui.lenpixels = lenpixels;
	fwplat->gui.lenpixels_uni = lenpixels_uni;
	fwplat->gui.clrscr = clrscr;
	fwplat->gui.sys_fontmetric = sys_fontmetric;
	fwplat->gui.sys_fontwdt = sys_fontwdt;
	fwplat->gui.sys_fonthgt = sys_fonthgt;
	fwplat->gui.sys_redraw = sys_redraw;
	fwplat->gui.sys_timer_redraw = sys_timer_redraw;
	fwplat->gui.sys_exit = sys_exit;
	fwplat->gui.sys_error_msg = sys_error_msg;
	fwplat->gui.screencopy = screencopy;
	fwplat->gui.screenput = screenput;
	fwplat->gui.screendelete = screendelete;

	hWinGdi = LoadLibrary("gdi32.dll");
	if(hWinGdi){

		dllLineTo = GetProcAddress(hWinGdi, "LineTo");
		dllMoveToEx = GetProcAddress(hWinGdi, "MoveToEx");
		dllRectangle = GetProcAddress(hWinGdi, "Rectangle");
		dllTextOut = GetProcAddress(hWinGdi, "TextOutA");
		dllTextOutW = GetProcAddress(hWinGdi, "TextOutW");
		dllGetTextExtentPoint = GetProcAddress(hWinGdi, "GetTextExtentPoint32A");
		dllGetTextExtentPointW = GetProcAddress(hWinGdi, "GetTextExtentPoint32W");
		dllSetPixel = GetProcAddress(hWinGdi, "SetPixel");
		dllCreatePen = GetProcAddress(hWinGdi, "CreatePen");
		dllCreateSolidBrush = GetProcAddress(hWinGdi, "CreateSolidBrush");
		dllSelectObject = GetProcAddress(hWinGdi, "SelectObject");
		dllDeleteObject = GetProcAddress(hWinGdi, "DeleteObject");
		dllSetTextColor = GetProcAddress(hWinGdi, "SetTextColor");
		dllSetBkColor = GetProcAddress(hWinGdi, "SetBkColor");
		dllGetTextFace = GetProcAddress(hWinGdi, "GetTextFaceA");
		dllCreateFont  = GetProcAddress(hWinGdi, "CreateFontA");
		dllCreateCompatibleDC = GetProcAddress(hWinGdi, "CreateCompatibleDC");
		dllBitBlt = GetProcAddress(hWinGdi, "BitBlt");
		dllCreateCompatibleBitmap = GetProcAddress(hWinGdi, "CreateCompatibleBitmap");
		dllGetTextMetrics = GetProcAddress(hWinGdi, "GetTextMetricsA");
		dllSetBkMode = GetProcAddress(hWinGdi, "SetBkMode");

		if(	!dllGetTextExtentPointW || !dllTextOutW || !dllTextOut || !dllGetTextExtentPoint ||
			!dllLineTo || !dllMoveToEx || !dllRectangle || !dllSetPixel || !dllCreatePen || !dllCreateSolidBrush ||
			!dllSelectObject || !dllDeleteObject || !dllSetTextColor || !dllSetBkColor || !dllGetTextFace || !dllCreateFont ||
			!dllCreateCompatibleDC || !dllBitBlt || !dllCreateCompatibleBitmap || !dllGetTextMetrics ||
			!dllSetBkMode
		){
			MessageBoxA(0, "Some required routines were NOT found in gdi32.dll", (fwplat->title ? fwplat->title : "Abort"), MB_ICONERROR | MB_OK);
			return 1;
		}


	    HICON hico = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(200));
	    memset(&wndclass, 0, sizeof(WNDCLASSEX));
	    wndclass.lpszClassName = szMainWndClass;
	    wndclass.cbSize = sizeof(WNDCLASSEX);
	    wndclass.style = CS_HREDRAW | CS_VREDRAW;
	    wndclass.lpfnWndProc = MainWndProc;
	    wndclass.hbrBackground = 0; //(HBRUSH)(COLOR_WINDOW + 1);
	    wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	    wndclass.hIcon = hico;
	    wndclass.hIconSm = hico;
	    RegisterClassEx(&wndclass);

	    hwndMain = CreateWindow(szMainWndClass, szMainWndClass, WS_OVERLAPPEDWINDOW,
	    		     posx, posy, wdt, hgt,
	    		     NULL, NULL, NULL, NULL);
	    if(hwndMain) {
	    	if(fwplat->title) SetWindowText(hwndMain, fwplat->title);
	    	fwplat->ongui_begin(fwplat);
	    	ShowWindow(hwndMain, SW_SHOWNORMAL);
	    	UpdateWindow(hwndMain);
	    } else {
			MessageBoxA(0, "Failed to create a window", (fwplat->title ? fwplat->title : "Abort"), MB_ICONERROR | MB_OK);
	    	return 2;
	    }


		while(!swExit && (bRet = GetMessage( &msg, hwndMain, 0, 0 )) != 0){
			if (bRet == -1){
				// handle the error and possibly exit
			} else {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}

		FreeLibrary(hWinGdi);
	} else {
		MessageBoxA(0, "Failed to load gdi32.dll", (fwplat->title ? fwplat->title : "Abort"), MB_ICONERROR | MB_OK);
		return 3;
	}
	return 0;
}

