/*
    Copyright (C) 2000, 2009 Vansteenhuyse Hubert <hubert.vansteenhuyse@freecode.be>

	This file is part of MathCandy.

    MathCandy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    MathCandy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MathCandy. If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <memory.h>
#include "platform.h"

#define ARAD (1.f / 180.f * M_PI)
#define ADEG (1.f / M_PI * 180.f)
#define SWITCH(v) ( v ? 0 : 1 )

/* Unicode output format: "%lc", UNICODE_x */
#define UNICODE_THETA 0x03B8
#define UNICODE_PHI 0x03C6
#define UNICODE_ANGLE 0x2220
#define UNICODE_DEGREE 0x00B0
#define UNICODE_PI 0x3C0
#define UNICODE_SQUARED 0x00B2
#define UNICODE_SQUAREROOT 0x221A
#define UNICODE_DELTA 0x2206
#define UNICODE_OMEGA 0x03C9
#define UNICODE_SUBSCRIPT_0	0x2080
#define UNICODE_SUBSCRIPT_X 0x2093


#define TIMER_INTERVAL 1
#define FIX_RAD_INTERVAL (.00001f)
#define FIX_RAD_INTERVAL_MAX (.25f)
#define FIX_RAD_INTERVAL_ADD ( FIX_RAD_INTERVAL_MAX / 100.f )

#define RUITJES 20

/**
 * golven
 */
#define LIST_DIM (1024 * 2)
int		ln = 0;
uT		utStart;
FP_TYPE	lcos[LIST_DIM],
		lsin[LIST_DIM],
		ltime[LIST_DIM];

//aaargh...!!! RGB max values can differ on platforms (win = 255, gtk = 65536)
#define MAXCOLOR 	fwplat->gui.color_maxval

#define COLOR_BACK fwplat->gui.set_bcolor(MAXCOLOR, MAXCOLOR, MAXCOLOR)
#define COLOR_FORE fwplat->gui.set_fcolor(0, 0, 0)
#define COLOR_WATERMARK fwplat->gui.set_fcolor(MAXCOLOR * 7 / 10, MAXCOLOR * 7 / 10, MAXCOLOR * 7 / 10)
#define COLOR_WATERMARK_DARK fwplat->gui.set_fcolor(MAXCOLOR * 5 / 8, MAXCOLOR * 5 / 8, MAXCOLOR * 5 / 8)
#define COLOR_RUITJES fwplat->gui.set_fcolor(MAXCOLOR * 7 / 8, MAXCOLOR * 7 / 8, MAXCOLOR)				/*licht blauw*/


#define COLOR_MENU fwplat->gui.set_fcolor(MAXCOLOR / 2, MAXCOLOR / 2, MAXCOLOR)
#define COLOR_HYP COLOR_WATERMARK
#define COLOR_COS fwplat->gui.set_fcolor(MAXCOLOR / 2, 0, 0)
#define COLOR_SEC fwplat->gui.set_fcolor(MAXCOLOR, 0, 0)
#define COLOR_SIN fwplat->gui.set_fcolor(0, MAXCOLOR / 2, 0)
#define COLOR_CSC fwplat->gui.set_fcolor(0, MAXCOLOR * 3 / 4, 0)
#define COLOR_TAN fwplat->gui.set_fcolor(0, 0, MAXCOLOR)
#define COLOR_COT fwplat->gui.set_fcolor(MAXCOLOR * 3 / 4, MAXCOLOR * 3 / 4, MAXCOLOR)
#define COLOR_CANS fwplat->gui.set_fcolor(MAXCOLOR / 2, MAXCOLOR / 2, 0)				/*co en secans donker geel*/




#ifdef __PLATFORM_WINDOWS__
char* FONTNAME = "Microsoft Sans Serif";
#endif
#ifdef __PLATFORM_GTK__
char* FONTNAME = "";
#endif
int FONTH = 10, FONTW = 10;

typedef enum SCREEN{
	SCR_HELP		= 0,
	SCR_GONIO 		= 1,
}SCREEN;

char *helpMotion = "motion: [r]otation, [+]faster, [-] slower, [d]irection, [space] mouse movement waves";
char *helpWave = "waves: [c]lear, [*]all, [0]none, [1]cos, [2]sin, [3]tan, [4]cot, [5]sec, [6]csc";
SCREEN scr = SCR_GONIO;

FP_TYPE	mX = 0, mY = 0;
IPLAT 	mB = 0;
int		swDoRotate = 0, swDoWave = 0,swWaveCos = 1, swWaveSin = 1, swWaveTan = 1, swWaveCot = 0, swWaveSec = 0, swWaveCsc = 0;
int		sw_caret_on = 1;
uT		utCaret;

FP_TYPE 	fix_rad = 50.f, fix_angle = 0,
		angle_interval = FIX_RAD_INTERVAL + FIX_RAD_INTERVAL_ADD * 10.f , rot_dir = 1;

int kwadrant(int lncos, int lnsin){
	if(lncos >= 0 && lnsin >= 0) return 1;
	else if(lncos <= 0 && lnsin >= 0) return 2;
	else if(lncos <= 0 && lnsin <= 0) return 3;
	return 4;
}

FP_TYPE radians_line(FP_TYPE lncos, FP_TYPE lnsin){
	FP_TYPE vhyp = sqrt(lncos*lncos + lnsin * lnsin),
			vcos = lncos / vhyp,
			theta = acos(vcos);
	if(kwadrant(lncos, lnsin) > 2) theta = M_PI * 2.f - theta;
	return theta;
}

void onmouse(struct PLATFORM_APP *fwplat, int x, int y, IPLAT b){
	FP_TYPE lncos, lnsin, vcos, vhyp;
	if(fwplat){
		mX = (FP_TYPE)(x);
		mY = (FP_TYPE)(y);
		mB = b;

			lncos = mX - (FP_TYPE)fwplat->gui.sys_maxx() / 2.f;
			lnsin = mY - (FP_TYPE)fwplat->gui.sys_maxy() / 2.f;
			lnsin *= -1;
			vhyp = sqrt(lncos*lncos + lnsin *lnsin);

			if(!swDoRotate){
				vcos = lncos / vhyp;
				fix_angle = acos(vcos);
				if(kwadrant(lncos, lnsin) > 2) fix_angle = M_PI * 2.f - fix_angle;
			}

			if(b == MBTN_LEFT_IN){
				fix_rad = vhyp;
			}
		if(!swDoRotate && !swDoWave)
			fwplat->gui.sys_redraw();
	}
}
void onskey(struct PLATFORM_APP *fwplat, PLATKEY k){

}
void onkey(PLATFORM_APP *fwplat, int k){
	FP_TYPE vhyp, vcos, lncos, lnsin;
	if(fwplat){
		//printf("key %c pressed...\n", k);
		if((char)k == 'q' || (char)k == 'Q' || (char)k == 'x' || (char)k == 'X'){
			fwplat->gui.sys_exit(0);
		}
		if((char)k == 'c'){
			ln = 0;
			fwplat->gui.sys_redraw();
		}
		if((char)k == 'r'){
			swDoRotate = SWITCH(swDoRotate);
			swDoWave = swDoRotate;

			fwplat->gui.sys_timer_redraw(TIMER_INTERVAL * swDoRotate);
			if(swDoRotate){
				ln = 0;
				utStart = utnow();
				/*angle theta = position of the mouse regarding to the screen center*/
				lncos = mX - ((FP_TYPE)fwplat->gui.sys_maxx() / 2.f);
				lnsin = mY - ((FP_TYPE)fwplat->gui.sys_maxy() / 2.f);
				vhyp = sqrt(lncos*lncos + lnsin*lnsin);
				vcos = lncos / vhyp;
				fix_angle = acos(vcos);
				lnsin *= -1;
				if(kwadrant(lncos, lnsin) > 2) fix_angle = M_PI * 2.f - fix_angle;
			}
		}
		if((char)k == '+'){
			angle_interval += FIX_RAD_INTERVAL_ADD;
			if(angle_interval > FIX_RAD_INTERVAL_MAX) angle_interval = FIX_RAD_INTERVAL_MAX;
		}
		if((char)k == '-'){
			angle_interval -= FIX_RAD_INTERVAL_ADD;
			if(angle_interval < 0) angle_interval = 0;
		}
		if((char)k == 'd'){
			rot_dir *= -1;
		}
		if((char)k == ' '){
			if(!swDoRotate){
				swDoWave = SWITCH(swDoWave);
				fwplat->gui.sys_timer_redraw(TIMER_INTERVAL * swDoWave);
				ln = 0;
				utStart = utnow();
			}
		}
		if((char)k == '0'){
			swWaveCos = swWaveCot = swWaveSin = swWaveTan = swWaveCsc = swWaveSec = 0;
			if(!swDoWave) fwplat->gui.sys_redraw();
		}
		if((char)k == '*'){
			swWaveCos = swWaveCot = swWaveSin = swWaveTan = swWaveCsc = swWaveSec = 1;
			if(!swDoWave) fwplat->gui.sys_redraw();
		}
		if((char)k == '1'){
			swWaveCos = SWITCH(swWaveCos);
			if(!swDoWave) fwplat->gui.sys_redraw();
		}
		if((char)k == '2'){
			swWaveSin = SWITCH(swWaveSin);
			if(!swDoWave) fwplat->gui.sys_redraw();
		}
		if((char)k == '3'){
			swWaveTan = SWITCH(swWaveTan);
			if(!swDoWave) fwplat->gui.sys_redraw();
		}
		if((char)k == '4'){
			swWaveCot = SWITCH(swWaveCot);
			if(!swDoWave) fwplat->gui.sys_redraw();
		}
		if((char)k == '5'){
			swWaveSec = SWITCH(swWaveSec);
			if(!swDoWave) fwplat->gui.sys_redraw();
		}
		if((char)k == '6'){
			swWaveCsc = SWITCH(swWaveCsc);
			if(!swDoWave) fwplat->gui.sys_redraw();
		}


	}
}


int druit = 0;
void drawFrame(PLATFORM_APP *fwplat, int x, int y){
	int i, mx, my;
	if(fwplat){
		my = fwplat->gui.sys_maxy();
		mx = fwplat->gui.sys_maxx();

		if(swDoWave){
			i = 0;
			if(ln) i = (int)(ltime[ln - 1] * RUITJES);
			druit = i % RUITJES;
			if(druit > RUITJES) druit = 0;
		}

		COLOR_RUITJES;
		for(i = 0; i < my; i += RUITJES) fwplat->gui.line(0, i, mx, i);
		for(i = 0; i < mx + RUITJES; i += RUITJES) fwplat->gui.line(i - druit, my, i - druit, 0);
		COLOR_WATERMARK;
		fwplat->gui.line(x, 0, x, my);
		fwplat->gui.line(0, y, mx, y);
		if(swDoWave){
			fwplat->gui.textout_uni(x + FONTW, 0, "X+");
			fwplat->gui.textout_uni(mx - FONTW * 3, y + FONTH, "T+");
		} else {
			fwplat->gui.textout_uni(x + FONTW, 0, "Y+");
			fwplat->gui.textout_uni(mx - FONTW * 3, y + FONTH, "X+");
		}
	}
}

int linesegment(PLATFORM_APP *fwplat,int x, int y, char *name, FP_TYPE length){
	int n = 0, nx = 0;
	if(fwplat){
		fwplat->gui.lenpixels_uni(&nx, 0, name);
		fwplat->gui.textout_uni(x, y, name, name);
		fwplat->gui.textout_uni(x + nx, y, " %.2f", length);
		nx = x + nx;
		fwplat->gui.line(x, y, nx, y);
	}
	return n;
}

void draw_waves(PLATFORM_APP *fwplat){
	int i;
	FP_TYPE dt, df, w;
	FP_TYPE xbase, ybase, ymax, tbase;
	FP_TYPE vhyp, vcos, vsin, vtan, vcoty, vsec, vcsc;
	FP_TYPE theta, thetamark = -1;
	FP_TYPE xtime, otime;
	FP_TYPE ocosy, cosy;
	FP_TYPE osiny, siny;
	FP_TYPE otany, tany;
	FP_TYPE ocoty, coty;
	FP_TYPE osecy, secy;
	FP_TYPE ocscy, cscy;


	if(fwplat && ln){
		vhyp = sqrt(lcos[ln-1]*lcos[ln-1] + lsin[ln-1]*lsin[ln-1]);
		xtime = ltime[ln - 1];
		otime = (ln > 2 ? ltime[ln - 2] : 0);
		dt = xtime - otime;
		cosy = radians_line(lcos[ln - 1], lsin[ln - 1]);
		ocosy =radians_line( (ln > 2 ? lcos[ln - 2] : 0), (ln > 2 ? lsin[ln - 2] : 0)  );
		df = fabs(cosy - ocosy);
		if(df > M_PI){
			if(ocosy > M_PI)
				cosy = M_PI * 2.f + cosy;
			else if(cosy > M_PI)
				ocosy = M_PI * 2.f + ocosy;
			df = fabs(cosy - ocosy);
		}

		w = (df * (60.f / dt)) / (2.f * M_PI);

		xbase = fwplat->gui.sys_maxx() * 7.25 / 10;
		ybase = fwplat->gui.sys_maxy() / 1.75;
		COLOR_WATERMARK;
		fwplat->gui.textout_uni(xbase, ybase + 1 * FONTH, "%lc%lc(t) = %.2f RPM (%.2fHz)", UNICODE_OMEGA, UNICODE_THETA, w, w / 60.f );
		fwplat->gui.textout_uni(xbase, ybase + 2 * FONTH, "%lc%lc = %f%lc", UNICODE_DELTA, UNICODE_THETA, df * ADEG, UNICODE_DEGREE);
		fwplat->gui.textout_uni(xbase, ybase + 3 * FONTH, "%lct%lc = %f msec", UNICODE_DELTA, UNICODE_SUBSCRIPT_X, dt);
		fwplat->gui.textout_uni(xbase, ybase + 4 * FONTH, "%lct%lc = %f sec", UNICODE_DELTA, UNICODE_SUBSCRIPT_0, ltime[ln - 1]);
		fwplat->gui.textout_uni(xbase, ybase + 5 * FONTH, "v%lc%lc = %.0f%%", UNICODE_OMEGA, UNICODE_THETA, angle_interval / FIX_RAD_INTERVAL_MAX  * 100.f);

		if(swWaveCos || swWaveSin || swWaveTan || swWaveCot || swWaveSec || swWaveCsc){
			i = ln;
			ymax = fwplat->gui.sys_maxx();
			xbase = fwplat->gui.sys_maxx() / 2;
			ybase = fwplat->gui.sys_maxy() / 2;
			tbase = (ltime[i - 1]);

			otime = xbase;
			ocosy = osiny = otany = ocoty = osecy = ocscy = ybase;
			do{
				i--;
				xtime = xbase + (ltime[i] - tbase) * RUITJES;
				vhyp = sqrt(lcos[i]*lcos[i] + lsin[i]*lsin[i]);
				vcos = lcos[i] / vhyp;
				vsin = lsin[i] / vhyp;
				vtan = vsin / vcos;
				vcoty = 1.f / vtan;
				vcsc = 1.f / vsin;
				vsec = 1.f / vcos;
				theta = acos(vcos);
				if(kwadrant(lcos[i], lsin[i]) > 2)
					theta = M_PI * 2.f - theta;
				theta = round(theta * ADEG);

				COLOR_WATERMARK;
				if(thetamark != theta && !((int)(theta) % 90) && theta != 360.f){
					fwplat->gui.textout_uni(xtime, ybase - FONTH, "%.0f%lc", theta, UNICODE_DEGREE);
					thetamark = theta;
				}

				if(swWaveCos){
					COLOR_COS;
					cosy = ybase - vcos * vhyp;
					fwplat->gui.line(otime, ocosy, xtime, cosy);
					ocosy = cosy;
				}

				if(swWaveSin){
					COLOR_SIN;
					siny = ybase - vsin * vhyp;
					fwplat->gui.line(otime, osiny, xtime, siny);
					osiny = siny;
				}

				if(swWaveTan){
					tany = ybase - vtan * vhyp;
					if( fabs(tany) > ymax ){
						otany = -FP_MAX;
					} else {
						COLOR_TAN;
						if(otany == -FP_MAX) otany = tany;
						fwplat->gui.line(otime, otany, xtime, tany);
						otany = tany;
					}
				}

				if(swWaveCot){
					coty = ybase - vcoty * vhyp;
					if( fabs(coty) > ymax ){
						ocoty = -FP_MAX;
					} else {
						COLOR_COT;
						if(ocoty == -FP_MAX) ocoty = coty;
						fwplat->gui.line(otime, ocoty, xtime, coty);
						ocoty = coty;
					}
				}

				if(swWaveCsc){
					cscy = ybase - vcsc * vhyp;
					if( fabs(cscy) > ymax ){
						ocscy = -FP_MAX;
					} else {
						COLOR_CSC;
						if(ocscy == -FP_MAX) ocscy = cscy;
						fwplat->gui.line(otime, ocscy, xtime, cscy);
						ocscy = cscy;
					}
				}
				if(swWaveSec){
					secy = ybase - vsec * vhyp;
					if( fabs(secy) > ymax ){
						osecy = -FP_MAX;
					} else {
						COLOR_SEC;
						if(osecy == -FP_MAX) osecy = secy;
						fwplat->gui.line(otime, osecy, xtime, secy);
						osecy = secy;
					}
				}
				otime = xtime;
			}while(i > 0 && xtime > 0);
		}
	}
}


void drawGonio(PLATFORM_APP *fwplat, int x, int y, FP_TYPE lncos, FP_TYPE lnsin){
	int 	txtwdt, posy, posx, n;
	FP_TYPE	vhyp = 0,	/*Hypotenusa =  radius cirkel*/
			vtan = 0,	/*tangens*/
			vcos = 0, 	/*cosinus*/
			vsin = 0,	/*sinus*/
			vcsc = 0,	/*cosecans*/
			vsec = 0,	/*secans*/
			vcot = 0,	/*cotangens*/
			theta = 0, 	/*hoek theta in radialen*/
			i = 0;

	/*gradenboog*/
	COLOR_WATERMARK;
	fwplat->gui.sys_fontmetric(FONTNAME, 10, 6);
	theta = 0;
	vhyp = ((FP_TYPE)(x) + (FP_TYPE)(y)) / 2.5f;
	i =  15.f * ARAD;
	while(theta < M_PI * 2.f - ARAD){
		vcos =  cos(theta) * vhyp;
		vsin = -sin(theta) * vhyp;
		fwplat->gui.line(x + vcos / 2.5f, y + vsin / 2.5f , x + vcos / 2.0, y + vsin / 2.0);
		fwplat->gui.textout_uni(x + vcos / 2.5f, y + vsin / 2.5f, "%.0f", theta * ADEG);
		theta += i;
	}

	fwplat->gui.sys_fontmetric(FONTNAME, 10, 10);
	for(i = 0; i < 360; i+= 15){
		vcos =  cos(i * ARAD) * (vhyp / 1.9f);
		vsin = -sin(i * ARAD) * (vhyp / 1.9f);
		if(!i) fwplat->gui.textout_uni(x + vcos, y + vsin, "2%lc", UNICODE_PI);
		else if(i == 30) fwplat->gui.textout_uni(x + vcos, y + vsin, "%lc/6", UNICODE_PI);
		else if(i == 45) fwplat->gui.textout_uni(x + vcos, y + vsin, "%lc/4", UNICODE_PI);
		else if(i == 60) fwplat->gui.textout_uni(x + vcos, y + vsin, "%lc/3", UNICODE_PI);
		else if(i == 90) fwplat->gui.textout_uni(x + vcos, y + vsin, "%lc/2", UNICODE_PI);
		else if(i == 120) fwplat->gui.textout_uni(x + vcos, y + vsin, "2%lc/3", UNICODE_PI);
		else if(i == 135) fwplat->gui.textout_uni(x + vcos, y + vsin, "3%lc/4", UNICODE_PI);
		else if(i == 150) fwplat->gui.textout_uni(x + vcos, y + vsin, "5%lc/6", UNICODE_PI);
		else if(i == 180) fwplat->gui.textout_uni(x + vcos, y + vsin, "%lc", UNICODE_PI);
		else if(i == 210) fwplat->gui.textout_uni(x + vcos, y + vsin, "7%lc/6", UNICODE_PI);
		else if(i == 225) fwplat->gui.textout_uni(x + vcos, y + vsin, "5%lc/4", UNICODE_PI);
		else if(i == 240) fwplat->gui.textout_uni(x + vcos, y + vsin, "4%lc/3", UNICODE_PI);
		else if(i == 270) fwplat->gui.textout_uni(x + vcos, y + vsin, "3%lc/2", UNICODE_PI);
		else if(i == 300) fwplat->gui.textout_uni(x + vcos, y + vsin, "5%lc/3", UNICODE_PI);
		else if(i == 315) fwplat->gui.textout_uni(x + vcos, y + vsin, "7%lc/4", UNICODE_PI);
		else if(i == 330) fwplat->gui.textout_uni(x + vcos, y + vsin, "11%lc/6", UNICODE_PI);
	}



	/*de kwadranten (liggen normaal gezien binnen de cirkel)*/
	fwplat->gui.sys_fontmetric(FONTNAME, (int)((FP_TYPE)(FONTH) * 1.25f), (int)((FP_TYPE)(FONTW) * 1.25f));
	if(kwadrant(lncos, lnsin) == 1) COLOR_FORE; else COLOR_WATERMARK;
	fwplat->gui.textout_uni(x + x / 2, y - y / 2, "Q1");
	if(kwadrant(lncos, lnsin) == 2) COLOR_FORE; else COLOR_WATERMARK;
	fwplat->gui.textout_uni(x - x / 2, y - y / 2, "Q2");
	if(kwadrant(lncos, lnsin) == 3) COLOR_FORE; else COLOR_WATERMARK;
	fwplat->gui.textout_uni(x - x / 2, y + y / 2, "Q3");
	if(kwadrant(lncos, lnsin) == 4) COLOR_FORE; else COLOR_WATERMARK;
	fwplat->gui.textout_uni(x + x / 2, y + y / 2, "Q4");

	/**
	 * Goniometrische getallen, test op maximum waarde straal cirkel
	 */
	vtan = 0;
	i = .7f;
	do{
		vhyp = sqrt(lncos*lncos + lnsin*lnsin);
		vcos = lncos / vhyp;
		vsin = lnsin / vhyp;
		if(vtan) break;
		vtan = (y <  x ? y : x);
		if(!vtan) return;
		if(vhyp > vtan * i) vhyp = vtan * i;
		else break;
		lncos = vcos * vhyp;
		lnsin = vsin * vhyp;
	}while(1);
	// graden theta in radialen op basis van de cosinus
	theta = acos(vcos);
	vcsc = 1.f / vsin;
	vsec = 1.f / vcos;
	vtan = vsin / vcos;
	vcot = 1.f / vtan;

	fwplat->gui.sys_fontmetric(FONTNAME, FONTH, FONTW);
	if(swDoWave){
		if(ln >= LIST_DIM){
			ln = LIST_DIM - 1;
			n = sizeof(FP_TYPE);
			memmove(ltime, ltime + 1, (LIST_DIM - 1) * n);
			memmove(lcos, lcos + 1, (LIST_DIM - 1) * n);
			memmove(lsin, lsin + 1, (LIST_DIM - 1) * n);
		}
		ltime[ln] = utsec( utdelta(utStart) );
		lcos[ln] = lncos;
		lsin[ln] = lnsin;
		ln++;
	}
	draw_waves(fwplat);

	COLOR_WATERMARK;
	fwplat->gui.textout_uni(x + x / 2.5,  1 * FONTH, "point P is in quadrant %d", kwadrant(lncos, lnsin));
	fwplat->gui.textout_uni(x + x / 2.5,  2 * FONTH, "hyp = %lc (cos%lc + sin%lc) = %.0f", UNICODE_SQUAREROOT, UNICODE_SQUARED, UNICODE_SQUARED, vhyp);
	COLOR_COS;
	fwplat->gui.textout_uni(x + x / 2.5,  3 * FONTH, "cos%lc = cos / hyp = %f", UNICODE_THETA, vcos);
	fwplat->gui.textout_uni(x + x / 2.5,  4 * FONTH, "%lc = arccos(cos%lc) => %.2f%lc", UNICODE_THETA, UNICODE_THETA, theta * ADEG, UNICODE_DEGREE);
	COLOR_SEC;
	fwplat->gui.textout_uni(x + x / 2.5,  5 * FONTH, "sec%lc = 1 / cos%lc = %f", UNICODE_THETA, UNICODE_THETA, vsec);
	COLOR_SIN;
	fwplat->gui.textout_uni(x + x / 2.5,  6 * FONTH, "sin%lc = sin / hyp = %f", UNICODE_THETA, vsin);
	fwplat->gui.textout_uni(x + x / 2.5,  7 * FONTH, "%lc = arcsin(sin%lc) = %.2f%lc", UNICODE_PHI, UNICODE_THETA, asin(vsin) * ADEG, UNICODE_DEGREE);
	COLOR_CSC;
	fwplat->gui.textout_uni(x + x / 2.5,  8 * FONTH, "csc%lc = 1 / sin%lc = %f", UNICODE_THETA, UNICODE_THETA, vcsc);
	COLOR_TAN;
	fwplat->gui.textout_uni(x + x / 2.5,  9 * FONTH, "tan%lc = sin / cos = %f", UNICODE_THETA, vtan);
	COLOR_COT;
	fwplat->gui.textout_uni(x + x / 2.5, 10 * FONTH, "cot%lc = 1 / tan%lc = %f", UNICODE_THETA, UNICODE_THETA, vcot);

	/*Goniometrische cirkel*/
	COLOR_WATERMARK;
	fwplat->gui.sys_fontmetric(FONTNAME, FONTH, FONTW);
	drawCircle(fwplat, x, y, vhyp);
	if(kwadrant(lncos, lnsin) > 2) theta = M_PI * 2.f - theta;
	COLOR_WATERMARK_DARK;
	i = ((FP_TYPE)(x) + (FP_TYPE)(y)) / 5.5f;
	drawArc(fwplat, x, y, i, 0.f, theta);
	drawArc(fwplat, x, y, i - 2.f, 0.f, theta);
	COLOR_WATERMARK;
	fwplat->gui.textout_uni(x + FONTW, y - FONTH, "%lc%lc %.2f%lc", UNICODE_ANGLE, UNICODE_THETA, fabs(theta * ADEG), UNICODE_DEGREE);


	/*hypotenusa in't blauw*/
	COLOR_HYP;
	fwplat->gui.line(x, y, x + lncos, y - lnsin);
	fwplat->gui.line(x + 1, y + 1, x + lncos + 1, y - lnsin + 1);
	linesegment(fwplat, x + lncos / 2, y - lnsin / 2 + FONTH, "HYP", vhyp);
	/*verlengde hypotenusa*/
	COLOR_WATERMARK_DARK;
	fwplat->gui.line(x + lncos, y - lnsin, x + lncos + vcos * sqrt(x*x+y*y), y - lnsin - vsin * sqrt(x*x+y*y));
	fwplat->gui.line(x, y, x + lncos - vcos * sqrt(x*x+y*y)*2, y - lnsin + vsin * sqrt(x*x+y*y)*2);

	/*Punt P*/
	COLOR_COS;
	fwplat->gui.textout_uni(x + (vcos * (vhyp + FONTW)), y - (vsin  * (vhyp + FONTH)), "P");
	COLOR_SIN;
	i = M_PI * (kwadrant(lncos, lnsin) / 2);
	n = 1;
	if(kwadrant(lncos, lnsin) == 2) n = -1;
	if(kwadrant(lncos, lnsin) == 3) n = -1;
	drawArc(fwplat, x, y, vhyp / 5, i, i + ( asin(vsin) * n) );
	drawArc(fwplat, x, y, 2 + vhyp / 5, i, i + ( asin(vsin) * n) );
	fwplat->gui.textout_uni(x + (vcos * (vhyp / 5.f)), y - (vsin * (vhyp / 5.f)), "%lc", UNICODE_PHI);

	/*cosinus in 't rood*/
	if(swWaveCos){
		COLOR_COS;
		fwplat->gui.line(x, y, x + lncos, y);
		fwplat->gui.line(x + 1, y + 1, x + lncos + 1, y + 1);
		fwplat->gui.dashline(x + lncos, y, x + lncos, y - lnsin);
		fwplat->gui.dashline(x + lncos + 1, y + 1, x + lncos + 1, y - lnsin + 1); //<--<< stippellijn
		fwplat->gui.lenpixels_uni(&txtwdt, 0, "COS%lc %.2f", UNICODE_THETA, vcos);
		linesegment(fwplat, x + lncos + (txtwdt * (lncos < 0 ? 0 : -1)), y + 4, "COS", lncos);
	}

	/*sinus in 't groen*/
	if(swWaveSin){
		COLOR_SIN;
		fwplat->gui.line(x, y, x, y - lnsin);
		fwplat->gui.line(x + 1, y + 1, x + 1, y - lnsin + 1);
		fwplat->gui.dashline(x, y - lnsin, x + lncos, y - lnsin);
		fwplat->gui.dashline(x + 1, y - lnsin + 1, x + lncos + 1, y - lnsin + 1);
		fwplat->gui.lenpixels_uni(&txtwdt, 0, "SIN%lc %.2f",UNICODE_THETA, vsin);
		txtwdt += FONTW;
		linesegment(fwplat, x + (txtwdt * (lncos < 0 ? 0 : -1)), y - lnsin, "SIN", lnsin);
	}

	/*tangens in't purper*/
	if(swWaveTan){
		COLOR_TAN;
		fwplat->gui.line(x + vhyp, y, x + vhyp, y - vtan * vhyp);
		fwplat->gui.line(x + vhyp + 1, y + 1, x + vhyp + 1, y - vtan * vhyp + 1);
		posy = y - vtan * vhyp;
		if(posy < 0) posy = 0;
		else if(posy > fwplat->gui.sys_maxy() - FONTH) posy = fwplat->gui.sys_maxy() - FONTH;
		fwplat->gui.textout_uni(x + vhyp + FONTW, posy, "TAN%lc %.2f",UNICODE_THETA, vtan);
	}

	/*cotangens*/
	if(swWaveCot){
		COLOR_COT;
		fwplat->gui.line(x, y - vhyp, x + vcot * vhyp, y - vhyp);
		fwplat->gui.line(x, y - vhyp + 1, x + vcot * vhyp, y - vhyp + 1);

		fwplat->gui.lenpixels_uni(&posy, 0, "COT%lc %.2f", UNICODE_THETA, vcot);
		posx = x + vcot * vhyp;
		if(posx < 0) posx = 0;
		else if(posx > fwplat->gui.sys_maxx() - posy) posx = fwplat->gui.sys_maxx() - posy;
		fwplat->gui.textout_uni(posx, y - vhyp - FONTH, "COT%lc %.2f", UNICODE_THETA, vcot);
	}

	/*cosecans en secans in't donker geel*/
	if(swWaveCsc || swWaveSec){
		COLOR_CANS;
		fwplat->gui.line(x, y - vcsc * vhyp, x + vsec * vhyp, y);
		fwplat->gui.line(x, y - vcsc * vhyp + 1, x + vsec * vhyp + 1, y);
		COLOR_CSC;
		fwplat->gui.lenpixels_uni(&posx, 0, "CSC%lc %.2f", UNICODE_THETA, vcsc);
		posy = y - vcsc * vhyp;
		if(posy < 0) posy = 0;
		else if(posy > fwplat->gui.sys_maxy()) posy = fwplat->gui.sys_maxy() - FONTH;
		fwplat->gui.textout_uni(x - posx, posy, "CSC%lc %.2f", UNICODE_THETA, vcsc);
		posx = x + vsec * vhyp;
		COLOR_SEC;
		fwplat->gui.lenpixels_uni(&posy, 0, "SEC%lc %.2f", UNICODE_THETA, vsec);
		if(posx < 0) posx = 0;
		else if(posx + posy > fwplat->gui.sys_maxx()) posx = fwplat->gui.sys_maxx() - posy;
		fwplat->gui.textout_uni(posx, y - FONTH, "SEC%lc %.2f", UNICODE_THETA, vsec);
	}

}

void onredraw(PLATFORM_APP *fwplat){
	FP_TYPE cx, cy, lnsin, lncos;

	if(fwplat){
		#ifdef __PLATFORM_GTK__
			FONTH = FONTW = 14;
		#endif
		#ifdef __PLATFORM_WINDOWS__
			FONTH = 12;
			FONTW = 8;
		#endif

		COLOR_FORE;
		COLOR_BACK;
		fwplat->gui.clrscr();

		cx = (FP_TYPE)fwplat->gui.sys_maxx() / 2.f;
		cy = (FP_TYPE)fwplat->gui.sys_maxy() / 2.f;

		fwplat->gui.sys_fontmetric(FONTNAME, FONTH, FONTW);
		FONTW = fwplat->gui.sys_fontwdt();
		FONTH = fwplat->gui.sys_fonthgt();

		if(scr && SCR_GONIO){
			/*cartesiaanse coordinaten*/
			drawFrame(fwplat, cx, cy);

			lncos = mX - (FP_TYPE)cx;
			lnsin = mY - (FP_TYPE)cy;
			/*goniometrische cirkel*/
			if(swDoRotate){
				fix_angle += angle_interval * rot_dir;
				if(fix_angle >= 2*M_PI) fix_angle = 0;
			}

			lncos = cos(fix_angle) * fix_rad;
			lnsin = -sin(fix_angle) * fix_rad;
			drawGonio(fwplat, cx, cy, lncos, -lnsin);

			COLOR_MENU;
			fwplat->gui.sys_fontmetric(FONTNAME, FONTH, FONTW);
			fwplat->gui.textout_uni(FONTW, fwplat->gui.sys_maxy() - FONTH * 4, "Press left mouse button and move the mouse");
			fwplat->gui.textout_uni(FONTW, fwplat->gui.sys_maxy() - FONTH * 3, "Options below (press the mentioned key between the brackets)...");
			fwplat->gui.textout_uni(FONTW, fwplat->gui.sys_maxy() - FONTH * 2, helpMotion);
			fwplat->gui.textout_uni(FONTW, fwplat->gui.sys_maxy() - FONTH * 1, helpWave);

			/*show the caret at the help position on the screen for about n seconds*/
			if(sw_caret_on){
				if(utsec(utdelta(utCaret)) > 10.f) sw_caret_on = 0;
				fwplat->gui.caret_set((sw_caret_on ? CARET_ON : CARET_OFF) , CARET_OVW, FONTW, fwplat->gui.sys_maxy() - FONTH * 4);
			}

		}else if(scr == SCR_HELP){
		}
	}
}

void ongui_begin(PLATFORM_APP *fwplat){
	if(fwplat){
		sw_caret_on = 1;
		utCaret = utnow();
		fix_rad = sqrt(fwplat->gui.sys_maxx() + fwplat->gui.sys_maxx() + fwplat->gui.sys_maxy() * fwplat->gui.sys_maxy()) / 2.f;
	} else
		exit(1);
}
void onexit(PLATFORM_APP *fwplat){
}

PLATFORM_APP* platform_app_init(){
	PLATFORM_APP *fwplat = 0;
	fwplat = (PLATFORM_APP*)malloc(sizeof(PLATFORM_APP));
	if(fwplat){
		memset(fwplat, 0, sizeof(PLATFORM_APP));
		fwplat->title = "MathCandy: Unit Circle";
		fwplat->onexit = onexit;
		fwplat->onkey = onkey;
		fwplat->onskey = onskey;
		fwplat->onmouse = onmouse;
		fwplat->onredraw = onredraw;
		fwplat->ongui_begin = ongui_begin;
	}
	return fwplat;
}
PLATFORM_APP* platform_app_destroy(PLATFORM_APP *fwplat){
	if(fwplat) free(fwplat);
	return 0;
}
