
/*
    Copyright (C) 2008 Vansteenhuyse Hubert <hubert.vansteenhuyse@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 ***********************************************************************************
 * list.c
 *
 *  Created on: Sat Nov 8, 2008 09:35
 *      Author: hubert.vansteenhuyse@freecode.be
 *
 * Purpose: keep a list of objects.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "muList.h"
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *++ LIJSTEN/STACKS
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int list_count = 0;

LIST* list_new(int block){
    LIST *lst = 0;
        lst = malloc(sizeof(LIST));
        if(lst){
        	list_count++;
            lst->cnt = 0;
            if(block < 0) block *= -1;
            if(block < LIST_MIN_BLOCK) block = LIST_MIN_BLOCK;
            lst->block = block * sizeof(LIST_DATA);
            lst->arr=0;
        }
    return lst;
}

LIST* list_destroy(LIST* lst){
    if(lst){
        if(lst->arr)free(lst->arr);
        free(lst);
        list_count--;
    }
    return 0;
}

LIST* list_reset(LIST *lst){
	if(lst)lst->cnt = 0;
	return lst;
}

LIST* list_ins(LIST* lst, LIST_INDEX idx, LIST_DATA elem){
	LIST_INDEX n;
	if(lst){
		if(idx >= 0 && idx < lst->cnt){
			lst = list_add(lst, 0);
			if(lst->arr){
				//memmove can't do it :(
				for(n = lst->cnt - 1; n > idx; n--)
					lst->arr[n] = lst->arr[n - 1];
				lst->arr[idx] = elem;
			}
		} else {
			lst = list_add(lst, elem);
		}

	}
    return lst;
}

LIST* list_add(LIST* lst, LIST_DATA elem){
    if(lst){
        if(!lst->arr){
            lst->arr = malloc(sizeof(LIST_DATA) * lst->block);
        } else if(!(lst->cnt % lst->block)){
            lst->arr = realloc(lst->arr, sizeof(LIST_DATA) * (lst->cnt + lst->block));
        }
        if(lst->arr){
        	lst->arr[lst->cnt] = elem;
        	lst->cnt++;
        }
    }
    return lst;
}

LIST* list_del(LIST* lst, LIST_INDEX idx){
	LIST_INDEX i;
    if(lst && lst->arr && idx >= 0 && idx < lst->cnt){
		for(i = idx; i < lst->cnt - 1; i++)
			lst->arr[i] = lst->arr[i + 1];

        lst->cnt--;
        if(lst->cnt && ( !(lst->cnt % lst->block) ) ){
            lst->arr = realloc(lst->arr, sizeof(LIST_DATA) * lst->cnt);
        }
    }
    return lst;
}
LIST* list_remove(LIST *lst, LIST_INDEX from, LIST_INDEX till){
	LIST_INDEX delta, i;
	if(lst && lst->cnt > 0){
		if(from > till){ //swap
			delta = till;
			till = from;
			from = delta;
		}
		if(till > lst->cnt - 1) till = lst->cnt - 1;
		if(from > lst->cnt - 1) return lst;
		delta = lst->cnt - 1 - till;

		i = 0;
		while(i < delta){
			lst->arr[from + i] = lst->arr[till + i + 1];
			i++;
		}

		delta = till - from;
        lst->cnt = lst->cnt - 1 - delta;
        //TODO: shrink list->arr in function of the amount of removed elements instead of the counter
        if(lst->cnt && ( !(lst->cnt % lst->block) ) )
            lst->arr = realloc(lst->arr, sizeof(LIST_DATA) * lst->cnt);
	}
	return lst;
}
LIST* list_swap(LIST* lst, LIST_INDEX idxa, LIST_INDEX idxb){
    void *t;
    if(lst && lst->arr && idxa >= 0 && idxa < lst->cnt && idxb >= 0 && idxb < lst->cnt){
        t = lst->arr[idxb];
        lst->arr[idxb] = lst->arr[idxa];
        lst->arr[idxa] = t;
    }
    return lst;
}

LIST* list_push(LIST* lst, LIST_DATA elem){
    return list_add(lst, elem);
}

LIST_DATA list_pop(LIST *lst){
    void *elem = 0;
    if(lst){
        elem = list_at(lst, lst->cnt - 1);
        list_del(lst, lst->cnt - 1);
    }
    return elem;
}



LIST* list_put(LIST* lst, LIST_INDEX idx, LIST_DATA val){
	LIST_INDEX 	blocks,		//allocated blocks in the list
				reqblocks;	//required blocks to write at position idx

        if(lst){
            idx++;
            blocks = lst->cnt / lst->block + (lst->cnt % lst->block ? 1 : 0);
            reqblocks = idx / lst->block + (idx % lst->block ? 1 : 0);

            if(!lst->arr){
                lst->arr = malloc(sizeof(LIST_DATA) * (reqblocks * lst->block));
            } else if(reqblocks > blocks){
                lst->arr = realloc(lst->arr, sizeof(LIST_DATA) * (reqblocks * lst->block));
            }

            if(lst->arr){
                lst->arr[idx - 1] = val;
                if(lst->cnt < idx)
                    lst->cnt = idx;
            }
        }
	return lst;
}

/********************************************************************************************************************
 * INDICES
 * Provides the data structure for indices on lists.
 * The logic must be implemented in the program using the structure.
 ********************************************************************************************************************/
LIST_INDICES* ilist_new(LIST *lst, int block){
	LIST_INDICES *ilist = 0;
	ilist = malloc(sizeof(LIST_INDICES));
	if(ilist){
		ilist->lst = lst;
		ilist->lstIndices = list_new(block);
		if(!ilist->lstIndices)
			ilist = ilist_destroy(ilist);
	}
	return ilist;
}

LIST_INDICES* ilist_destroy(LIST_INDICES *ilist){
	if(ilist){
		ilist = ilist_clear(ilist);
		list_destroy(ilist->lstIndices);

		free(ilist);
		ilist = 0;
	}
	return ilist;
}

LIST_INDICES* ilist_add(LIST_INDICES *ilist, LIST_INDEX ival, LIST_INDEX idx){
	LIST_IDX	*nidx;
	if(ilist && ilist->lstIndices){
		nidx = malloc(sizeof(LIST_IDX));
		if(nidx){
			nidx->idx = idx;
			nidx->ival = ival;
			ilist->lstIndices = list_add(ilist->lstIndices, LIST_ELEMENT_CAST(nidx));
		}
	}
	return ilist;
}

LIST_INDICES* ilist_clear(LIST_INDICES *ilist){
	LIST_INDEX	i;
	LIST_IDX	*idx;
	if(ilist && ilist->lstIndices){
		for(i = 0; i < ilist->lstIndices->cnt; i++){
			idx = (LIST_IDX*)list_at(ilist->lstIndices, i);
			if(idx) free(idx);
		}
		ilist->lstIndices->cnt = 0;
	}

	return ilist;
}

LIST_IDX* ilist_index(LIST_INDICES *ilist, LIST_INDEX ival, int swLowest){
	LIST_INDEX	i = 0;
	LIST_IDX	*idx, *nearest = 0;
	if(!ilist || !ilist->lstIndices || !ilist->lstIndices->arr || !ilist->lst) return 0;

	while(swLowest ? i < ilist->lstIndices->cnt : i >= 0){
		idx = (LIST_IDX*)list_at(ilist->lstIndices, i);
		i += (swLowest ? 1 : -1);
		if(!idx) continue;

		if(swLowest && idx->ival > ival) return nearest;
		if(!swLowest && idx->ival < ival) return nearest;

		nearest = idx;
	}
	return nearest;
}

