
/*
    Copyright (C) 2008 Vansteenhuyse Hubert <hubert.vansteenhuyse@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 ***********************************************************************************
 * list.h
 *
 *  Created on: Sat Nov 8, 2008 09:35
 *      Author: hubert.vansteenhuyse@freecode.be
 *
 * Purpose: keep a list of objects.
 */
#ifndef LIB_MULIST
#define LIB_MULIST

#ifdef __cplusplus
	extern "C" {
#endif

#define LIST_MIN_BLOCK			16
#define LIST_DATA 				unsigned long long*	/*data type for the list elements*/
#define LIST_INDEX				unsigned long long		/*data type of the indexes for accessing elements*/
#define LIST_ELEMENT_CAST(a) 	(LIST_DATA)(a)			/*cast macro*/

extern int list_count;


/*
 * LIJSTEN/STACKS
 */
typedef struct LIST{
	LIST_INDEX	cnt, block;
    LIST_DATA	*arr;
}LIST;

/*
 *routines
 */
LIST* list_new(int block);	/*block is the amount of memory the list will allocate when it needs more memory to store elements*/
LIST* list_destroy(LIST*);
LIST* list_reset(LIST*);
LIST* list_ins(LIST*, LIST_INDEX idx, LIST_DATA elem);
LIST* list_add(LIST*, LIST_DATA elem);
LIST* list_del(LIST*, LIST_INDEX idx);		//is a very slow operation, use list_remove for deleting more consecutive elements!
LIST* list_remove(LIST*, LIST_INDEX from, LIST_INDEX till);	//removes n elements
LIST* list_swap(LIST*, LIST_INDEX idxa, LIST_INDEX idxb);
LIST* list_push(LIST*, LIST_DATA);
LIST_DATA list_pop(LIST*);
LIST* list_put(LIST* lst, LIST_INDEX idx, LIST_DATA val);

/**
 * Macro's
 */

/*returns element i out of the list*/
#define list_at(lst, i) 	( !lst || !lst->arr || i < 0 || i >= lst->cnt ? 0 : lst->arr[i] )

/********************************************************************************************************************
 * INDICES
 * Provides the data structure and basic routines for indices on lists.
 * The logic for the indices must be implemented in the program using the structure.
 * The program using the indices is responsible to add them in a ascendantly sorted manner.
 * The structure allows to associate numerical values to indexes of elements inside a referenced list.
 * Routines are provided to create, add, retrieve, empty and destroy indices data structures.
 ********************************************************************************************************************/
typedef struct LIST_IDX{
	LIST_INDEX	ival,	//a value associated to idx
				idx;	//index location of an element from lst see LIST_INDICES
}LIST_IDX;
typedef struct LIST_INDICES{
	LIST 	*lst,			//the referenced list
			*lstIndices;	//the list containing the indices (LIST_IDX objects)
}LIST_INDICES;

LIST_INDICES* ilist_new(LIST *refList, int block);
LIST_INDICES* ilist_destroy(LIST_INDICES*);									//does not destroy the referenced list!
LIST_IDX* ilist_index(LIST_INDICES *ilist, LIST_INDEX ival, int swLowest);	//cf. macros below
LIST_INDICES* ilist_add(LIST_INDICES*, LIST_INDEX ival, LIST_INDEX idx);
LIST_INDICES* ilist_clear(LIST_INDICES*);

#define ilist_cnt(ili) 				( ili && ili->lstIndices ? ili->lstIndices->cnt : 0 )
#define ilist_index_lo( ili,  ival)	( ilist_index(ili, ival, 1) )	//returns the nearest smaller or equal LIST_IDX for ival
#define ilist_index_hi( ili,  ival)	( ilist_index(ili, ival, 0) )	//returns the nearest highest or equal LIST_IDX for ival

#ifdef __cplusplus
}
#endif

#endif /*LIB_MULIST*/
